
/*
 *	$Id: ax25.c,v 1.1.1.1 2002/02/15 22:36:40 Klaus Exp $
 *
 *	ax25.c		AX.25-Verwaltungroutinen. 
 *
 *			Hier sind u.a. die Funktionen zu finden, um
 *			Verbindungen aufzubauen oder abzubauen. 
 *
 *
 *	(c) 2000,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "ax25_l1.h"
#include "ax25_l2.h"
#include "ax25_dg.h"
#include "ax25.h"
#include "ax25_sup.h"
#include "mm.h"
#include "string.h"
#include "queue.h"

/*---------------------------------------------------------------------------*/


/*
 *	Initialisiert den AX.25-Controlblock <axcb> und
 *	traegt ihn in die Controlblock-Tabelle ax25tab ein.
 *
 *	Wrapper-Funktion fuer ax25_create(). Die Auslagerung von
 *	ax25_create nach ax25_sup.c wurde durchgefuehrt, um diese
 *	Quelldatei sauberer zu gestalten.
 *	
 */
struct AX25_cb xdata * ax25_open (unsigned char type)
{
	switch (type) {
	case TYPE_SEQPACKET:
	case TYPE_DGRAM:
		break;

	default:
		return NULL;
	}

	return ax25_create(type);
}


/*
 *	Gibt alle vom Controlblock reservierten Resourcen wieder
 *	frei und traegt ihn aus der Control-Block-Tabelle aus.
 *
 */
void ax25_close (struct AX25_cb xdata * axcb)
{
	if (axcb->type != TYPE_SEQPACKET) {
		ax25_destroy(axcb);
		return;
	}

	/*
	 *	Wenn die Verbindung bereits disconnected ist,
	 *	kann der ControlBlock auch sofort entsorgt
	 *	werden.
	 *
	 */
	if (axcb->l2.state == AX25_STATE_DISCONNECTED) {
		ax25_destroy(axcb);
		return;
	}

	/*
	 *	AX.25-CB in den Linger-Modus setzen. Das bedeutet,
	 *	dass das Layer 2 nun versucht alle Pakete in der
	 *	TX-Queue auszusenden, um danach die Verbindung
	 *	abzubauen.
	 *
	 */
	axcb->state = AX25_LINGER;

	/*
	 *	Wenn keine Daten mehr in der TX-Queue stehen, wird
	 *	nun der Disconnect angestossen.
	 *
	 */
	if (!q_first(&axcb->outq))
		l2_disconnect(axcb);
}


/*
 *	Stoesst ein Connect zu dem durch <path> angegebenen Ziel an.
 *	Bei DGRAM-AXCBs wird lediglich der komplette AX.25-Pfad
 *	gesetzt.
 *
 *	Im Fehlerfall wird 0 zurueckgegeben, ansonsten 1.
 *
 */
char ax25_connect (struct AX25_cb xdata * axcb, 
 		   struct AX25_smart_path xdata * path)
{
	if (axcb->state != AX25_IDLE)
		return 0;

	/*
	 *	Gibt es bereits einen entsprechenden Controlblock?
	 *	Wenn ja, dann wird connect mit einem Fehler
	 *	abgebrochen.
	 *
	 */
	if (ax25_lookup(&path->peer, axcb->type) != NULL)
		return 0;

	l1_mkpath(&axcb->path, path);
	l1_prepare_path(&axcb->path);
	
	switch (axcb->type) {
	case TYPE_DGRAM:
		axcb->state = AX25_ESTABLISHED;
		break;
	
	case TYPE_SEQPACKET:
		l2_connect(axcb);
		axcb->state = AX25_CONNECTING;
		break;
	
	default:
		return 0;
	}

	return 1;
}


/*
 *	Setzt den AX25-Controlblock <axcb> in den Zustand listen. Das
 *	Funktioniert nur mit TYPE_SEQPACKET Control-Bloecken!!
 *
 *	Im Fehlerfall wird 0 zurueckgegeben, ansonsten 1
 *
 */
char ax25_listen (struct AX25_cb xdata * axcb)
{
	if (axcb->state != AX25_IDLE ||
	    axcb->type != TYPE_SEQPACKET)
		return 0;

	axcb->state = AX25_LISTENING;

	return 1;
}


/*
 *	Akzeptiert den von der Gegenstelle empfangenen
 *	Verbindungswunsch. 
 *
 *	Wenn Accept geklappt hat wird 1 zurueckgegeben, ansonsten 0.
 *
 */
char ax25_accept (struct AX25_cb xdata * axcb)
{
	if (axcb->type != TYPE_SEQPACKET ||
	    axcb->state != AX25_LISTENING)
		return 0;
		
	if (l2_accept(axcb)) {
		axcb->state = AX25_ESTABLISHED;
		return 1;
	}
	
	return 0;
}


/*
 *	Wenn ein Event anliegt, wird er zurueckgeliefert, 
 *	ansonsten wird EVENT_NOEVENT oder bei leerer TX-Queue
 *	EVENT_TXQEMPTY zurueckgegeben.
 *
 */
char ax25_event (struct AX25_cb xdata * axcb)
{
	unsigned char event = axcb->event;

	axcb->event = EVENT_NOEVENT;

	/*
	 *	Wenn kein Event gespeichert ist und die
	 *	TX-Queue leer ist, wird EVENT_TXQEMPTY
	 *	zurueckgegeben.
	 *
	 */
	if (event == EVENT_NOEVENT)
		if (!q_first(&axcb->outq))
			return EVENT_TXQEMPTY;
			
	return event;
}



