
/*
 *	$Id: ax25_io.c,v 1.1.1.1 2002/02/15 22:36:40 Klaus Exp $
 *
 *	ax25_io.c	zusatz-IO-Funktionen fuer den AX.25 Protokollstack
 *
 *			Mit diesem Modul ist es z.B. moeglich die empfangenen
 *			Daten vom AX.25-Stack zeilenweise auszulesen.
 *
 *			TODO:	ax25_gets() mit effizienterem Algorithmus
 *				reimplementieren.
 *	
 *
 *	(c) 2000,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include <stdarg.h>

#include "mm.h"
#include "ax25.h"
#include "ax25_io.h"
#include "queue.h"
#include "buf.h"
#include "string.h"

/*
 *	Assembly-Puffer fuer die ax25_printf-Support-Funktionen
 *
 */
static char xdata astr[255];
static char xdata *aptr;
 
/*
 *	Puffer fuer ax25_printf + Schreibzeiger
 *
 */
static char xdata str[255];
static char xdata *sptr;

/*---------------------------------------------------------------------------*/


/*
 *	Initialisiert die IO-Instanz <io>
 *
 */
void ax25_init_io (struct AX25_io xdata *io)
{
	io->line = NULL;
	io->wptr = NULL;
	io->eptr = NULL;
}


/*
 *	Gibt die vom AX.25-IO-System reservierten
 *	Resourcen wieder frei und verwirft dabei ggf.
 *	schon fuer ax25_gets() eingelesene Zeichen.
 *
 */
void ax25_clean_io (struct AX25_cb xdata *axcb)
{
	if (axcb->io.line != NULL)
		free(axcb->io.line);

	ax25_init_io(&axcb->io);
}


/*---------------------------------------------------------------------------*/


/*
 *	Simple Wrapper-Funktion. Das macht erst so richtig Sinn, wenn
 *	der AX.25-Stack komplett Event-gesteurt funktioniert.
 *
 */
struct Buffer xdata * ax25_recv_buf (struct AX25_cb xdata *axcb)
{
	return q_dequeue(&axcb->inq);
}


/*
 *	Eine komplette Zeile aus der Empfangs-Queue von <axcb> auslesen.
 *	Es wird eine dynamisch erzeugte Kopie der Zeile zurueckgegeben.
 *	Die aufrufende Funktion muss die Zeile mit free() wieder loeschen!
 *
 *	Wenn keine Zeile ausgelesen werden konnte, wird NULL zurueckgegeben.
 *
 */
char xdata * ax25_gets (struct AX25_cb xdata *axcb,unsigned int maxlen)
{
	struct Buffer xdata * idata buf;
	char xdata * idata line;
	int idata ch;

	if ( !(buf = q_first(&axcb->inq)) )
		return NULL;

	if (!axcb->io.line) {
		if ( !(axcb->io.line = malloc(maxlen + 1)) )
			return NULL;

		axcb->io.wptr = axcb->io.line;
		axcb->io.eptr = axcb->io.line + maxlen;
	}

	while ( (ch = buf_getch(buf)) != -1 ) {
		if ((char)ch == '\r')
			goto line_ready;
	
		*axcb->io.wptr++ = (char)ch;

		if (axcb->io.wptr == axcb->io.eptr)
			goto line_ready;
	}

	buf = q_dequeue(&axcb->inq);
	buf_delete(buf);
	
	return NULL;

line_ready:

	*axcb->io.wptr = 0;

	line = axcb->io.line;
	ax25_init_io(&axcb->io);
	
	return line;
}


/*---------------------------------------------------------------------------*/


/*
 *	Legt im Assembly-Puffer das Call (incl. SSID) als
 *	ASCII-Text ab.
 *
 */
static void print_call (struct AX25_address xdata *addr)
{
	unsigned char n;
	unsigned char ssid;
	unsigned char xdata *call;

	call = addr->call;
	ssid = (addr->ssid >> 1) & 0x0f;

	for (n = 0; n < AX25_CALL_SIZE; n++) {
		if (*call == ' ' << 1)
			break;

		*aptr++ = *call++ >> 1;
	}

	if (ssid) {
		*aptr++ = '-';
		
		if (ssid > 9)
			*aptr++ = '1';

		ssid %= 10;

		*aptr++ = '0' + ssid;
	}
}


/*
 *	Kopiert den XDATA-String <s> in den Assembly-Puffer
 *
 */
static void print_string_x (char xdata *s)
{
	while (*s)
		*aptr++ = *s++;
}


/*
 *	Kopiert den CODE-String <s> in den Assembly-Puffer
 *
 */
static void print_string_c (char code *s)
{
	while (*s)
		*aptr++ = *s++;
}


/*
 *	Gibt den smart-Pfad <path> aus.
 *
 */
static void print_smart_path (struct AX25_smart_path xdata *path)
{
	print_call(&path->peer);

	if (path->digi_cnt) {
		unsigned char n;
	
		print_string_c(" via");

		for (n = 0; n != path->digi_cnt; n++) {
			print_string_c(" ");
			print_call(&path->digi[n]);
		}
	}
}


/*
 *	Gibt die Zahl <num> als Dezimalzahl in den Assembly-Puffer
 *	aus.
 *
 */
static void print_unum_10 (unsigned long num)
{
	char xdata a[10];
	unsigned char cnt;

	if (!num) {
		*aptr++ = '0';
		return;
	}

	cnt = 0;

	while (num) {
		a[cnt++] = '0' + (num % 10);
		num /= 10;
	}

	do {
		cnt--;
		*aptr++ = a[cnt];
	} while (cnt);
}


/* 
 *	Gibt die Zahl <num> als Hexadezimalzahl in den Assembly-Puffer
 *	aus. Dabei wird als erweiterte Ziffern ab <extdigit>
 *	durchnummeriert.
 *
 */
static void print_unum_16 (unsigned long num,unsigned char extdigit)
{
	char xdata a[10];
	unsigned char cnt;

	if (!num) {
		*aptr++ = '0';
		return;
	}

	cnt = 0;

	while (num) {
		unsigned char digit;

		digit = (unsigned char)(num & 0x0f);
	
		if (digit < 10)
			digit += '0';
		else
			digit += (extdigit - 10);
			
		a[cnt++] = digit;
		num >>= 4;
	}

	do {
		cnt--;
		*aptr++ = a[cnt];
	} while (cnt);
}


/*
 *	Kopiert den Assembly-Puffer ohne Formatierungen in
 *	den printf-Puffer
 *
 */
static void print ()
{
	char xdata *ptr = astr;

	while (ptr != aptr)
		*sptr++ = *ptr++;
}


/*
 *	Schreibt <cnt> <pad> Zeichen in den printf-Puffer
 *
 */
static void print_padding (unsigned char cnt,char pad)
{
	while (cnt) {
		*sptr++ = pad;
		cnt--;
	}
}


/*
 *	Kopiert den Assembly-Puffer linksbuendig 
 *	in den printf-Puffer und benutzt dabei das
 *	Padding-Zeichen <pad>
 *
 */
static void print_left (unsigned char width,char pad)
{
	unsigned char cnt;

	print();

	cnt = (unsigned char)(aptr - astr);

	if (cnt < width)
		print_padding(width - cnt,pad);
}


/*
 *	Kopiert den Assembly-Puffer rechtsbuendig in
 *	den printf-Puffer und benutzt dabei das
 *	Padding-Zeichen <pad>.
 *
 */
static void print_right (unsigned char width,char pad)
{
	unsigned char cnt;

	cnt = (unsigned char)(aptr - astr);

	if (cnt < width)
		print_padding(width - cnt,pad);

	print();
}


/*
 *	Funktion zerlegt die durch ax25_printf() erzeugte Ausgabe
 *	in Puffer und reicht sie an das AX.25-L2 weiter. Evtl.
 *	noch nicht volle Puffer werden zuerst aufgefuellt.
 *
 */
static unsigned char print_pushout (struct AX25_cb xdata *axcb)
{
	struct Buffer xdata * buf;
	struct Buffer xdata * new_buf;
	char xdata * ptr;
	unsigned int cnt;
	unsigned int tailroom;

	/* 	
	 *	Der Puffer, der spaeter evtl. gebraucht wird, muss
	 *	unbedingt vorher erzeugt werden, damit zu diesem
	 *	Zeitpunkt klar ist, ob die Ausgabe komplett
	 *	erfolgreich sein wird!
	 *
	 */
	if ( !(new_buf = buf_new(AX25_DATA_MAX)) )
		return 0;

	ptr = str;
	cnt = (unsigned int)(sptr - str);

	/*
	 *	erstmal probieren wir einen Teil der Daten in dem
	 *	letzten Puffer der Output-Queue unterzubringen.
	 *
	 */
	if ( (buf = q_last(&axcb->outq)) && 
	     !buf->locked &&
	     (tailroom = buf_tailroom(buf)) )
	{
		if (cnt < tailroom)
			tailroom = cnt;
	
		buf_append(buf,ptr,tailroom);
		cnt -= tailroom;

		if (!cnt) {
			buf_delete(new_buf);
			return 1;
		}

		ptr += tailroom;
	}
	
	/*
	 *	Fuer die restlichen Daten muss der neue Puffer
	 *	new_buf verwendet werden.
	 *
	 */
	buf_append(new_buf,ptr,cnt);
	
	if (!q_enqueue(&axcb->outq,new_buf)) {
		/*
		 *	Das ist ein Problem. Es konnte nicht der
		 *	komplette Text in der out-Queue gespeichert
		 *	werden! D.h. bevor ax25_printf() benutzt wird,
		 *	muss ueberprueft werden, ob noch Platz in der
		 *	out-Queue ist.
		 *
		 */
		buf_delete(new_buf);
		return 0;
	}

	return 1;
}


/*
 *	Baut einen String zusammen und legt ihn in der Sende-Queue
 *	von <axcb> ab.
 *
 *	Wenn ax25_printf() erfolgreich war, wird 1 zurueckgegeben,
 *	ansonsten 0;
 *
 *	%-Direktiven, die von ax25_printf() unterstuetzt werden:
 *
 *	%[0|-][0-9]+[l][%|a|c|m|s|u|x|X]
 *
 *	0	rechtsbuendig mit 0-Padding
 *	-	linksbuendig mit ' '-Padding
 *
 *	nnn	Feldgroesse
 *
 *	l	long
 *
 *	%	%
 *	a	AX.25-Call
 *	p	smart-Path
 *	b	Boolscher Wert <on|off>
 *	c	ein Zeichen
 *	m	MyCall
 *	s	XDATA-String
 *	S	CODE-String
 *	u	unsigned int (dezimal)
 *	x	unsigned int (hex) [0-9a-f]
 *	X	unsigned int (hex) [0-9A-F]
 *
 *
 */
unsigned char ax25_printf (struct AX25_cb xdata *axcb,
			    char code *format, ...)
{
	va_list arg;
	char c;
	char adjust_left;
	char zero_padding;
	char long_type;
	unsigned char width;
	
	if (q_isbusy(&axcb->outq))
		return 0;

	va_start(arg,format);
	
	sptr = str;

	while ( (c = *format++) ) {
		/* 
		 *	Alles was literal ist wird einfach
		 *	umkopiert.
		 *
		 */
		if (c != '%') {
			*sptr++ = c;
			continue;
		}

		aptr = astr;

		/*
		 *	Artung der Formatierung einlesen:
		 *	- lenksbuendig %-
		 *	- rechtsbuendig mit 0-Padding %0
		 *
		 */
		adjust_left = 0;
		zero_padding = 0;
		 
		c = *format++;

		switch (c) {
		case '-':
			adjust_left = 1;
			c = *format++;
			break;

		case '0':
			zero_padding = 1;
			c = *format++;
			break;
		}

		/*
		 *	Feldgroesse einlesen, falls sie 
		 *	angegeben ist.
		 *
		 */
		width = 0;

		while (c >= '0' && c <= '9') {
			width *= 10;
			width += c - '0';
			c = *format++;
		}

		/*
		 *	Optionaler type-Specifier
		 *
		 */
		if (c == 'l') {
			long_type = 1;
			c = *format++;
		} else
			long_type = 0;

		/*
		 *	Type decodieren
		 *
		 */
		switch (c) {
		case 'c':
			*aptr++ = va_arg(arg,char);
			break;

		case 'm':
			print_call(&l1_mycall);
			break;

		case 'a':
			print_call((struct AX25_address xdata *)
				   va_arg(arg,__XPTR__));
			break;

		case 'p':
			print_smart_path((struct AX25_smart_path xdata *)
					 va_arg(arg, __XPTR__));
			break;

		case 'b':
			if (va_arg(arg,unsigned char))
				print_string_c((char code *)"on");
			else
				print_string_c((char code *)"off");
			break;

		case 's':
			print_string_x((char xdata *)va_arg(arg,__XPTR__));
			break;
			
		case 'S':
			print_string_c((char code *)va_arg(arg,__CPTR__));
			break;

		case 'u':
			if (long_type)
				print_unum_10(va_arg(arg,unsigned long));
			else
				print_unum_10(va_arg(arg,unsigned int));
			break;
			
		case 'x':
			if (long_type)
				print_unum_16(va_arg(arg,unsigned long),'a');
			else
				print_unum_16(va_arg(arg,unsigned int),'a');
			break;
			
		case 'X':
			if (long_type)
				print_unum_16(va_arg(arg,unsigned long),'A');
			else
				print_unum_16(va_arg(arg,unsigned int),'A');
			break;

		case '%':
			*aptr++ = '%';
			break;
		}

		if (!c)
			break;

		/*
		 *	Assembly-Puffer muss nun in den printf-Puffer
		 *	umkopiert werden und dabei formatiert werden.
		 *
		 */
		if (adjust_left)
			print_left(width,' ');
		else if (zero_padding)
			print_right(width,'0');
		else if (width)
			print_right(width,' ');
		else
			print();
	}

	va_end(arg);

	return print_pushout(axcb);
}




