
#ifndef _AX25_L1_H_
#define _AX25_L1_H_

#include "buf.h"

/*
 *	AX.25 Paket-Typen
 *
 */
#define AX25_I		0x00
#define AX25_RR		0x01
#define AX25_RNR	0x05
#define AX25_REJ	0x09
#define AX25_SABM	0x2f
#define AX25_DISC	0x43
#define AX25_DM		0x0f
#define AX25_UA		0x63
#define AX25_FRMR	0x87
#define AX25_UI		0x03

/*
 *	Bitmasken zur Ausdecodierung von
 *	AX.25-Adressen
 *
 */
#define AX25_MASK_END	0x01
#define AX25_MASK_SSID	0x1e
#define AX25_MASK_RR	0x60
#define AX25_MASK_H	0x80
#define AX25_MASK_C	0x80

/*
 *	Frame-Attribute
 *
 */
#define AX25_COMMAND	0x80
#define AX25_RESPONSE	0x00
#define AX25_POLL	0x90
#define AX25_FINAL	0x10

/*
 *	Protokoll-IDs von Info-Bloecken
 *
 */
#define AX25_PID_TEXT	0xf0

/*
 *	Konstanten fuer AX.25-Adressen
 *	
 */
#define AX25_CALL_SIZE	6
#define AX25_DIGIS_MAX	8

/*
 *	Limitationen eines AX.25-Blocks in Bytes
 *
 */
#define AX25_BLK_MAX	328
#define AX25_PATH_MAX	70
#define AX25_DATA_MAX	255

/*
 *	Fehler-Flags im Datenbereich von FRMR-Paketen
 *
 */
#define AX25_W		0x01
#define AX25_X		0x02
#define AX25_Y		0x04
#define AX25_Z		0x08

/* 
 *	Invertiert das Flag f (Senderichtung <-> Empfangsrichtung)
 *
 */
#define INVERT_FLAG(f)	((f) ^ AX25_COMMAND)


/*
 *	Eine AX.25-Adresse
 *
 */
struct AX25_address {
	unsigned char	call[AX25_CALL_SIZE];	/* geshifted */
	unsigned char	ssid;			/* geshifted */
};


/*
 *	Besonders platzsparende Pfadangabe, um einen
 *	Weg in einem gut gerouteten AX.25-Netz
 *	anzugeben.
 *
 */
struct AX25_smart_path {
	struct AX25_address	peer;

	unsigned char		digi_cnt;
	struct AX25_address	digi[2];
};


/*
 *	Der komplette Pfad zu einer Gegenstation. 
 *
 */
struct AX25_path {
	unsigned char	path_size;

	union {
		/*
		 *	Pfad in seinen Komponenten
		 *
		 */
		struct {
			struct AX25_address	peer;
			struct AX25_address	local;
			struct AX25_address	digi[AX25_DIGIS_MAX];
		} addr;
	
		/*
		 *	fertig vorberechneter Pfad.. 
		 *
		 */
		unsigned char	path[AX25_PATH_MAX];
	} path;
};


/*
 *	Ein decodierter AX.25-Block
 *
 */
struct AX25_block {
	unsigned char	ctl_raw;	/* unbearbeitetes Control-Feld */
	unsigned char	ctl;		/* Controlfeld ohne NR, NS, PF-Flag */

	unsigned char	flag;		/* Comman/Response/Poll/Final */

	unsigned char	nr;		/* Empfangsfolgenzaehler */
	unsigned char	ns;		/* Sendefolgenzaehler */
	
	unsigned char	pid;		/* Layer3 Protocol-ID */
	
	struct Buffer	xdata *buf;	/* Daten */
};


char l1_xmit (struct AX25_path xdata *axpath,
	      unsigned char ctl,
	      unsigned char ns,
	      unsigned char nr,
	      unsigned char flag,
	      unsigned char pid,
	      struct Buffer xdata *buf);

#define l1_xmit_i(axpath,ns,nr,flag,pid,buf) \
	l1_xmit(axpath,AX25_I,ns,nr,flag,pid,buf)

#define l1_xmit_rr(axpath,nr,flag) \
	l1_xmit(axpath,AX25_RR,0,nr,flag,0,NULL)
	
#define l1_xmit_rnr(axpath,nr,flag) \
	l1_xmit(axpath,AX25_RNR,0,nr,flag,0,NULL)

#define l1_xmit_rej(axpath,nr,flag) \
	l1_xmit(axpath,AX25_REJ,0,nr,flag,0,NULL)

#define l1_xmit_sabm(axpath,flag) \
	l1_xmit(axpath,AX25_SABM,0,0,flag,0,NULL)

#define l1_xmit_disc(axpath,flag) \
	l1_xmit(axpath,AX25_DISC,0,0,flag,0,NULL)

#define l1_xmit_dm(axpath,flag) \
	l1_xmit(axpath,AX25_DM,0,0,flag,0,NULL)

#define l1_xmit_ua(axpath,flag) \
	l1_xmit(axpath,AX25_UA,0,0,flag,0,NULL)

#define l1_xmit_ui(axpath,flag,pid,buf) \
	l1_xmit(axpath,AX25_UI,0,0,flag,pid,buf)

#define l1_xmit_frmr(axpath,flag,buf) \
	l1_xmit(axpath,AX25_FRMR,0,0,flag,0,buf)


extern struct AX25_address xdata l1_mycall;
extern struct AX25_address code l1_nocall;

char l1_cmpaddr (struct AX25_address xdata *a1,struct AX25_address xdata *a2);
void l1_recv (struct Buffer xdata *buf);
void l1_prepare_path (struct AX25_path xdata *p);
void l1_mkpath (struct AX25_path xdata *p, struct AX25_smart_path xdata *sp);
void l1_set_mycall (struct AX25_address xdata *mycall);

#endif

