
/*
 *	$Id: ax25_l2.c,v 1.1.1.1 2002/02/15 22:36:40 Klaus Exp $
 *
 *	ax25_l2.c	AX.25 Layer 2
 *
 *
 *
 *	(c) 2000,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "system.h"
#include "ax25.h"
#include "ax25_sup.h"
#include "ax25_l1.h"
#include "ax25_l2.h"
#include "buf.h"
#include "timer.h"
#include "mm.h"
#include "config.h"

/*
 *	Konfigurations-Daten von dem Konfigurations-Modul
 *	beziehen.
 *
 */
#define l2_unack	conf.l2_unack
#define l2_n2 		conf.l2_n2
#define l2_t1 		conf.l2_t1
#define l2_t2		conf.l2_t2
#define l2_t3		conf.l2_t3

/*
 *	Event, der beim Bearbeiten des RX-Zweigs
 *	gespeichert wurde.
 *
 */
static unsigned char qevent;

/*--------------------------------------------------------------------------*/


/*
 *	Setzt die Layer2 Daten komplett zurueck
 *
 */
void l2_clear (struct AX25_l2data xdata *axl2) 
{
	axl2->state     = AX25_STATE_DISCONNECTED;
	axl2->connected = 0;
	
	axl2->own_busy  = 0;
	axl2->peer_busy = 0;
	axl2->rej_sent  = 0;

	axl2->n2	= l2_n2;
	axl2->vr	= 0;
	axl2->vs	= 0;
	axl2->va	= 0;

	t_stop(&axl2->t1);
	t_stop(&axl2->t2);
	t_stop(&axl2->t3);
}


/* 
 *	Initialisiert die Layer2 Daten
 *
 */
void l2_init (struct AX25_cb xdata *axcb)
{
	t_set(&axcb->l2.t1, l2_t1, T_RES1S);
	t_set(&axcb->l2.t2, l2_t2, T_RES100MS);
	t_set(&axcb->l2.t3, l2_t3, T_RES10S);
	
	l2_clear(&axcb->l2);
}


/*
 *	Loest einen Connect aus. 
 *
 */
void l2_connect (struct AX25_cb xdata *axcb)
{
	if (axcb->l2.state != AX25_STATE_DISCONNECTED)	
		return;

	l2_clear(&axcb->l2);
	axcb->l2.state = AX25_STATE_CON_PENDING;
	l1_xmit_sabm(&axcb->path,AX25_POLL);
	t_start(&axcb->l2.t1);
}


/*
 *	Stoesst einen Disconnect an
 *
 */
void l2_disconnect (struct AX25_cb xdata *axcb)
{
	if (axcb->l2.state == AX25_STATE_DISCONNECTED)
		return;

	l2_clear(&axcb->l2);
	axcb->l2.state = AX25_STATE_DISC_PENDING;
	l1_xmit_disc(&axcb->path,AX25_POLL);
	t_start(&axcb->l2.t1);
}


/*
 *	Akzeptiert den Verbindungswunsch der Gegenstelle. Wenn
 *	Accept funktioniert hat, wird 1 zurueckgegeben, ansonsten 0.
 *
 */
char l2_accept (struct AX25_cb xdata *axcb)
{
	if (axcb->l2.state != AX25_STATE_DISCONNECTED)
		return 0;

	/*
	 *	Genaugenommen duerfte hier nicht einfach mit dem 
	 *	FINAL-Flag geantwortet werden. Da in der Regel
	 *	das SABM mit dem POLL-Flag ausgesendet wird, wird
	 *	hier diese Vereinfachung verwendet. Das sollte
	 *	jedoch nicht zu Problemen fuehren...
	 *
	 */
	if (!l1_xmit_ua(&axcb->path,AX25_FINAL))
		return 0;

	l2_clear(&axcb->l2);
	axcb->l2.state = AX25_STATE_CONNECTED;
	t_start(&axcb->l2.t3);

	return 1;
}


/*--------------------------------------------------------------------------*/


/*
 *	Entfernt alle bis <nr> bestaetigten Puffer aus der Sendequeue.
 *	Wenn dabei ein Fehler unterlaufen ist und daraufhin der Zustand
 *	des Control-Blocks veraendert wurde, wird 0 zurueckgegeben.
 *	Wenn alles geklappt hat, wird 1 zurueckgegeben.
 *
 */
static char l2_ack (struct AX25_cb xdata *axcb,unsigned char nr)
{
	unsigned char idata ack_cnt;
	unsigned char idata sent_cnt;
	
	ack_cnt  = (nr - axcb->l2.va) & 7;

	if (!ack_cnt)
		return 1;
	
	sent_cnt = (axcb->l2.vs - axcb->l2.va) & 7;

	if (ack_cnt > sent_cnt) {
		/*
		 *	Protokollfehler!!!
		 *
		 *	TODO: FRMR-Handling
		 *
		 */
		l2_disconnect(axcb);
		return 0;
	}

	for (; ack_cnt; ack_cnt--) {
		struct Buffer xdata *buf;

		if ( (buf = q_dequeue(&axcb->outq)) != NULL )
			buf_delete(buf);
	}

	axcb->l2.vs = axcb->l2.va = nr;

	/*
	 *	siehe Link-Reset-Handling in ax25_recv_un().
	 *
	 */
	axcb->l2.connected = 1;

	if (!q_first(&axcb->outq))
		qevent = EVENT_TXQEMPTY;

	return 1;
}


/*
 *	Sendet das zum Status vom Control-Block passende Supervisory
 *	Frame aus, mit dem Flag <flag>.
 *
 *	Wenn das Senden erfolgreich war, wird 1 zurueckgegeben, 
 *	ansonsten 0.
 *
 */
static char l2_xmit_super (struct AX25_cb xdata *axcb,unsigned char flag)
{
	if (axcb->l2.own_busy)
		return l1_xmit_rnr(&axcb->path,axcb->l2.vr,flag);

	if (axcb->l2.rej_sent)
		return l1_xmit_rej(&axcb->path,axcb->l2.vr,flag);

	return l1_xmit_rr(&axcb->path,axcb->l2.vr,flag);
}


/*
 *	Bearbeitet ein empfangenes Supervisory-Frame.
 *	Im Fehlerfall wird 0 zurueckgegeben, ansonsten 1.
 *
 */
static char l2_recv_super (struct AX25_cb xdata *axcb,
			   struct AX25_block idata *axb)
{
	if (!l2_ack(axcb,axb->nr))
		return 0;

	if (axb->flag == AX25_POLL)
		l2_xmit_super(axcb,AX25_FINAL);

	return 1;
}


/*
 *	Sendet maximal maxcnt Info-Frames aus. Es wird immer vom
 *	Anfang der Sendequeue an ausgesendet! Wenn Frames
 *	ausgesendet wurden, wird auch der Timer1 gestartet!
 *	Ausgesendete Puffer werden gelocked, damit sie nicht
 *	nachtraeglich von den support-Funktionen modifiziert
 *	werden!
 *
 *	Es wird die Anzahl der tatsaechlich ausgesendeten Frames
 *	zurueckgegeben.
 *
 */
static char l2_xmit_info (struct AX25_cb xdata *axcb,unsigned char maxcnt)
{
	struct Buffer xdata * buf;
	unsigned char idata cnt;

	cnt = 0;
	axcb->l2.vs = axcb->l2.va;
	buf = q_first(&axcb->outq);

	while (buf && maxcnt) {
		if (!l1_xmit_i(&axcb->path,
			       axcb->l2.vs,
			       axcb->l2.vr,
			       AX25_COMMAND,
			       AX25_PID_TEXT,
			       buf))
		{
			break;
		}

		axcb->l2.vs = (axcb->l2.vs + 1) & 7;
		buf->locked = 1;

		maxcnt--;
		cnt++;

		buf = q_next(buf);
	}

	if (cnt) {
		t_start(&axcb->l2.t1);
		t_stop(&axcb->l2.t3);
	}

	return cnt;
}


/*
 *	Alle empfangenen I-Frames werden von dieser Funktion
 *	entgegengenommen und bearbeitet.
 *
 *	Im Fehlerfall wird 0 zurueckgegeben, ansonsten 1.
 *
 */
static char l2_recv_info (struct AX25_cb xdata *axcb,
                          struct AX25_block idata *axb)
{
	if (!l2_ack(axcb,axb->nr))
		return 0;

	t_stop(&axcb->l2.t2);

	if (axcb->l2.own_busy) {
		if (axb->flag == AX25_POLL)
			l2_xmit_super(axcb,AX25_FINAL);
		else
			t_start(&axcb->l2.t2);

		return 1;
	}

	if (axb->ns == axcb->l2.vr) {
		axcb->l2.rej_sent = 0;
	
		if (axb->pid == AX25_PID_TEXT) {
			if (q_enqueue(&axcb->inq,axb->buf)) {
				axb->buf = NULL;
				axcb->l2.vr = (axcb->l2.vr + 1) & 7;
			} else
				axcb->l2.own_busy = 1;
		} else {
			buf_delete(axb->buf);
			axb->buf = NULL;
			axcb->l2.vr = (axcb->l2.vr + 1) & 7;
		}
			
		if (axb->flag == AX25_POLL)
			l2_xmit_super(axcb,AX25_FINAL);
		else
			t_start(&axcb->l2.t2);

		return 1;
	} 

	if (axcb->l2.rej_sent)
		return 1;
	
	l1_xmit_rej(&axcb->path, axcb->l2.vr, INVERT_FLAG(axb->flag));
	axcb->l2.rej_sent = 1;

	return 1;
}


/*
 *	Funktion bearbeitet alle unnummerierten Frames, die in
 *	den Zustaenden AX25_STATE_CONNECTED und AX25_STATE_WAITING_ACK
 *	empfangen wurden.
 *
 */
static void l2_recv_un (struct AX25_cb xdata *axcb,
                        struct AX25_block idata *axb)
{
	switch (axb->ctl) {
	case AX25_SABM:
		/*
		 * 	Evtl. ist das kein Link-Reset! Wenn
		 *	wir auf ein SABM mit UA geantwortet
		 *	haben und die Gegenstelle das UA nicht
		 *	empfangen hat, besteht in Wirklichkeit
		 *	noch keine Verbindung! In diesem Fall
		 *	muss die Link-Reset Meldung an das
		 *	hoehere Layer unterdrueckt werden!
		 *
		 */
		if (axcb->l2.connected)
			qevent = EVENT_LINKRESET;

		l1_xmit_ua(&axcb->path,INVERT_FLAG(axb->flag));
		l2_clear(&axcb->l2);
		t_start(&axcb->l2.t3);
		axcb->l2.state = AX25_STATE_CONNECTED;
		break;

	case AX25_DISC:
		l1_xmit_ua(&axcb->path,INVERT_FLAG(axb->flag));
		/* kein break ! */
		
	case AX25_DM:
		l2_clear(&axcb->l2);
		qevent = EVENT_DISCONNECT;
		break;

	case AX25_FRMR:
		l2_disconnect(axcb);
	}
}


/*--------------------------------------------------------------------------*/


/*
 *	Funktion wird nach Ablauf des Timers 1 aufgerufen
 *
 */
static void l2_timer1 (struct AX25_cb xdata *axcb)
{
	switch (axcb->l2.state) {
	case AX25_STATE_CON_PENDING:
		if (!--axcb->l2.n2) {
			l2_clear(&axcb->l2);
			ax25_notify(axcb,EVENT_CONTOUT);
			break;
		}
		
		l1_xmit_sabm(&axcb->path,AX25_POLL);
		t_start(&axcb->l2.t1);
		break;

	case AX25_STATE_DISC_PENDING:
		if (!--axcb->l2.n2) {
			l2_clear(&axcb->l2);
			ax25_notify(axcb,EVENT_DISCONNECT);
			break;
		}

		l1_xmit_disc(&axcb->path,AX25_POLL);
		t_start(&axcb->l2.t1);
		break;

	case AX25_STATE_CONNECTED:
		axcb->l2.state = AX25_STATE_WAITING_ACK;
		/* kein break ! */
	
	case AX25_STATE_WAITING_ACK:
		if (!--axcb->l2.n2) {
			l2_clear(&axcb->l2);
			ax25_notify(axcb,EVENT_LINKFAILURE);
			break;
		}
	
		/* TODO: evtl. I-Poll */ 
		l2_xmit_super(axcb,AX25_POLL);
		t_start(&axcb->l2.t1);
		break;
	}
}


/*
 *	Funktion wird nach Ablauf des Timers 2 aufgerufen Wenn Frames
 *	ausgesendet wurden, wird auch der Timer1 gestartet!
 *
 */
static void l2_timer2 (struct AX25_cb xdata *axcb)
{
	switch (axcb->l2.state) {
	case AX25_STATE_CONNECTED:
		if (l2_xmit_info(axcb,l2_unack))
			break;

		/* kein break ! */
		
	case AX25_STATE_WAITING_ACK:
		l2_xmit_super(axcb,AX25_RESPONSE);
	}
}


/*
 *	Funktion wird nach Ablauf des Timers 3 aufgerufen
 *
 */
static void l2_timer3 (struct AX25_cb xdata *axcb)
{
	if (axcb->l2.state != AX25_STATE_CONNECTED)
		return;

	l2_xmit_super(axcb,AX25_POLL);
	t_start(&axcb->l2.t1);
}


/*--------------------------------------------------------------------------*/


/*
 *	Wenn keine Verbindung existiert, der Control-Block
 *	jedoch einen Connect annehmen kann, wird beim Empfang
 *	eines Frames diese Funktion aufgerufen.
 *
 */
static void l2_state_disconnected (struct AX25_cb xdata *axcb,
                                   struct AX25_block idata *axb)
{
	switch (axb->ctl) {
	case AX25_SABM:
		qevent = EVENT_CONREQ;
		break;

	case AX25_DM:
		/*
		 *	Auf ein DM darf nicht mit einem DM geantwortet
		 *	werden, da es sonst DM-Loops gibt!
		 *
		 */
		break;
		
	default:
		l1_xmit_dm(&axcb->path,INVERT_FLAG(axb->flag));
	}
}


/*
 *	Wenn versucht wird eine Verbindung zur Gegenstelle
 *	aufzubauen und ein Frame empfangen wurde, dann wird
 *	diese Funktion aufgerufen.
 *
 */
static void l2_state_con_pending (struct AX25_cb xdata *axcb,
			          struct AX25_block idata *axb)
{
	switch (axb->ctl) {
	case AX25_UA:
		l2_clear(&axcb->l2);
		axcb->l2.state = AX25_STATE_CONNECTED;
		axcb->l2.connected = 1;
		t_start(&axcb->l2.t3);
		qevent = EVENT_CONNECT;
		break;

	case AX25_DM:
	case AX25_FRMR:
		l2_clear(&axcb->l2);
		qevent = EVENT_CONREFUSED;

	/*
	 *	Alle weiteren Frame-Typen muessen ignoriert werden. Es kann
	 *	sein, dass die Gegenstelle unser SABM mit einer UA, I - Folge
	 *	bestaetigt hat. Wenn wir in diesem Fall nur das I empfangen,
	 *	darf kein DM gesendet werden!
	 *
	 */
	}
}


/*
 *	Wenn versucht wird die bestehende Verbindung abzubauen
 *	und ein Frame empfangen wurde, so wird diese
 *	Funktion aufgerufen.
 *
 */
static void l2_state_disc_pending (struct AX25_cb xdata *axcb,
				   struct AX25_block idata *axb)
{
	switch (axb->ctl) {
	case AX25_FRMR:
		l1_xmit_dm(&axcb->path,INVERT_FLAG(axb->flag));
		/* kein break! */
	
	case AX25_UA:
	case AX25_DM:
		l2_clear(&axcb->l2);
		qevent = EVENT_DISCONNECT;
		break;
	
	default:
		break;
	}
}


/*
 *	Wenn eine Verbindung aufgebaut ist und ein Frame
 *	empfangen wurde, wird diese Funktion aufgerufen.
 *
 */
static void l2_state_connected (struct AX25_cb xdata *axcb,
			        struct AX25_block idata *axb)
{
	switch (axb->ctl) {
	case AX25_I:
		t_stop(&axcb->l2.t1);
		if (!l2_recv_info(axcb,axb))
			return;
		break;
	
	case AX25_RR:
		t_stop(&axcb->l2.t1);
		axcb->l2.peer_busy = 0;
		if (!l2_recv_super(axcb,axb))
			return;
		l2_xmit_info(axcb,l2_unack);
		break;
	
	case AX25_RNR:
		t_stop(&axcb->l2.t1);
		axcb->l2.peer_busy = 1;
		if (!l2_recv_super(axcb,axb))
			return;
		break;
	
	case AX25_REJ:
		t_stop(&axcb->l2.t1);
		axcb->l2.peer_busy = 0;
		if (!l2_recv_super(axcb,axb))
			return;
		l2_xmit_info(axcb,1);
		break;
		
	default:
		l2_recv_un(axcb,axb);
		return;
	}

	if (!t_isrunning(&axcb->l2.t1))
		t_start(&axcb->l2.t3);
}


/*
 *	Wenn auf eine Bestaetigung gewartet wird und ein Frame
 *	empfangen wurde, wird diese Funktion aufgerufen.
 *
 */
static void l2_state_waiting_ack (struct AX25_cb xdata *axcb,
			          struct AX25_block idata *axb)
{
	switch (axb->ctl) {
	case AX25_I:
		if (!l2_recv_info(axcb,axb))
			return;
		break;
		
	case AX25_RR:
		axcb->l2.peer_busy = 0;
		if (!l2_recv_super(axcb,axb))
			return;
		break;
	
	case AX25_RNR:
		axcb->l2.peer_busy = 1;
		if (!l2_recv_super(axcb,axb))
			return;
		break;
	
	case AX25_REJ:
		axcb->l2.peer_busy = 0;
		if (!l2_recv_super(axcb,axb))
			return;
		break;
		
	default:
		l2_recv_un(axcb,axb);
		return;
	}

	if (axb->flag == AX25_FINAL) {
		axcb->l2.state = AX25_STATE_CONNECTED;
		axcb->l2.n2 = l2_n2;
		t_stop(&axcb->l2.t1);
		t_start(&axcb->l2.t3);
	}
}


/*
 *	Nimmt alle vom L1 aussortierten AX.25-Bloecke entgegen und
 *	wertet sie nach den Regeln des AX.25-Protokoll Version 2
 *	aus.
 *
 */
void l2_recv (struct AX25_cb xdata *axcb,struct AX25_block idata *axb)
{
	qevent = EVENT_NOEVENT;

	switch (axcb->l2.state) {
	case AX25_STATE_DISCONNECTED:
		l2_state_disconnected(axcb,axb);
		break;
		
	case AX25_STATE_CON_PENDING:
		l2_state_con_pending(axcb,axb);
		break;
	
	case AX25_STATE_DISC_PENDING:
		l2_state_disc_pending(axcb,axb);
		break;
		
	case AX25_STATE_CONNECTED:
		l2_state_connected(axcb,axb);
		break;
		
	case AX25_STATE_WAITING_ACK:
		l2_state_waiting_ack(axcb,axb);
		break;
		
	default:
		l2_disconnect(axcb);
	}

	if (axb->buf)
		buf_delete(axb->buf);

	if (qevent != EVENT_NOEVENT)
		ax25_notify(axcb,qevent);
}


/*--------------------------------------------------------------------------*/


/*
 *	Fuehrt zyklische Aufgaben im AX.25-Stack durch.
 *
 *	TODO:	Diese Aufgaben sollten besser Ereignisgesteuert
 *		sein!
 *
 */
void l2_run (struct AX25_cb xdata * axcb)
{
	if (t_isexpired(&axcb->l2.t1))
		l2_timer1(axcb);

	if (t_isexpired(&axcb->l2.t2))
		l2_timer2(axcb);

	if (t_isexpired(&axcb->l2.t3))
		l2_timer3(axcb);

	if (axcb->l2.state != AX25_STATE_CONNECTED)
		return;

	if (q_first(&axcb->outq) &&
	    !axcb->l2.peer_busy &&
	    !t_isrunning(&axcb->l2.t1) &&
	    !t_isrunning(&axcb->l2.t2))
	{
		l2_xmit_info(axcb,l2_unack);
	} 

	/*
	 *	Der Busy-Zustand wird erst dann wieder verlassen,
	 *	wenn die input-Queue leer ist.
	 *
	 */
	if (axcb->l2.own_busy && !q_first(&axcb->inq)) {
		axcb->l2.own_busy = 0;
		l2_xmit_super(axcb,AX25_POLL);
		t_start(&axcb->l2.t1);
		t_stop(&axcb->l2.t3);
	}
}





