
/*
 *	ax25_sup.c	AX.25-Support-Funktionen fuer das
 *			Layer1 und Layer2. Wurde aus
 *			ax25.c extrahiert.
 *
 *
 *
 *	(c) 2000,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "ax25_sup.h"
#include "ax25_dg.h"
#include "ax25.h"
#include "queue.h"
#include "mm.h"
#include "ax25_io.h"

/* 
 *	Anzahl der AX25-Controlbloecke in der Controlblock-Liste
 *
 */
unsigned int ax25_cbcnt = 0;

/*
 *	Liste aller AX.25-Controlbloecke
 *
 */
static struct AX25_cb xdata * axcbtab = NULL;

/*
 *	Im naechsten Zyklus zu barbeitender AX.25-Controlblock
 *
 */
static struct AX25_cb xdata * axcb_next = NULL;

/*--------------------------------------------------------------------------*/


/*
 *	Erzeugt einen neuen AX.25-Controlblock und
 *	initialisiert ihn.
 *
 *	Im Fehlerfall wird NULL zurueckgegeben ansonsten
 *	der Zeiger auf den neuen Controlblock.
 *	
 */
struct AX25_cb xdata * ax25_create (unsigned char type)
{
	struct AX25_cb xdata * idata axcb;

	axcb = malloc(sizeof(struct AX25_cb));

	if (!axcb)
		return NULL;

	axcb->type = type;
	axcb->state = AX25_IDLE;
	axcb->event = EVENT_NOEVENT;

	q_init(&axcb->inq,8);
	q_init(&axcb->outq,4);

	ax25_init_io(&axcb->io);

	if (type == TYPE_SEQPACKET) 
		l2_init(axcb);

	if (axcbtab)
		axcbtab->prev = axcb;

	axcb->prev = NULL;
	axcb->next = axcbtab;
	axcbtab = axcb;

	ax25_cbcnt++;

	return axcb;
}


/*
 *	Setzt den kompletten AX25-Controlblock
 *	zurueck.
 *	
 */
void ax25_clear (struct AX25_cb xdata * axcb)
{
	q_clear(&axcb->inq);
	q_clear(&axcb->outq);

	ax25_clean_io(axcb);

	if (axcb->type == TYPE_SEQPACKET)
		l2_clear(&axcb->l2);

	axcb->state = AX25_IDLE;
	axcb->event = EVENT_NOEVENT; 
}


/*
 *	Gibt alle vom Controlblock belegten Resourcen frei
 *	und entfernt danach den Controlblock selbst.
 *
 */
void ax25_destroy (struct AX25_cb xdata * axcb)
{
	if (axcb_next == axcb)
		axcb_next = axcb_next->next;

	ax25_clear(axcb);

	if (axcb->next)
		axcb->next->prev = axcb->prev;

	if (axcb->prev)
		axcb->prev->next = axcb->next;
	else
		axcbtab = axcbtab->next;

	free(axcb);

	ax25_cbcnt--;
}


/*
 *	Sucht den AX.25-Controlblock heraus, der zum Peer <peer>
 *	gehoert.
 *
 */
struct AX25_cb xdata * ax25_lookup (struct AX25_address xdata *peer,
			            unsigned char type)
{
	struct AX25_cb xdata * idata axcb;

	for (axcb = axcbtab; axcb != NULL; axcb = axcb->next) {
		if (axcb->type != type)
			continue;
	
		switch (axcb->state) {
		case AX25_CONNECTING:
		case AX25_ESTABLISHED:
		case AX25_LINGER:
			break;

		default:
			continue;
		}
	
		if (l1_cmpaddr(peer,&axcb->path.path.addr.peer) == 1)
			break;
	}

	return axcb;
}


/*
 *	Gibt den ersten gefundenen Listener zurueck. Es werden
 *	nur alle SEQPACKET Control-Bloecke durchsucht. Wenn
 *	kein Listener gefunden wurde, wird NULL zurueckgegeben.
 *
 */
struct AX25_cb xdata * ax25_listener ()
{
	struct AX25_cb xdata * idata axcb;

	for (axcb = axcbtab; axcb != NULL; axcb = axcb->next)
		if (axcb->state == AX25_LISTENING)
			break;
	
	return axcb;
}


/*
 *	Speichert das angegebene Ereignis im <axcb>, damit es
 *	dem hoeheren Layer weitergemeldet wird.
 *
 */
void ax25_notify (struct AX25_cb xdata * axcb,char event)
{
	switch (event) {
	case EVENT_TXQEMPTY:
		if (axcb->state == AX25_LINGER)
			l2_disconnect(axcb);
		break;
		
	case EVENT_CONNECT:
		axcb->state = AX25_ESTABLISHED;
		break;

	case EVENT_NOEVENT:
	case EVENT_CONREQ:
	case EVENT_LINKRESET:
		break;

	default:
		/*
		 *	Bei allen Events, die einen Verbindungsabbruch oder
		 *	einen Fehler melden, wird der AX.25-Controlblock 
		 *	zurueckgesetzt und ggf. geloescht.
		 *
		 */
		if (axcb->state == AX25_LINGER) {
			ax25_destroy(axcb);
			return;
		}
	
		ax25_clear(axcb);
		axcb->state = AX25_IDLE;
	}

	axcb->event = event;
}


/*
 *	Bearbeitet den naechsten anstehenden AX.25-Controlblock
 *
 */
void ax25_run ()
{
	struct AX25_cb xdata * idata axcb;

	if (!axcb_next) {
		if (!axcbtab)
			return;

		axcb_next = axcbtab;
	}

	axcb = axcb_next;
	axcb_next = axcb_next->next;
		
	switch (axcb->type) {
	case TYPE_SEQPACKET:
		l2_run(axcb);
		break;
	
	case TYPE_DGRAM:
		dg_run(axcb);
	}
}


