
/*
 *	$Id: boot.c,v 1.1.1.1 2002/02/15 22:36:40 Klaus Exp $
 *
 *	boot.c		Funktionen zum starten des Systems
 *
 *
 *
 *	(c) 2001,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "boot.h"
#include "system.h"
#include "flash.h"
#include "version.h"
#include "mm.h"
#include "crc.h"
#include "bank.h"
#include "eeprom.h"
#include "ax25.h"
#include "ax25_io.h"

/*
 *	Zustaende der Flash-Baenke
 *
 */
#define STATE_DISABLED	0
#define STATE_ENABLED	1

/*
 *	Blockgroesse, die bei der Ueberpruefung einer
 *	Flash-Bank verwendet wird.
 *
 */
#define FLASH_BLKSIZE 	256
 
/*
 *	Boot-Daten, die aus dem EEPROM gelesen wurden.
 *
 */
static struct BootConf xdata bconf;

/*
 *	Die Firmware-Informationen, dieser Firmware
 *
 */
struct FWInfo code fw_info = 
{	
	FW_MAGIC,

	VERSION_MAJOR,
	VERSION_MINOR,
	VERSION_PATCH
};

/*--------------------------------------------------------------------------*/


/*
 *	Gibt die Boot-Konfiguration aus
 *
 */
void boot_print_state (struct AX25_cb xdata *axcb)
{
	unsigned char bank;
	unsigned char active_bank;

	if (IS_CSEG_CODE())
		active_bank = BOOT_EPROM;
	else 
		active_bank = FLASH_ACTIVE_BANK();

	ax25_printf(axcb, "bank  state  start-cnt  version\r");

	for (bank = BOOT_FLASH0; bank <= BOOT_FLASH1; bank++) {
		struct Bank xdata *b = &bconf.bank[bank];
		char ch;

		ch = (bank == active_bank) ? '*' : ' ';
	
		ax25_printf(axcb, 
		            "%u%c%9b%11u  %u.%u.%u\r",
			    (unsigned int)bank,
			    ch,
			    b->state,
			    b->start_cnt,
			    (unsigned int)b->version_major,
			    (unsigned int)b->version_minor,
			    (unsigned int)b->version_patch);
	}			   
}


/*
 *	Prueft, ob die Firmware in der Flash-Bank korrekt eingespielt
 *	wurde und ob der notwendige boot-Code vorhanden ist.
 *
 *	Wenn die Firmware Ok ist wird 1 zurueckgegeben, ansonsten 0.
 *
 */
char boot_check (unsigned char bank)
{	
	unsigned char xdata *buf;
	unsigned long addr;
	unsigned int crc;
	struct BankInfo xdata binf;
	struct FWInfo xdata fwinf;

	/*
	 *	zuerst wird ueberprueft, ob sich eine gueltige Firmware
	 *	in der flash-Bank befindet.
	 *
	 */
	flash_read(bank, (unsigned int)&bank_info, 
	           &binf, sizeof(binf));

	flash_read(bank, (unsigned int)binf.fw_info, 
	           &fwinf, sizeof(fwinf));

	if (fwinf.magic != FW_MAGIC)
		return 0;

	/*
	 *	Ueberpruefung der Firmware anhand der CRC.
	 *
	 */
	if ( !(buf = malloc(FLASH_BLKSIZE)) )
		return 0;

	crc = CRC_THP_INITIAL;

	for (addr = 0; addr < 0x10000UL; addr += FLASH_BLKSIZE) {
		flash_read(bank, addr, buf, FLASH_BLKSIZE);
		crc = crc_thp_buf(crc, buf, FLASH_BLKSIZE);
	}

	free(buf);

	if (bconf.bank[bank].crc != crc)
		return 0;

	return 1;
}


/*
 *	Laedt die default boot-Konfiguration
 *
 */
static void boot_load_defaults ()
{
	struct Bank xdata *b;
	
	bconf.magic = BOOT_CONF_MAGIC;
	bconf.active_bank = BOOT_EPROM;

	for (b = &bconf.bank[0]; b <= &bconf.bank[1]; b++) {
		b->state     = STATE_DISABLED;
		b->crc       = 0;
		b->start_cnt = 0;

		b->version_minor = 0;
		b->version_major = 0;
		b->version_patch = 0;
	}
}


/*
 *	Schreibt die aktuelle Boot-Konfiguration in
 *	das EEPROM
 *
 */
static void boot_conf_write ()
{
	bconf.crc = crc_crc16_buf(CRC_CRC16_INITIAL, 
                                  (unsigned char xdata *)&bconf, 
				  sizeof(bconf) - sizeof(bconf.crc));
				  
	ee_write(BOOT_CONF_ADDR, 
		 (unsigned char xdata *)&bconf, 
		 sizeof(bconf));
}


/*
 *	Setzt die Bank, die beim booten gestartet werden soll.
 *	Wenn die bank aktiviert werden konnte wird 1 zurueckgegeben,
 *	ansonsten 0.
 *
 */
char boot_set_bank (unsigned char bank)
{
	switch (bank) {
	case BOOT_FLASH0:
	case BOOT_FLASH1:
		if (bconf.bank[bank].state != STATE_ENABLED)
			return 0;
		break;

	case BOOT_EPROM:
		break;

	default:
		return 0;
	}

	bconf.active_bank = bank;
	boot_conf_write();
	return 1;
}


/*
 *	Aktiviert die angegebene Bank
 *
 */
void boot_enable_bank (unsigned char bank, unsigned int crc)
{
	struct Bank xdata *b = &bconf.bank[bank];

	b->state     = STATE_ENABLED;
	b->crc       = crc;
	b->start_cnt = 0;
	
	b->version_major = 0;
	b->version_minor = 0;
	b->version_patch = 0;

	if (IS_CSEG_CODE()) {
		struct BankInfo xdata binf;
		struct FWInfo xdata fwinf;

		flash_read(bank, (unsigned int)&bank_info, 
			   &binf, sizeof(binf));

		flash_read(bank, (unsigned int)binf.fw_info, 
			   &fwinf, sizeof(fwinf));

		if (fwinf.magic == FW_MAGIC) {
			b->version_major = fwinf.version_major;
			b->version_minor = fwinf.version_minor;
			b->version_patch = fwinf.version_patch;
		}
	}
	
	boot_conf_write();
}


/*
 *	Deaktiviert die angegebene Bank
 *
 */
void boot_disable_bank (unsigned char bank)
{
	bconf.bank[bank].state = STATE_DISABLED;
	boot_conf_write();
}


/*
 *	Startet die in der boot-Konfiguration angegebene Firmware	
 *
 */
void boot ()
{
	unsigned int crc;
	
	/*
	 *	Boot-Konfiguration einlesen und ueberpruefen. Wenn
	 *	die Daten nicht korrekt sind, wird die default-
	 *	Konfiguration geladen und sicherheitshalber ein
	 *	Reset durchgefuert.
	 *
	 */
	ee_read(BOOT_CONF_ADDR, &bconf, sizeof(bconf));
	
	crc = crc_crc16_buf(CRC_CRC16_INITIAL, 
	                    (unsigned char xdata *)&bconf, 
			    sizeof(bconf) - sizeof(bconf.crc));
			    
	if (bconf.magic != BOOT_CONF_MAGIC || crc != bconf.crc) {
		boot_load_defaults();
		boot_conf_write();
		RESET();
	}

	/*
	 *	Eine Flash-Bank ist bereits eingeblendet. D.h die Umschaltung
	 *	hat schon stattgefunden und der normale start-Vorgang
	 *	muss fortgesetzt werden.
	 *	Zuvor wird jedoch noch der starup-Counter aktualisiert.
	 *
	 */
	if (IS_CSEG_FLASH()) {
		unsigned char bank = FLASH_ACTIVE_BANK();

		bconf.bank[bank].start_cnt++;
		boot_conf_write();

		return;
	}

	/*
	 *	Wenn der TEST-Taster beim boot-Prozess gedrueckt wird
	 *	wird die Firmware aus dem EPROM benutzt.
	 *
	 */
	if (TEST_KEY)
		return;

	/*
	 *	Nur wenn die Flash-Baenke gestartet werden sollen,
	 *	muessen wir weitermachen. Ansosten wird der normale
	 *	start-Vorgang (in xos.c) fortgefuert.
	 *	
	 *	Um dem User zu signalisieren was passiert, wird die
	 *	ACT-LED fuer Bank 0 und die CON-LED fuer die Bank 1
	 *	eingeschaltet.
	 *
	 */
	switch (bconf.active_bank) {
	case BOOT_FLASH0:
		ON(LED_ACT);
		break;
		
	case BOOT_FLASH1:
		ON(LED_CON);
		break;
		
	case BOOT_EPROM:
	default:
		return;
	}

	/*
	 *	Wenn die Ueberpruefung der zu startenden Firmware
	 *	fehlschlaegt, wird die EPROM-Firmware verwendet.
	 *	Zuvor wird jedoch der Flash-Bank-Zustand und die
	 *	aktive Flash-Bank in der boot-Konfiguration korrigiert.
	 *
	 */
	if (!boot_check(bconf.active_bank)) {
		unsigned char bank = bconf.active_bank;
	
		bconf.active_bank = BOOT_EPROM;
		boot_disable_bank(bank);
		return;
	}

	OFF(LED_ACT);
	OFF(LED_CON);
	
	/*
	 *	Und nun wird die gewuenschte Firmware gestartet.
	 *	Dazu wird P3 auf ~0x20 gesetzt. Naeheres siehe
	 *	in bank.s
	 *
	 */
	EA = 0;
	FLASH_ACTIVATE(bconf.active_bank);
	WATCHDOG();
	bank_start();

	ON(LED_ACT);
	ON(LED_CON);
	RESET();
}



