
/*
 *	$Id: config.c,v 1.6 2002/05/12 12:12:43 klaus Exp $
 *
 *	config.c	Konfigurations-Modul
 *
 *
 *
 *	(c) 2001,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "system.h"
#include "config.h"
#include "eeprom.h"
#include "crc.h"
#include "ax25_l1.h"
#include "ax25_l2.h"
#include "string.h"
#include "scc.h"
#include "master.h"
#include "types.h"
#include "repeater.h"
#include "sched.h"

/*
 *	Konfigurationssegmente fuer conf_load_defaults()
 *
 */
#define SEG_HOT_DATA	1
#define SEG_DATA	2

/*
 *	Konfigurationsdaten
 *
 */
struct Config xdata conf;

/*---------------------------------------------------------------------------*/


/*
 *	Setzt die default-Parameter.
 *
 *	Mit <seg> kann festgelegt werden welches Segment
 *	zurueckgesetzt wird.
 *
 */
static void conf_load_defaults (unsigned char seg)
{
	if (seg & SEG_HOT_DATA) {
		/*
		 *	Passwort loeschen.
		 *
		 */
		*conf.hot_data.passwd = 0;
		
		/*
		 *	Standard-Rufzeichen setzten
		 *
		 */
		memcpy_xc(&conf.hot_data.mycall, 
		          &l1_nocall, 
			  sizeof(l1_nocall));

		/*
		 *	reservierten Bereich mit 0 initialisieren
		 *
		 */
		memset_x(conf.hot_data.reserved, 0,
			 sizeof(conf.hot_data.reserved));
	}

	if (seg & SEG_DATA) {
		/*
		 *	Allgemeine Konfigurationsdaten
		 *	zuruecksetzten
		 *
		 */
		*conf.ctext      = 0;
		conf.ident_addr  = 0;
		*conf.ident_text = 0;

		/*
		 *	AX.25-Parameter setzen
		 *
		 */
		conf.l2_unack	= L2_DEFAULT_UNACK;
		conf.l2_n2	= L2_DEFAULT_N2;
		conf.l2_t1	= L2_DEFAULT_T1;
		conf.l2_t2	= L2_DEFAULT_T2;
		conf.l2_t3	= L2_DEFAULT_T3;

		/*
		 *	SCC-Parameter setzen
		 *
		 */
		conf.scc_txinvert = SCC_DEFAULT_TXINVERT;
		conf.scc_rxinvert = SCC_DEFAULT_RXINVERT;
		conf.scc_txhead   = SCC_DEFAULT_TXHEAD;
		conf.scc_txtail   = SCC_DEFAULT_TXTAIL;
		conf.scc_preamble = SCC_DEFAULT_PREAMBLE;

		/*
		 *	Master-Tabelle loeschen
		 *
		 */
		m_clear();

		/*
		 *	Repeater-Konfiguration setzen
		 *
		 */
		rp_trigger_clear();
	}
}


/*
 *	Konfigurationsdaten laden und ueberpruefen.
 *
 */
void conf_init ()
{
	ee_read(CONF_ADDR, &conf, sizeof(conf));
	
	/*
	 *	Wenn beim Initialisieren des Config-Moduls die TEST-Taste
	 *	gedrueckt ist, wird die Default-Konfiguration gesetzt.
	 *
	 */
	if (TEST_KEY)
		conf_load_defaults(SEG_DATA | SEG_HOT_DATA);
	else {
		unsigned int crc_data;
		unsigned int crc_hot_data;
		unsigned char seg;
	
		/*
		 *	Pruefsummen ueber die beiden Konfigurationssegmente
		 *	berechnen und ggf. default-Werte laden.
		 *	
		 */
		crc_hot_data = crc_crc16_buf(CRC_CRC16_INITIAL,
		                    (unsigned char xdata *)&conf.hot_data,
				    sizeof(conf.hot_data));
	
		crc_data = crc_crc16_buf(CRC_CRC16_INITIAL,
		                    (unsigned char xdata *)&conf,
				    sizeof(conf));

		seg = 0;
			
		if (conf.magic != CONF_MAGIC)
			seg |= SEG_DATA | SEG_HOT_DATA;
		
		if (crc_data != 0)
			seg |= SEG_DATA;

		if (crc_hot_data != 0)
			seg |= SEG_HOT_DATA;
		
		conf_load_defaults(seg);
	}
	
	/*
	 *	post-processing
	 *
	 */
	conf.ctext[sizeof(conf.ctext) - 1] = 0;
	conf.ident_text[sizeof(conf.ident_text) - 1] = 0;
	
 	conf.hot_data.passwd[sizeof(conf.hot_data.passwd) - 1] = 0;
	l1_set_mycall(&conf.hot_data.mycall);
}


/*
 *	Konfigurationsdaten in das EEPROM
 *	schreiben.
 *
 *	Die CRCs werden im little-Endian-Format abgelegt,
 *	damit die Pruefsummenkontrolle einfacher gestaltet
 *	werden kann.
 *
 */
void conf_save ()
{
	unsigned int crc;

	/*
	 *	Pruefsummen berechnen und Konfiguration speichern
	 *
	 */
	conf.magic = CONF_MAGIC;
	
	crc = crc_crc16_buf(CRC_CRC16_INITIAL,
		            (unsigned char xdata *)&conf.hot_data, 
		            sizeof(conf.hot_data) - sizeof(unsigned int));

	conf.hot_data.crc16 = s_le(crc);
	
	crc = crc_crc16_buf(CRC_CRC16_INITIAL,
		            (unsigned char xdata *)&conf, 
			    sizeof(conf) - sizeof(unsigned int));
	
	conf.crc16 = s_le(crc);

	ee_write(CONF_ADDR, &conf, sizeof(conf));
}

 
