
/*      
 *	$Id: eeprom.c,v 1.4 2002/04/14 14:27:32 Klaus Exp $
 *
 *	eeprom.c	EEPROM-Zugriffsfunktionen. Die Lese- und
 *			Schreib-Funktionen arbeiten mit einer
 *			Seitengroesse von 16 Bytes, damit moeglichst alle
 *			EEPROM-Typen angesprochen werden koennen.
 *
 *
 *
 *	(c) 2001,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "system.h"
#include "eeprom.h"

/*
 *      Zustaende der State-Machine
 *
 */
#define STATE_IDLE		0
#define STATE_READ		1
#define STATE_READ_PAGE		2
#define STATE_WRITE		3
#define STATE_WRITE_END		4

/*
 *      verwendete Zustaende des
 *      I2C-Senders/Empfaengers (s. 80C552 Datenbuch S.760ff)
 *
 */
#define I2C_START_SENT		0x08
#define I2C_REP_START_SENT	0x10
#define I2C_SLAW_SENT_ACK	0x18
#define I2C_SLAW_SENT_NACK	0x20
#define I2C_DATA_SENT_ACK	0x28
#define I2C_DATA_SENT_NACK	0x30
#define I2C_ARBITRATION_LOST	0x38
#define I2C_SLAR_SENT_ACK	0x40
#define I2C_SLAR_SENT_NACK	0x48
#define I2C_DATA_RCVD_ACK	0x50
#define I2C_DATA_RCVD_NACK	0x58

/*
 *      Puffer fuer lese- und schreib-Operationen
 *
 */
static unsigned char xdata *buf;
static unsigned int buflen;

/*
 *	EEPROM-Adresszaehler
 *
 */
static unsigned int addr;

/*
 *      Zustand des EEPROM-Moduls.
 *      Variable wird fuer die State-Machine verwendet.
 */
static unsigned char volatile state;

/*--------------------------------------------------------------------------*/


/*
 *      Initialisiert das EEPROM-Modul.
 *
 */
void ee_init () 
{
	ES1 = 0;	/* Interrupts von SIO1 abschalten */
	SI  = 0;	/* ggf. anstehenden Interrupt loeschen */
	SDA = 1;	/* P1.6 = P1.7 = 1, damit I2C-Bus funktioniert */
	SCL = 1;
	
	state = STATE_IDLE;
	
	/*
	 *    I2C-Bus/SIO1 konfigurieren
	 *
	 */
	EE_WP = 0;  	/* Schreibzugriff auf EEPROM erlauben */
	S1CON = 0xc1;	/* Bit-Frequenz = 100kHz, I2C enable */
	PS1   = 0;	/* SIO1 Interrupt mit niedriger Prioritaet */
}



/*
 *	liest ab der Adresse <_addr> das EEPROM aus. Dabei werden
 *	<_buflen> byte im Puffer <buf> abgelegt.
 *
 *	ACHTUNG: 	<_buflen> muss > 0 sein, sonst funktioniert
 *			der read-Algorithmus nicht richtig.
 *
 */
void ee_read (unsigned int _addr, void xdata *_buf, unsigned int _buflen)
{
	unsigned int mask;

	mask = sys_cli();

	ES1 = 1;	/* Intterupts zulassen */
	SI  = 0;	/* Anstehenden Interrupt loeschen */
	EA  = 1;	/* Interrupting einschalten */

	/*
	 *	Warten bis das EEPROM fuer den Zugriff
	 *	frei ist. Evtl. loest der Watchdog aus!
	 *
	 */
	WATCHDOG();
	while (state != STATE_IDLE);

	addr   = _addr;
	buf    = _buf;
	buflen = _buflen;

	state = STATE_READ;
	STA   = 1;

	/*
	 *	Warten bis der Lesevorgang beendet ist.
	 *
	 */
	WATCHDOG();
	while (state != STATE_IDLE);

	sys_sti(mask);
}


/*
 *	Schreibt alle Bytes aus dem Puffer <_buf>/<_buflen> in das
 *	EEPROM an der Adresse <_addr>
 *
 */
void ee_write (unsigned int _addr, void xdata *_buf, unsigned int _buflen)
{
	unsigned int mask;

	mask = sys_cli();

	ES1 = 1;	/* Intterupts zulassen */
	SI  = 0;	/* Anstehenden Interrupt loeschen */
	EA  = 1;	/* Interrupting einschalten */

	/*
	 *	Warten bis das EEPROM fuer den Zugriff
	 *	frei ist. Evtl. loest der Watchdog aus!
	 *
	 */
	WATCHDOG();
	while (state != STATE_IDLE);

	addr   = _addr;
	buf    = _buf;
	buflen = _buflen;
	
	state = STATE_WRITE;
	STA   = 1;

	/*
	 *	Warten bis alle Bytes aus dem gemappten
	 *	Bereich geschrieben wurden. Wenn die
	 *	Schreiboperation zu lange dauert, wird
	 *	der Watchdog ausgeloest!
	 *
	 */
	WATCHDOG();
	while (state != STATE_IDLE);

	sys_sti(mask);
}


/*--------------------------------------------------------------------------*/


/*
 *      Funktion setzt den Adresszaehler des EEPROMs
 *
 *      partielle State-Machine fuer den Master Receiver Mode.
 */
static void sio1_isr_read () using REGB_LP
{
	switch (S1STA) {
	case I2C_START_SENT:
	case I2C_REP_START_SENT:
		STA = 0;
		S1DAT = (EE_ADDR << 1) | ((addr >> 7) & 0x0e);
		break;

	case I2C_SLAW_SENT_ACK:
		S1DAT = addr & 0xff;	 /* A7 .. A0 */
		break;

	case I2C_DATA_SENT_ACK:
		state = STATE_READ_PAGE; /* Nun kann eine komplette Seite */
		STA = 1;                 /* gelesen werden */
		break;

	default:
		STA = 1;	/* Einfach nochmal versuchen */
	}
}



/*
 *      Funktion greift lesend auf das EEPROM zu.
 *      Diese Funktion wird von sio1_isr() aufgerufen.
 *
 *      partielle State-Machine fuer den Master Receiver Mode.
 */
static void sio1_isr_read_page () using REGB_LP
{
	switch (S1STA) {
	case I2C_START_SENT:
	case I2C_REP_START_SENT:
		STA = 0;
		S1DAT = (EE_ADDR << 1) | ((addr >> 7) & 0x0e) | 0x01;
		break;
	
	case I2C_DATA_RCVD_ACK:
	case I2C_DATA_RCVD_NACK:
		*buf = S1DAT;
		buf++;
		buflen--;
		addr++;

		if (!buflen) {
			state = STATE_IDLE;
			STO = 1;
			break;
		}

		if ( !(addr & 0x0f) ) {
			/*
			 *    Es muss eine neue Seite angebrochen werden,
			 *    Dazu muss zunaechst der Adresszeiger vom EEPROM
			 *    gesetzt werden.
			 *
			 */
			state = STATE_READ;
			STA = 1;
			break;
		} 

		/* kein break! */

	case I2C_SLAR_SENT_ACK:
		if ( buflen == 1 || (addr & 0x0f) == 0x0f )
			AA = 0;
		else
			AA = 1;
		break;
	
	default:
		state = STATE_READ;	/* Nochmal probieren */
		STA = 1;
	}
}


/*
 *      Funktion greift schreibend auf das EEPROM zu.
 *      Diese Funktion wird von sio1_isr() aufgerufen.
 *
 *      State-Machine fuer den Master Transmitter Mode.
 *      Modifiziertes Datenbyte in das EEPROM schreiben.
 */
static void sio1_isr_write () using REGB_LP
{
	switch (S1STA) {
	case I2C_START_SENT:
	case I2C_REP_START_SENT:
		STA = 0;
		S1DAT = (EE_ADDR << 1) | ((addr >> 7) & 0x0e);
		break;

	case I2C_SLAW_SENT_ACK:
		S1DAT = addr & 0xff;	/* A7 .. A0 */
		break;

	case I2C_DATA_SENT_ACK:
		if (!buflen) {
     			state = STATE_IDLE;
     			STO = 1;
      			break;
   		}

		S1DAT = *buf;
		buf++;
		buflen--;
		addr++;

		if ( !(addr & 0x0f) )
			state = STATE_WRITE_END;

		break;
		
	default:
		STA = 1;
	}
}


/*
 *	Nachdem eine komplette Speicherseite geschrieben
 *	wurde, wird diese Funktion aufgerufen. Hier wird
 *	das Schreiben der naechsten Speicherseite
 *	angestossen.
 *
 */
static void sio1_isr_write_end () using REGB_LP
{
	switch (S1STA) {
	case I2C_DATA_SENT_ACK:
		STO = 1;
	
		if (!buflen) {
			state = STATE_IDLE;
			break;
		}

		/*
		 *	Wenn noch Daten zu schreiben sind,
		 *	wird nach dem STOP sofort wieder
		 *	ein START gesendet...
		 *
		 */

	default:
		state = STATE_WRITE;
		STA = 1;
	}
}


/*
 *      Interrupt subroutine der SIO1
 *
 *      Diese Funktion wird bei jedem Zustandswechsel des
 *      I2C-Bus Senders/Empfaengers aufgerufen.
 */
void sio1_isr () interrupt INT_S1 using REGB_LP 
{
	switch (state) {
	case STATE_READ:
		sio1_isr_read();
		break;

	case STATE_READ_PAGE:
		sio1_isr_read_page();
		break;

	case STATE_WRITE:
		sio1_isr_write();
		break;

	case STATE_WRITE_END:
		sio1_isr_write_end();
		break;

	default:
		state = STATE_IDLE;
	}
	
	SI = 0;
}



