
/*
 *	$Id: flash.c,v 1.1.1.1 2002/02/15 22:36:40 Klaus Exp $
 *
 *	flash.c		Funktionen zum Loeschen und Programmieren
 *			der Flahs AMD Am29F010 und ATMEL AT29C010A. 
 *			Die Flash-Funktionieren funktionieren nur dann, 
 *			wenn das EPROM in das code-Segment gemapped ist.
 *
 *
 *
 *	(c) 2001,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "flash.h"
#include "system.h"

/*
 *	Makro zum direkten Zugriff auf das xdata Segment
 *
 */
#define xmem(addr)	*((unsigned char volatile xdata *)(addr))

/*
 *	Flash-Typ, Hersteller und Flash-ID
 *
 */
unsigned char flash_type;
unsigned char flash_manufacturer;
unsigned char flash_device_code;

/*--------------------------------------------------------------------------*/
/*		GENERISCHE FLASH-ZUGRIFFSROUTINEN			    */
/*--------------------------------------------------------------------------*/


/*
 *	Wartet bis das Flash mit der letzten schreib/loesch-Operation
 *	fertig ist.
 *
 *	ACHTUNG:	Diese Funktion darf _nicht_ auf das xdata-Segment
 *			zugreifen!
 *
 */
static void flash_wait ()
{
	unsigned char idata old;
	unsigned char idata new;

	old = xmem(0x0000);

	while (old ^ (new = xmem(0x0000)))
		old = new;
}


/*
 *	Wartet ca. 10s darauf, dass das Flash die letzte Operation beendet.
 *	Beim Loeschen kann das Flash bis zu 8 Sekunden Zeit brauchen...
 *
 *	ACHTUNG:	Diese Funktion darf _nicht_ auf das xdata-Segment
 *			zugreifen.
 *	
 */
static void flash_long_wait ()
{
	unsigned char idata old;
	unsigned char idata new;
	unsigned long idata cntr;

	cntr = 200000; 

	old = xmem(0x0000);

	while (old ^ (new = xmem(0x0000))) {
		old = new;

		if (!--cntr)
			break;

		WATCHDOG();
	}
}


/*
 *	Wartet ca. <ms> Millisekunden. Die genaue Wartezeit
 *	haengt von dem erzeugten Code ab.
 *
 */
static void flash_sleep (unsigned char ms)
{
	register unsigned char n;

	for (; ms; ms--)
		for (n = 0; n < 255; n++);
}


/*
 *	Identifiziert das Flash, das auf dem RPC steckt und
 *	fuehrt danach einen Reset des Flashs durch.
 *
 */
void flash_init ()
{
	unsigned char idata man;	/* manifacturer */
	unsigned char idata dc;		/* device code */

	/*
	 *	Die Firmware muss im EPROM laufen, damit
	 *	die Flash-Daten ermittelt werden koennen.
	 *
	 */
	if (IS_CSEG_FLASH()) {
		flash_type = FLASH_UNTESTED;
		flash_manufacturer = 0;
		flash_device_code  = 0;
		return;
	}
	
	FLASH_A16(0);
	
	/*
	 *	Flash Identifizieren
	 *
	 */
	XSEG_FLASH();

	xmem(0x5555) = 0xaa;
	xmem(0x2aaa) = 0x55;
	xmem(0x5555) = 0x90;

	flash_sleep(10);

	man = xmem(0x0000);
	dc  = xmem(0x0001);

	XSEG_XDATA();

	/*
	 *	Ausgelesene Daten auswerten
	 *
	 */
	flash_manufacturer = man;
	flash_device_code  = dc;

	if (man == 0x1f && dc == 0xd5)
		flash_type = FLASH_AT29C010A;
	else
		flash_type = FLASH_AM29F010;

	/*
	 *	Reset des Flashs durchfuehren
	 *
	 */
	XSEG_FLASH();

	xmem(0x5555) = 0xaa;
	xmem(0x2aaa) = 0x55;
	xmem(0x5555) = 0xf0;

	XSEG_XDATA();

	flash_sleep(10);
}


/*
 *	Liest aus der Flash-Bank <bank> ab Adresse <addr>
 *	<buflen> Bytes und speichert sie in <buf>
 *
 */
void flash_read (unsigned char bank,
		 unsigned int addr,
		 void xdata *buf,
		 unsigned int buflen)
{
	unsigned char idata b;
	unsigned int idata raddr = addr;

	FLASH_A16(bank);

	while (buflen) {
		WATCHDOG();
	
		EA = 0;
		XSEG_FLASH();

		b = xmem(raddr);

		XSEG_XDATA();
		EA = 1;
		
		*((unsigned char xdata *)buf) = b;
		buf = ((unsigned char xdata *)buf) + 1;
		buflen--;

		raddr++;
	}
}


/*--------------------------------------------------------------------------*/
/*		FLASH-FUNKTIONEN FUER DAS AMD AM29F010			    */
/*--------------------------------------------------------------------------*/


/*
 *	Loescht die angegebene Speicherbank
 *
 */
static void amd_erase (unsigned char bank)
{
	unsigned int idata sec_addr = 0x0000;
	bit a16 = (bank);

	EA = 0;
	XSEG_FLASH();	

	do {
		WATCHDOG();
	
		FLASH_A16(0);
		
		xmem(0x5555)   = 0xaa;
		xmem(0x2aaa)   = 0x55;
		xmem(0x5555)   = 0x80;
		xmem(0x5555)   = 0xaa;
		xmem(0x2aaa)   = 0x55;

		FLASH_A16(a16);
		xmem(sec_addr) = 0x30;

		flash_long_wait();
		
		sec_addr += 0x4000;
		
	} while (sec_addr);

	XSEG_XDATA();
	EA = 1;
}


/*
 *	Schreibt den Puffer <buf>/<buflen> in die Flash-Bank <bank>
 *	ab der Adresse <addr>
 *
 */
static void amd_write (unsigned char bank, 
		       unsigned int addr, 
                       unsigned char xdata *buf, 
		       unsigned int buflen)
{
	unsigned char idata b;
	unsigned int idata raddr = addr;
	bit a16 = (bank);
	
	while (buflen) {
		WATCHDOG();
	
		/*
		 *	Bevor ein Byte geschrieben werden kann, muessen
		 *	alle notwendigen Daten in Registern oder im 
		 *	idata-Segment gespeichert werden, da beim
		 *	Programmieren das Flash in das xdata-Segment
		 *	eingeblendet ist!
		 *	
		 */
		b = *buf;
		buf++;
		buflen--;

		/*
		 *	Byte programmieren
		 *
		 */
		EA = 0;
		XSEG_FLASH();
		 
		FLASH_A16(0);

		xmem(0x5555) = 0xaa;
		xmem(0x2aaa) = 0x55;
		xmem(0x5555) = 0xa0;

		FLASH_A16(a16);
		xmem(raddr) = b;
		
		flash_wait();

		/*
		 *	RAM wieder in das xdata-Segment mappen
		 *
		 */
		XSEG_XDATA();
		EA = 1;

		raddr++;
	}
}


/*--------------------------------------------------------------------------*/
/*		FLASH-FUNKTIONEN FUER DAS ATMEL AT29C010A		    */
/*--------------------------------------------------------------------------*/


/*
 *	ATMEL-Flashes werden Sektor-weise geschrieben. Es werden
 *	128 Bytes geschrieben und dann eine kurze Zeit gewartet (in der
 *	Wartezeit schreibt das Flash die zwischengespeicherten Daten).
 *	Damit keine Dateninkonsistenz entsteht wird zuerst der entsprechende
 *	Sektor mit den alten Daten aus dem Flash gefuellt (siehe Datenblatt).
 *
 */
void atmel_write (unsigned char bank, 
		  unsigned int addr, 
                  unsigned char xdata *buf, 
		  unsigned int buflen)
{
	unsigned char idata b;
	unsigned int idata raddr = addr;
	unsigned int idata fa;

	FLASH_A16(bank);

	while (buflen) {
		WATCHDOG();
		
		fa = raddr & 0xff80;

		EA = 0;
		XSEG_FLASH();

		while (fa != raddr) {
			xmem(fa) = xmem(fa);
			fa++;
		}
	
		XSEG_XDATA();

		do {
			b = *buf;
			buf++;
			buflen--;

			XSEG_FLASH();
			xmem(raddr) = b;
			XSEG_XDATA();

			raddr++;
		
		} while (buflen && (raddr & 0x007f));

		XSEG_FLASH();

		while (raddr & 0x007f) {
			xmem(raddr) = xmem(raddr);
			raddr++;
		}

		XSEG_XDATA();
		EA = 1;
		
		flash_sleep(5);

		EA = 0;
		XSEG_FLASH();
		flash_wait();
		XSEG_XDATA();
		EA = 1;
	}

	WATCHDOG();
}


/*--------------------------------------------------------------------------*/


/*
 *	Loescht die angegebene Bank das Flashs
 *
 */
void flash_erase (unsigned char bank)
{
	if (flash_type == FLASH_AM29F010)
		amd_erase(bank);
}


/*
 *	Schreibt den Puffer <buf>/<buflen> ab der Adresse <addr>
 *	in die Flash-Bank <bank>.
 *
 */
void flash_write (unsigned char bank, 
		  unsigned int addr, 
                  void xdata *buf, 
		  unsigned int buflen)
{
	switch (flash_type) {
	case FLASH_AM29F010:
		amd_write(bank, addr, buf, buflen);
		break;
	
	case FLASH_AT29C010A:
		atmel_write(bank, addr, buf, buflen);
	}
}



