
/*
 *	$Id: kiss.c,v 1.1.1.1 2002/02/15 22:36:40 Klaus Exp $
 *
 *	kiss.c		Funktionen zum Codieren und Decodieren
 *			eines KISS-Datenstroms.
 *
 *
 *
 *
 *	(c) 1999,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "system.h"
#include "kiss.h"
#include "buf.h"
#include "queue.h"
#include "serial.h"
#include "ax25_l1.h"
#include "mm.h"
#include "crc.h"

/*
 *	Zustaende der KISS-Decoder Statemachine
 *
 */
#define STATE_RX_WAIT_START	0
#define STATE_RX_WAIT_CTL	1
#define STATE_RX_DATA		2
#define STATE_RX_ESCAPE		3

/*
 *	Zustaende der KISS-Encoder Statemachine
 *
 */
#define STATE_TX_OFF		0
#define STATE_TX_FEND		1
#define STATE_TX_CTL		2
#define STATE_TX_DATA		3
#define STATE_TX_TFEND		4
#define STATE_TX_TFESC		5

/*
 *	Definitionen fuer <what>-Parameter bei
 *	kiss_activity.
 *
 */
#define KISS_RX			0
#define KISS_TX			1

/*
 *	In kiss_ctl wird das control-Byte der entsprechenden
 *	KISS-Variante gespeichert. 
 *
 */
static unsigned char idata kiss_ctl = KISS_CTL_DATA;

/*
 *	Sende- und Empfangspuffer. Die Sendepuffer werden
 *	in einer Queue gehalten.
 *
 */
static struct Buffer xdata * idata rxbuf;
static struct Buffer xdata * idata txbuf;
static struct Queue xdata txq;
 
/*
 *	Zustand der Empfangs- und Sende-Statemachine
 *
 */
static unsigned char idata rxstate;
static unsigned char idata txstate;

/*---------------------------------------------------------------------------*/



/*
 *	Initialisiert das KISS-Layer
 *
 */
void kiss_init ()
{
	q_init(&txq,KISS_MAXTXFMS);
	txstate = STATE_TX_OFF;

	rxbuf = NULL;
	rxstate = STATE_RX_WAIT_START;

	ser_putch(KISS_FEND);
}


/*
 *	Steuert die Activity-LED vom RPC
 *
 */
static void kiss_activity (unsigned char what, unsigned char state)
{
	static bit act_rx = 0;
	static bit act_tx = 0;

	if (what == KISS_RX)
		act_rx = state;
	else
		act_tx = state;

	if (act_rx || act_tx)
		ON(LED_ACT);
	else
		OFF(LED_ACT);
}


/*
 *	Prueft ob das empfangene Frame Ok ist.
 *	Rueckgabe: 0 Frame fehlerhaft
 *		   1 Frame Ok
 *
 */
static char kiss_rx_check ()
{
	unsigned int crc;
	
	switch (kiss_ctl) {
	case KISS_CTL_DATA:
		return 1;

	case KISS_CTL_SMACK:
		crc = crc_crc16(CRC_CRC16_INITIAL, KISS_CTL_SMACK);
		crc = crc_crc16_buf(crc, buf_data(rxbuf), buf_size(rxbuf));
		
		if (!crc) {
			buf_getch_tail(rxbuf);
			buf_getch_tail(rxbuf);
			return 1;
		}

		break;	
	
	case KISS_CTL_FLEX:
		crc = crc_flex(CRC_FLEX_INITIAL, KISS_CTL_FLEX);
		crc = crc_flex_buf(crc, buf_data(rxbuf), buf_size(rxbuf));

		if (crc == 0x7070) {
			buf_getch_tail(rxbuf);
			buf_getch_tail(rxbuf);
			return 1;
		}
	
		break;
	}
	
	return 0;
}


/*
 *	Holt alle empfangenen Zeichen von der seriellen Schnittstelle
 *	und nimmt eine KISS-Decodierung vor. Sobald ein komplettes
 *	Frame ausdecodiert wurde, wird es an das AX.25 Layer 1
 *	weitergegeben.
 *
 */
static void kiss_rx ()
{
	int idata sb;

	while ( (sb = ser_getch()) != -1 ) {
		unsigned char idata b = sb;
		
		switch (rxstate) {
		case STATE_RX_WAIT_START:
			if (b == KISS_FEND)
				rxstate = STATE_RX_WAIT_CTL;

			break;
		
		case STATE_RX_WAIT_CTL:
			if (b == KISS_FEND)
				break;

			if (b != KISS_CTL_DATA  &&
			    b != KISS_CTL_SMACK &&
			    b != KISS_CTL_FLEX)
			{
				rxstate = STATE_RX_WAIT_START;
				break;
			}
			
			kiss_ctl = b;

			if ( (rxbuf = buf_new(KISS_MAXBUFLEN)) ) {
				rxstate = STATE_RX_DATA;
				kiss_activity(KISS_RX, 1);
		        } else		
				rxstate = STATE_RX_WAIT_START;
				
			break;

		case STATE_RX_ESCAPE:
			if (b == KISS_TFEND) 
				b = KISS_FEND;
			else if (b == KISS_TFESC)
				b = KISS_FESC;
			else {
				buf_delete(rxbuf);
				rxbuf = NULL;
				rxstate = STATE_RX_WAIT_START;
				kiss_activity(KISS_RX, 0);
				break;
			}

			rxstate = STATE_RX_DATA;
			goto store_byte;

		case STATE_RX_DATA:
			if (b == KISS_FEND) {
				/*
				 *	Hier wird das ausdecodierte Frame
				 *	an das hoehere Layer weitergegeben.
				 *
				 */
				if (kiss_rx_check()) {
					WATCHDOG();
					l1_recv(rxbuf); 
				} else
					buf_delete(rxbuf);

				rxbuf = NULL;

				/*
				 *	Es kann sein, dass zwei Frames nur
				 *	von einem FEND begrenzt werden, daher
				 *	muss als naechstes auf ein Control-
				 *	Byte gewartet werden!
				 *
				 */
				rxstate = STATE_RX_WAIT_CTL;
				kiss_activity(KISS_RX, 0);
				break;	
			}
		
			if (b == KISS_FESC) {
				rxstate = STATE_RX_ESCAPE;
				break;
			}
store_byte:
			if (!buf_putch(rxbuf,b)) {
				rxstate = STATE_RX_WAIT_START;
				buf_delete(rxbuf);
				rxbuf = NULL;
				kiss_activity(KISS_RX, 0);
			}

			break;

		default:
			if (rxbuf) {
				buf_delete(rxbuf);
				rxbuf = NULL;
			}

			rxstate = STATE_RX_WAIT_START;
			kiss_activity(KISS_RX, 0);
		}
	}
}



/*
 *	Wenn Frames in der Sendequeue stehen, werden sie
 *	codiert und ueber die serielle Schnittstelle ausgegeben.
 *
 */
static void kiss_tx ()
{
	while (!ser_istxbusy()) {
		unsigned char idata ch;
		int idata chi;
	
		switch (txstate) {
		case STATE_TX_OFF:
			if ( !(txbuf = q_dequeue(&txq)) )
				return;

			kiss_activity(KISS_TX, 1);
			txstate = STATE_TX_FEND;
			break;
		
		case STATE_TX_FEND:
			ser_putch(KISS_FEND);
			txstate = STATE_TX_CTL;
			break;

		case STATE_TX_CTL:
			ser_putch(kiss_ctl);
			txstate = STATE_TX_DATA;
			break;

		case STATE_TX_DATA:
			if ( (chi = buf_getch(txbuf)) == -1 ) {
				buf_delete(txbuf);
				ser_putch(KISS_FEND);
				txstate = STATE_TX_OFF;
				kiss_activity(KISS_TX, 0);
				break;
			}
		
			ch = (unsigned char)chi;

			if (ch == KISS_FEND) {
				ser_putch(KISS_FESC);
				txstate = STATE_TX_TFEND;
				break;
			}

			if (ch == KISS_FESC) {
				ser_putch(KISS_FESC);
				txstate = STATE_TX_TFESC;
				break;
			}

			ser_putch(ch);
			break;

		case STATE_TX_TFEND:
			ser_putch(KISS_TFEND);
			txstate = STATE_TX_DATA;
			break;

		case STATE_TX_TFESC:
			ser_putch(KISS_TFESC);
			txstate = STATE_TX_DATA;
			break;

		default:
			txstate = STATE_TX_OFF;
			kiss_activity(KISS_TX, 0);
		}
	}
}



/*
 *	Fuehrt die KISS Empfangs- und Sendefunktionen aus.
 *
 */
void kiss_run ()
{
	kiss_rx();
	kiss_tx();
}



/*
 *	Reiht das Frame *buf in die Sendequeue ein. Wenn die
 *	die maximale Frame-Anzahl ueberschritten wurde, wird
 *	0 zurueckgegeben und das Frame nicht eingereiht!
 *	Wenn alles klappt, wird 1 zurueckgegeben.
 *
 */
char kiss_xmit (struct Buffer xdata *buf)
{
	unsigned int crc;

	switch (kiss_ctl) {
	case KISS_CTL_DATA:
		break;
	
	case KISS_CTL_SMACK:
		crc = crc_crc16(CRC_CRC16_INITIAL, KISS_CTL_SMACK);
		crc = crc_crc16_buf(crc, buf_data(buf), buf_size(buf));
		
		if (!buf_putch(buf,(unsigned char)(crc & 0xff)) ||
		    !buf_putch(buf,(unsigned char)(crc >> 8)))
			return 0;

		break;
	
	case KISS_CTL_FLEX:
		crc = crc_flex(CRC_FLEX_INITIAL, KISS_CTL_FLEX);
		crc = crc_flex_buf(crc, buf_data(buf), buf_size(buf));
	
		if (!buf_putch(buf,(unsigned char)(crc >> 8)) ||
		    !buf_putch(buf,(unsigned char)(crc & 0xff)))
			return 0;

		break;
	
	default:
		kiss_ctl = KISS_CTL_DATA;
	}

	return q_enqueue(&txq,buf);
}






