
/*
 *	$Id: master.c,v 1.2 2002/04/07 14:36:08 klaus Exp $
 *
 *	master.c	Simple Verwaltungsfunktionen rund um die
 *			Master-Tabelle
 *
 *
 *
 *	(c) 2001,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "master.h"
#include "ax25.h"
#include "ax25_l1.h"
#include "ax25_io.h"
#include "config.h"
#include "mm.h"
#include "string.h"
#include "scp.h"
#include "rtc.h"
#include "system.h"

/*
 *	Master-Tabelle aus dem Konfigurationsmodul beziehen.
 *	
 */
#define mtab	conf.mtab

/*
 *	Zeit in Zehntelsekunden, die zwischen zwei Connect-Versuchen
 *	verstreichen muss.
 *
 */
#define TRY_DELAY	300

/*--------------------------------------------------------------------------*/


/*
 *	Loescht alle Eintraege der Master-Tabelle
 *
 */
void m_clear ()
{
	struct Master xdata *m;

	for (m = mtab; m != mtab + MASTER_MAX; m++)
		m->in_use = 0;
}


/*
 *	Sucht den Master-Eintrag, der zur Adresse <addr> gehoert
 *	aus der Mastertabelle heraus. Wenn kein Eintrag gefunden
 *	wurde wird NULL zurueckgegeben.
 *
 */
struct Master xdata * m_lookup (struct AX25_address xdata *addr)
{
	struct Master xdata *m;

	for (m = mtab; m != mtab + MASTER_MAX; m++) {
		if (!m->in_use)
			continue;

		if (l1_cmpaddr(addr, &m->path.peer) == 1)
			return m;
	}

	return NULL;
}


/*
 *	Fuegt den Master mit dem Pfad <path> der Master-Tabelle
 *	hinzu. Wenn die Master-Tabelle bereits voll ist oder bereits
 *	ein Eintrag fuer den Master existiert, wird 0 zurueckgegeben, 
 *	ansonsten 1.
 *
 */
char m_add (struct AX25_smart_path xdata *path)
{
	struct Master xdata *m;

	if (m_lookup(&path->peer))
		return 0;

	for (m = mtab; m != mtab + MASTER_MAX; m++)
		if (!m->in_use) {
			memcpy_xx(&m->path, path, sizeof(*path));
			m->in_use  = 1;
			m->cmode   = 0;
			*m->passwd = 0;
			return 1;
		}
	
	return 0;
}


/*
 *	Entfernt den Master, der zur Adresse <addr> gehoert aus der
 *	Master-Tabelle. Es wird die Anzahl der entfernten Master
 *	zurueckgegeben.
 *
 */
char m_del (struct AX25_address xdata *addr)
{
	struct Master xdata *m;

	if ( !(m = m_lookup(addr)) )
		return 0;

	m->in_use = 0;
	return 1;
}


/*
 *	Gibt den Inhalt der aktuellen Master-Tabelle ueber den
 *	AX.25-Controlblock <axcb> aus.
 *
 */
void m_list (struct AX25_cb xdata *axcb)
{
	struct Master xdata *m;

	ax25_printf(axcb,"master                             cmode  pwd\r");
	/*                DL3KHB-15 via DB0AAA-10 DB0BBB-10     on   on */

	for (m = mtab; m != mtab + MASTER_MAX; m++) {
		if (!m->in_use)
			continue;

		ax25_printf(axcb, "%-35p%5b%5b\r", 
		            (__XPTR__)&m->path,
			    (unsigned char)m->cmode,
			    (unsigned char)*m->passwd);
	}
}


/*--------------------------------------------------------------------------*/


/*
 *	Wenn der RPC keine Verbindung zu einem Master hat, wird
 *	hier versucht einen Master zu connecten und mit dem SCP
 *	zu verbinden.
 *
 */
void m_run ()
{
	static struct Master xdata *master = mtab;
	static struct AX25_cb xdata *axcb = NULL;
	static stime_t last_try = 0x8000;

	/*
	 *	Haben wir bereits einen AX.25-Control-Block, zum
	 *	connecten des Masters erzeugt? Solange versucht
	 *	wird den Master zu connecten, wird die CON-LED
	 *	blinken gelassen.
	 *
	 */
	if (axcb) {
		static stime_t last_flip;
		stime_t cur_time;

		cur_time = rtc_time();

		if (cur_time - last_flip > 5) {
			FLIP(LED_CON);
			last_flip = cur_time;
		}
	
		switch (ax25_event(axcb)) {
		case EVENT_NOEVENT:
		case EVENT_TXQEMPTY:
			break;

		case EVENT_CONNECT:
			scp_setup(axcb);
			axcb = NULL;
			break;
		
		default:
			ax25_close(axcb);
			axcb = NULL;

			if (scp_connection())
				ON(LED_CON);
			else
				OFF(LED_CON);
		}
	
		return;
	}
	
	/*
	 *	Wenn bereits eine aktive SCP-Verbindung besteht,
	 *	muessen wir nichts weiter machen.
	 *
	 */
	if (scp_connection())
		return;

	if (master == mtab && rtc_time() - last_try < TRY_DELAY)
		return;

	/*
	 *	Wenn bereits alle Master durchprobiert wurden,
	 *	wird eine kurze Zeit gewartet und dann erneut
	 *	mit dem ersten begonnen.
	 *
	 */
	if (master >= mtab + MASTER_MAX) {
		master = mtab;
		last_try = rtc_time();
		return;
	}

	if (!master->in_use || !master->cmode) {
		master++;
		return;
	}
	
	if ( !(axcb = ax25_open(TYPE_SEQPACKET)) ) {
		master++;
		return;
	}

	if (!ax25_connect(axcb, &master->path)) {
		ax25_close(axcb);
		axcb = NULL;
	}
	
	master++;
}


