
/*
 *	$Id: page.c,v 1.9 2002/04/01 12:56:57 klaus Exp $
 *
 *	page.c		Funkruf- und PageQueue-Verwaltungsfunktionen
 *
 *
 *
 *
 *	(c) 2002,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "page.h"
#include "mm.h"

/*--------------------------------------------------------------------------*/


/*
 *	Erzeugt eine neue Page-Instanz, mit einem Puffer, der
 *	<cw_cnt> Codewoerter fassen kann. Der Funkruf wird immer
 *	in der Domaene PAGE_DOMAIN_PAGER angelegt.
 *
 *	Im Fehlerfall wird NULL zurueckgegeben, ansonsten der 
 *	Zeiger auf die neue Instanz.
 *
 */
struct Page xdata * page_new (unsigned char cw_cnt, unsigned char frame)
{
	struct Page xdata *p;

	p = (struct Page xdata *)
	    malloc(sizeof(struct Page) + 4 * (unsigned int)cw_cnt);

	if (p) {
		p->next   = NULL;
		p->frame  = frame & 0x07;
		p->domain = PAGE_DOMAIN_PAGER;
		p->cw_cnt = cw_cnt;
		p->cw_buf = (unsigned long xdata *)(p + 1);
	}

	return p;
}


/*
 *	Trennt von der Page-Sequenz <*ps> max. <cnt> Codewoerter
 *	ab und bildet daraus einen Funkruf. Es wird der Zeiger auf
 *	den erzeugten Funkruf zurueckgegeben oder im Fehlerfall NULL.
 *
 */
struct Page xdata * page_split (struct PageSequence xdata *ps,
			        unsigned char cnt)
{
	struct Page xdata *p;

	if (ps->cw_cnt == 0)
		return NULL;

	if ( !(p = page_new(0, (ps->cw_off >> 1) & 0x07)) )
		return NULL;

	p->cw_buf = ps->cw_buf;
	
	if ((unsigned int)cnt >= ps->cw_cnt)
		p->cw_cnt = ps->cw_cnt;
	else
		p->cw_cnt = cnt;
		
	ps->cw_buf += p->cw_cnt;
	ps->cw_off += p->cw_cnt;
	ps->cw_cnt -= p->cw_cnt;

	return p;
}


/*--------------------------------------------------------------------------*/
/*									    */
/*	PageQueue - Verwaltungsfunktionen				    */
/*									    */
/*--------------------------------------------------------------------------*/


/*
 * 	Initialisiert die Page-Queue <pq>
 *
 */
void pq_init (struct PageQueue xdata *pq)
{
	pq->first    = NULL;
	pq->last     = NULL;
	pq->page_cnt = 0;
}


/*
 *	Loescht alle Funkrufe aus der Queue <pq>
 *
 */
void pq_clear (struct PageQueue xdata *pq)
{
	struct Page xdata *p;

	while (pq->first != NULL) {
		p = pq->first;
		pq->first = pq->first->next;

		page_delete(p);
	}

	pq_init(pq);
}


/*
 *	Loescht alle Funkrufe der Domain <domain> aus
 *	der PageQueue <pq>.
 *
 */
void pq_clear_dom (struct PageQueue xdata *pq, unsigned char domain)
{
	struct Page xdata *p;
	struct Page xdata *prev;

	prev = NULL;
	p    = pq->first;

	while (p != NULL)
		if (p->domain != domain) {
			prev = p;
			p    = p->next;
		} else {
			struct Page xdata *next = p->next;
		
			if (prev == NULL)
				pq->first = next;
			else
				prev->next = next;
			
			page_delete(p);
			pq->page_cnt--;
			p = next;
		}

	pq->last = prev;
}


/*
 *	Fuegt den Funkruf <page> zu der Page-Queue <pq>
 *	hinzu. Wenn <head> != 0 ist, wird der Funkruf am Anfang der
 *	Queue hinzugefuegt, ansonsten am Ende.
 *
 */
void pq_enqueue (struct PageQueue xdata *pq, 
		 struct Page xdata *page, 
		 unsigned char head)
{
	page->next = NULL;

	if (!pq->first) 
	{
		pq->first = page;
		pq->last = page;
	} 
	else if (head) 
	{
		page->next = pq->first;
		pq->first = page;
	} 
	else 
	{
		pq->last->next = page;
		pq->last = page;
	}

	pq->page_cnt++;
}


/*
 *	Entfernt den ersten Funkruf aus der Queue <pq> und
 *	gibt einen Zeiger auf ihn zurueck. Wenn kein Funkruf
 *	in der Queue ist, wird NULL zurueckegeben.
 *
 */
struct Page xdata * pq_dequeue (struct PageQueue xdata *pq)
{
	struct Page xdata *p;

	if (!pq->first)
		return NULL;

	p = pq->first;

	if ( !(pq->first = p->next) )
		pq->last = NULL;

	pq->page_cnt--;
	
	return p;
}


/*
 *	Fuehrt das Batchen von Funkrufen aus.
 *
 *	Es werden soviele Funkrufe von <src> nach <dest> umgehaengt,
 *	dass bei der Aussendung der Funkrufe aus <dest> nicht mehr als
 *	<max_bcnt> Batches benoetigt werden.
 *
 */
void pq_batch (struct PageQueue xdata *dest,
	       struct PageQueue xdata *src, 
	       unsigned char max_bcnt)
{
	struct Page xdata *p;
	unsigned int cw_cnt;

	/*
	 *	Berechnen, wieviele Codewoerter benoetigt werden,
	 *	um die Funkrufe zu codieren, die schon in <dest>
	 *	stehen.
	 *
	 */
	cw_cnt = 0;	
	 
	for (p = dest->first; p != NULL; p = p->next) {
		/*
		 *	Zuerst alle notwendigen IDLE-Codewoerter addieren
		 *	und dann die Funkruf-Codewoerter hinzuzaehlen.
		 *
		 */
		cw_cnt += (2 * p->frame - (cw_cnt & 0x0f)) & 0x0f;
		cw_cnt += p->cw_cnt;
	}

	/*
	 *	Nun werden soviele Funkrufe von <src> nach <dest> umgehaengt,
	 *	dass zum Codieren aller Funkrufe in <dest> nicht mehr als
	 *	<max_bcnt> Batches benoetigt werden.
	 *
	 */
	while ( (p = pq_dequeue(src)) != NULL ) {
		/*
		 *	Zuerst alle notwendigen IDLE-Codewoerter addieren
		 *	und dann die Funkruf-Codewoerter hinzuzaehlen.
		 *
		 */
		cw_cnt += (2 * p->frame - (cw_cnt & 0x0f)) & 0x0f;
		cw_cnt += p->cw_cnt;

		if ( (cw_cnt + 15) / 16 > max_bcnt ) {
			pq_enqueue(src, p, 1);
			break;
		}

		/*
		 *	Funkruf passt noch, also umhaengen.
		 *
		 */
		pq_enqueue(dest, p, 0);
	}
}


/*
 *	Fuegt alle Funkrufe aus <src> am Anfang
 *	von <dest> ein.
 *
 */
void pq_insert (struct PageQueue xdata * dest,
		struct PageQueue xdata * src)
{
	if (!src->first)
		return;

	if (dest->first) {
		src->last->next = dest->first;
		dest->first     = src->first;
	} else {
		dest->first = src->first;
		dest->last  = src->last;
	}

	dest->page_cnt += src->page_cnt;
	pq_init(src);
}



