
/*
 *	$Id: pocsag.c,v 1.13 2002/05/13 22:21:49 klaus Exp $
 *
 *	pocsag.c	Modul, zum Codieren und Decodieren von Funkrufen 
 *			nach dem POCSAG-Protokoll. Die Behandlung von
 *			Synchron- und Idle-Codewoertern wird hier nicht
 *			vorgenommen.
 *
 *
 *
 *	(c) 2001,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "pocsag.h"
#include "page.h"
#include "string.h"
#include "mm.h"
#include "system.h"
#include "scc.h"

/*
 *	Makro berechnet aus einer Adresse und der Angabe der Funktions-Bits
 *	ein Adresscodewort.
 *
 */
#define ENCODE_ACW(addr,func)	\
			(((((unsigned long)(addr)) & 0x001ffff8UL) << 10) | \
                        ((((unsigned long)(func)) & 0x00000003UL) << 11))

/*
 *	Baut ein Nachrichten-Codewort zusammen
 *
 */
#define ENCODE_MCW(msg)	(0x80000000UL | \
      		        ((((unsigned long)(msg)) & 0x000fffffUL) << 11))
			
/*
 *	Extrahiert die Adresse aus dem Adress-Codewort und gibt
 *	sie zurueck. Die untersten 3 Bits der Adresse sind 0 und muessen
 *	ueber die zugehoerige Framenummer ermittelt werden.
 *
 */
#define DECODE_ACW(cw)	(((unsigned long)(cw) >> 10) & 0x001ffff8UL)

/*
 *	Extrahiert die Funktionsbits aus dem Adresscodewort <cw> und
 *	gibt sie zurueck.
 *
 */
#define DECODE_FUNC(cw)	(((unsigned long)(cw) >> 11) & 0x00000003UL)

/*
 *	Extrahiert aus dem Message-Codewort (cw) den Datenbereich
 *	und gibt ihn zurueck.
 *
 */
#define DECODE_MCW(cw)	(((unsigned long)(cw) >> 11) & 0x000fffffUL)

/*
 *	Wenn das Codewort <cw> ein Message-Codewort ist, wird
 *	!= 0 zurueckgegeben, ansonsten 0.
 *
 */
#define IS_MCW(cw)	(((unsigned long)(cw) >> 24) & 0x80)

/*--------------------------------------------------------------------------*/
/*									    */
/*	Funktionen und Makros zum Codieren und Decodieren von 	  	    */	
/*	Funkruf-Texten.							    */
/*									    */
/*--------------------------------------------------------------------------*/


/*
 *	Tabelle zum Umwandeln von ISO-Codierten Zeichen in
 *	POCSAG-Codierte Zeichen (Umwandlung + Bitspiegelung).
 *	Zeichen, die nicht umgewandelt werden koennen, werden
 *	auf einen Punkt '.' gemapped.
 *
 */
static unsigned char code isotab[256] =
{
	 0x00, 0x40, 0x20, 0x60, 0x10, 0x50, 0x30, 0x70,
	 0x08, 0x48, 0x28, 0x68, 0x18, 0x58, 0x38, 0x78,
	 0x04, 0x44, 0x24, 0x64, 0x14, 0x54, 0x34, 0x74,
	 0x0c, 0x4c, 0x2c, 0x6c, 0x1c, 0x5c, 0x3c, 0x7c,
	 0x02, 0x42, 0x22, 0x62, 0x12, 0x52, 0x32, 0x72,
	 0x0a, 0x4a, 0x2a, 0x6a, 0x1a, 0x5a, 0x3a, 0x7a,
	 0x06, 0x46, 0x26, 0x66, 0x16, 0x56, 0x36, 0x76,
	 0x0e, 0x4e, 0x2e, 0x6e, 0x1e, 0x5e, 0x3e, 0x7e,
	 0x01, 0x41, 0x21, 0x61, 0x11, 0x51, 0x31, 0x71,
	 0x09, 0x49, 0x29, 0x69, 0x19, 0x59, 0x39, 0x79,
	 0x05, 0x45, 0x25, 0x65, 0x15, 0x55, 0x35, 0x75,
	 0x0d, 0x4d, 0x2d, 0x6d, 0x1d, 0x5d, 0x3d, 0x7d,
	 0x03, 0x43, 0x23, 0x63, 0x13, 0x53, 0x33, 0x73,
	 0x0b, 0x4b, 0x2b, 0x6b, 0x1b, 0x5b, 0x3b, 0x7b,
	 0x07, 0x47, 0x27, 0x67, 0x17, 0x57, 0x37, 0x77,
	 0x0f, 0x4f, 0x2f, 0x6f, 0x1f, 0x5f, 0x3f, 0x7f,
	 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a,
	 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a,
	 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a,
	 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a,
	 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x01,
	 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a,
	 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a,
	 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a,
	 0x3a, 0x3a, 0x3a, 0x3a, 0x6d, 0x3a, 0x3a, 0x3a,
	 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a,
	 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x1d, 0x3a,
	 0x3a, 0x3a, 0x3a, 0x3a, 0x5d, 0x3a, 0x3a, 0x3f,
	 0x3a, 0x3a, 0x3a, 0x3a, 0x6f, 0x3a, 0x3a, 0x3a,
	 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a,
	 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x1f, 0x3a,
	 0x3a, 0x3a, 0x3a, 0x3a, 0x5f, 0x3a, 0x3a, 0x3a
};


/*
 *	Nimmt die Codierungdes Zeichens <ch> vor. Es wir
 *	das fertig codierte Zeichen zurueckgegeben.
 *
 */
#define poc_encode_char(ch)	isotab[(unsigned char)(ch)]

/*
 *	Die ISO-Tabelle eigenet sich in der unteren Haelfte
 *	auch zum decodieren von Zeichen. Wenn es notwendig ist,
 *	sollte spaeter eine zweite Tabelle eingefuehrt werden.
 *
 */
#define poc_decode_char(ch)	isotab[(unsigned char)(ch)];


/*
 *	Codiert die Ziffer <ch> in das POCSAG-Format. Alle
 *	nicht codierbaren Zeichen werden in das SPARE-Zeichen
 *	0x05 umcodiert.
 *
 */
static unsigned char poc_encode_digit (unsigned char ch)
{
	static unsigned char code mirror_tab[10] = {
		0x00, 0x08, 0x04, 0x0c, 0x02, 
		0x0a, 0x06, 0x0e, 0x01, 0x09
	};

	if (ch >= '0' && ch <= '9')
		return mirror_tab[ch - '0'];

	switch (ch) {
	case ' ':
		return 0x03;

	case 'U':
		return 0x0d;

	case '_':
		return 0x0b;

	case '[':
		return 0x0f;

	case ']':
		return 0x07;
	}

	return 0x05;
}


/*
 *	Decodiert die nach dem POCSAG-Protokoll codierte
 *	Ziffer <ch>. Es wird die klartext-Ziffer zurueckgegeben.
 *
 */
static unsigned char poc_decode_digit (unsigned char ch)
{
	static unsigned char code num_tab[16] = {
		'0', '8', '4', ' ',
		'2', ' ', '6', ']',
		'1', '9', '5', '_',
		'3', 'U', '7', '['
	};

	return num_tab[ch & 0x0f];
}


/*--------------------------------------------------------------------------*/
/*									    */
/*	Funktionen zum Codieren von kompletten Funkrufen		    */
/*									    */
/*--------------------------------------------------------------------------*/


/*
 *	Berechnet die CRC und die Paritaet fuer das Codewort <cw>.
 *	Es wird das Codewort mit eingebauter CRC und Paritaet
 *	zurueckgegeben.
 *	Voraussetzung fuer diese Funktion ist, dass die Bits 0..10
 *	beim <cw> auf 0 gesetzt sind.
 *
 */
static unsigned long poc_crc (unsigned long cw)
{
	unsigned long idata crc;
	unsigned long idata d;
	unsigned long idata m;
	unsigned char idata p;

	/*
	 * 	Berechnung der CRC
	 *
	 */
	crc = cw;
	d = 0xed200000;

	for (m = 0x80000000; !(m & 0x400); m >>= 1) {
		if (crc & m)
			crc ^= d;

		d >>= 1;
	}

	cw |= crc;

	/*
	 * 	Berechnung der Paritaet
	 *
	 */
	p = ((cw >> 24) & 0xff) ^
	    ((cw >> 16) & 0xff) ^
	    ((cw >>  8) & 0xff) ^
	            (cw & 0xff);

	p ^= (p >> 4);
	p ^= (p >> 2);
	p ^= (p >> 1);
	p &= 0x01;
	
	return cw | p;
}


/*
 *	Codiert den String <text> nach dem POCSAG-Protokoll und gibt
 *	den Zeiger auf das Resultat zurueck. Wenn ein Fehler aufgetreten
 *	ist, wird NULL zurueckgegeben.
 *
 */
struct Page xdata * poc_encode_str (unsigned long addr, 
 			            unsigned char func,
				    const char xdata *text)
{
	struct Page xdata *p;
	unsigned char cw_cnt;
	unsigned long xdata *cw_buf;
	unsigned long idata msg;
	unsigned char idata msg_bits_left;

	/*
	 *	Anzahl der benoetigten Codewoerter berechnen. und ein
	 *	gross genuges Page-Objekt erzeugen.
	 *
	 */
	cw_cnt = 1 + 
	         (POC_BITS_PER_CHAR * strlen_x(text) + POC_BITS_PER_CW - 1) / 
		 POC_BITS_PER_CW;

	if ( (p = page_new(cw_cnt, addr & 7)) == NULL )
		return NULL;

	cw_buf = p->cw_buf;

	/*
	 *	Adress-Codewort erzeugen und im Puffer speichern.
	 *
	 */
	*cw_buf++ = poc_crc(ENCODE_ACW(addr, func));

	/*
	 *	Komplette Nachricht codieren und speichern.
	 *
	 */
	msg_bits_left = POC_BITS_PER_CW;
	 
	for (; *text != 0; text++) {
		unsigned char ch = poc_encode_char(*text);
	
		/*
		 *	Der codierung-Algorithmus kann sehr viel Zeit
		 *	brauchen. Daher muss hier der Watchdog
		 *	zurueckgesetzt werden!
		 *
		 */
		WATCHDOG();
	
		if (msg_bits_left >= POC_BITS_PER_CHAR) 
		{
			msg <<= POC_BITS_PER_CHAR;
			msg |= ch;
			msg_bits_left -= POC_BITS_PER_CHAR;

			if (!msg_bits_left) {
				*cw_buf++ = poc_crc(ENCODE_MCW(msg));
				msg_bits_left = POC_BITS_PER_CW;
			}
		} 
		else 
		{
			msg <<= msg_bits_left;
			msg |= ch >> (POC_BITS_PER_CHAR - msg_bits_left);
			
			*cw_buf++ = poc_crc(ENCODE_MCW(msg));

			msg = ch;
			msg_bits_left = POC_BITS_PER_CW - POC_BITS_PER_CHAR +
			                msg_bits_left;
		} 
	}

	if (msg_bits_left != POC_BITS_PER_CW) {
		msg <<= msg_bits_left;
		*cw_buf = poc_crc(ENCODE_MCW(msg));
	}

	return p;
}


/*
 *	Erzeugt einen verschluesselten alphanumerischen Funkruf. Der 
 *	Algorithmus funktioniert folgendermassen.
 *
 *	#0x00 .. #0x20 -> #0x21 -> .. -> #0x7d -> #0x7e .. 0x7f
 *		   ^                                |
 *		   |                                |
 *		   +--------------------------------+
 *
 *	D.h. Zeichen im Intervall [0x20..0x7e] werden zyklisch nach
 *	oben geschoben. Aus 0x7e wird nach dem Schieben 0x20.
 *	
 *	Wenn die Codierung erfolgreich war, wird der fertige Funkruf
 *	zurueckgegeben, ansonsten NULL.
 *
 */
struct Page xdata * poc_encode_scrambled (unsigned long addr, 
 			                  unsigned char func,
				          const char xdata *text)
{
	char xdata *sctext;
	char xdata *w;
	const char xdata *r;
	struct Page xdata *p;

	if ( !(sctext = malloc(strlen_x(text) + 1)) )
		return NULL;

	r = text;
	w = sctext;

	while (*r != 0) {
		if (*r >= 0x20 && *r <= 0x7d)
			*w = *r + 1;
		else if (*r == 0x7e)
			*w = 0x20;
		else
			*w = *r;
		
		r++;
		w++;
	}

	*w = 0;

	p = poc_encode_str(addr, func, sctext);
	free(sctext);

	return p;
}


/*
 *	Codiert den numerischen Funkruftext <text> nach dem
 *	POCSAG-Protokoll und gibt den Zeiger auf den generierten
 *	Funkruf zurueck. Im Fehlerfall wird NULL zurueckgegeben.
 *
 */
struct Page xdata * poc_encode_num (unsigned long addr,
				    unsigned char func,
				    const char xdata *text)
{
	struct Page xdata *p;
	unsigned char cw_cnt;
	unsigned long xdata *cw_buf;
	unsigned long idata msg;
	unsigned char idata msg_bits_left;

	/*
	 *	Anzahl der benoetigten Codewoerter berechnen
	 *
	 */
	cw_cnt = 1 + 
		 (POC_BITS_PER_DIGIT * strlen_x(text) + POC_BITS_PER_CW - 1) /
		 POC_BITS_PER_CW;
		 
	if ( (p = page_new(cw_cnt, addr & 7)) == NULL )
		return NULL;

	cw_buf = p->cw_buf;
	
	/*
	 *	Adress-Codewort erzeugen und im Puffer speichern.
	 *
	 */
	*cw_buf++ = poc_crc(ENCODE_ACW(addr, func));
	
	/*
	 *	Komplette Nachricht codieren und speichern.
	 *
	 */
	msg_bits_left = POC_BITS_PER_CW;

	for (; *text != 0; text++) {
		unsigned char ch = poc_encode_digit(*text);
	
		/*
		 *	Der codierung-Algorithmus kann sehr viel Zeit
		 *	brauchen. Daher muss hier der Watchdog
		 *	zurueckgesetzt werden!
		 *
		 */
		WATCHDOG();

		msg <<= POC_BITS_PER_DIGIT;
		msg |= ch;
		msg_bits_left -= POC_BITS_PER_DIGIT;

		if (!msg_bits_left) {
			*cw_buf++ = poc_crc(ENCODE_MCW(msg));
			msg_bits_left = POC_BITS_PER_CW;
		}
	}

	/*
	 *	Wenn das letzte Codewort nicht komplett ist, wird
	 *	es mit Spaces aufgefuellt.
	 *
	 */
	if (msg_bits_left != POC_BITS_PER_CW) {
		while (msg_bits_left) {
			msg <<= POC_BITS_PER_DIGIT;
			msg |= 0x03; /* Space */
			msg_bits_left -= POC_BITS_PER_DIGIT;
		}

		*cw_buf = poc_crc(ENCODE_MCW(msg));
	}

	return p;
}


/*
 * Eine recht simple Wrapper-Funktion, die je nach Funkruftext das
 * entsprechende Codierungsverfahren auswaehlt. Das erste Zeichen
 * des Textes <text> gibt die Codierungsart an:
 *
 *	#	numerisch
 *	$	scrambled
 *	@	alphanumerisch
 *
 * Wenn der Text mit keinem der Zeichen anfaengt wird alphanumerisch
 * codiert.
 * Wenn die Codierung erfolgreich war, wird die Adresse des generierten
 * Funkrufs zurueckgegeben, ansonsten NULL.
 *
 */
struct Page xdata * poc_encode (unsigned long addr,
				unsigned char func,
				const char xdata *text)
{
	switch (*text) {
	case '#':
		return poc_encode_num(addr, func, text + 1);
		
	case '$':
		return poc_encode_scrambled(addr, func, text + 1);
		
	case '@':
		return poc_encode_str(addr, func, text + 1);
	}

	return poc_encode_str(addr, func, text);
}


/*--------------------------------------------------------------------------*/
/*									    */
/*	Funktionen zum Decodieren von Funkrufen				    */
/*									    */
/*--------------------------------------------------------------------------*/


/*
 *	Entfernt alle Nachrichten-Codewoerter und IDLE-Codewoerter
 *	am Anfang der Page-Sequenz.
 *
 */
void poc_trim (struct PageSequence xdata *ps)
{
	while (ps->cw_cnt != 0) {
		/*
		 *	Wenn z.B. lange Idle-Sequenzen in <ps> enthalten
		 *	sind, braucht dieser Algorithmus entsprechend viel
		 *	Zeit. Deshalb muss unbedingt der Watchdog-Timer
		 *	zurueckgesetzt werden.
		 *
		 */
		WATCHDOG();
	
		if (!IS_MCW(*ps->cw_buf) && *ps->cw_buf != CW_IDLE)
			break;

		ps->cw_buf++;
		ps->cw_cnt--;
		ps->cw_off++;
	}

	ps->cw_off &= 0x0f;
}


/*
 *	Trennt von der PageSequence <*ps> den ersten Funkruf ab und gibt
 *	einen Zeiger auf ihn zurueck. Wenn das Abtrennen nicht geklappt hat,
 *	wird NULL zurueckgegeben.
 *
 */
struct Page xdata * poc_splitoff (struct PageSequence xdata *ps)
{
	struct Page xdata *p;

	/*
	 *	Der Anfang eines Funkrufs wird immer durch ein Adress-Codewort
	 *	eingeleitet. Deshalb werden vorsorglich alle nicht notwendigen
	 *	Codewoerter am Anfang der Sequenz entfernt.
	 *
	 */
	poc_trim(ps);

	if (ps->cw_cnt == 0)
		return NULL;

	/*
	 *	Funkruf-Objekt reservieren und Anfang des Funkrufs darin
	 *	vermerken.
	 *
	 */
	if ( !(p = page_new(0, (ps->cw_off >> 1) & 0x07)) )
		return NULL;

	p->cw_buf = ps->cw_buf;
	p->cw_cnt = 0;

	/*
	 *	Ende des Funkrufs suchen und die Laenge in p->cw_cnt
	 *	vermerken.
	 *
	 */
	do {
		/*
		 *	Bei sehr langen Funkrufen wuerde der Watchdog
		 *	ausloesen, deshalb wird er hier zurueckgesetzt.
		 *
		 */
		WATCHDOG();
	
		ps->cw_buf++;
		ps->cw_cnt--;
		ps->cw_off++;

		p->cw_cnt++;

		if (!IS_MCW(*ps->cw_buf))
			break;
		
	} while (ps->cw_cnt != 0 && p->cw_cnt < PAGE_CWS_MAX);
	 
	/*
	 *	Offset anpassen. Und abgespaltenen Funkruf
	 *	zurueckgeben.
	 *
	 */
	ps->cw_off &= 0x0f;

	return p;
}


/*
 *	Gibt die Adresse zurueck, an den Funkruf <page>
 *	gerichtet ist.
 *
 */
void poc_decode_addr (struct Page xdata *page,
		      unsigned long xdata *addr,
		      unsigned char xdata *func)
{
	if (addr) {
		*addr = DECODE_ACW(*page->cw_buf);
		*addr |= page->frame;
	}

	if (func)
		*func = DECODE_FUNC(*page->cw_buf);
}


/*
 *	Decodiert den Funkruf <page> so, als waehre er ein
 *	alphanumerischer Funkruf. Es wird der Zeiger auf den
 *	dynamisch reservierten Funkruftext zurueckgegeben.
 *
 */
char xdata * poc_decode_str (struct Page xdata *page)
{
	unsigned char cw_cnt;
	unsigned long xdata *cw_buf;
	unsigned char idata bit_cnt;
	unsigned long idata bit_buf;
	unsigned int len;
	char xdata *str;
	char xdata *s;

	if (!page || page->cw_cnt <= 1)
		return NULL;

	cw_cnt = page->cw_cnt - 1;
	cw_buf = page->cw_buf + 1;

	/*
	 *	Benoetigte String-Laenge ausrechnen und
	 *	Speicher reservieren.
	 *
	 */
	len = ((unsigned int)cw_cnt * POC_BITS_PER_CW) / POC_BITS_PER_CHAR + 2;

	if ( !(str = malloc(len)) )
		return NULL;

	s = str;
	bit_cnt = 0;
	bit_buf = 0;

	/*
	 *	Laeuft durch alle Codewoerter, gruppuert die Message-Bits
	 *	zusammen, extrahiert einzelne Zeichen daraus, decodiert
	 *	sie und speichert sie in dem Ergebnis-String.
	 *
	 */
	for (;;) {
		unsigned char ch;

		WATCHDOG();

		/*
		 *	Wenn nicht mehr genuegend Bits im Bit-Buffer
		 *	sind, wird er mit den Daten des naechsten
		 *	Codeworts gefuellt. Wenn kein Codewort mehr
		 *	vorhanden ist, ist die Decodierung abgeschlossen.
		 *
		 */
		if (bit_cnt < POC_BITS_PER_CHAR) {
			if (cw_cnt == 0)
				break;
		
			bit_buf |= DECODE_MCW(*cw_buf) 
				   << (32 - bit_cnt - POC_BITS_PER_CW);
				   
			bit_cnt += POC_BITS_PER_CW;

			cw_buf++;
			cw_cnt--;
		}

		/*
		 *	Zeichen extrahieren, decodieren und im
		 *	String speichern.
		 *
		 */
		ch = (bit_buf & 0xfe000000UL) >> 25;
		bit_buf <<= POC_BITS_PER_CHAR;
		bit_cnt  -= POC_BITS_PER_CHAR;

		*s++ = poc_decode_char(ch);
	}
	
	*s = 0;

	return str;
}


/*
 *	Decodiert den Funkruf <page>, als waehre es ein
 *	verschluesselter Funkruf. Ansonsten verhaelt sich
 *	diese Funktion genauso wie poc_decod_str().
 *	Fuer den Verschluesselungsalgorithmus siehe 
 *	poc_encode_scrambled().
 *
 */
char xdata * poc_decode_scrambled (struct Page xdata *page)
{
	char xdata *str;
	char xdata *s;

	if ( !(str = poc_decode_str(page)) )
		return NULL;

	for (s = str; *s != 0; s++)
		if (*s >= 0x21 && *s <= 0x7e)
			(*s)--;
		else if (*s == 0x20)
			*s = 0x7e;

	return str;
}



/*
 *	Decodiert den Funkruf <page>, als sei es ein 
 *	numerischer Funkruf. Es wird der Zeiger auf
 *	den decodierten numerischen String zurueckgegeben.
 *	Die aufrufende Funktion muss den String mit free()
 *	entsorgen.
 *
 */
char xdata * poc_decode_num (struct Page xdata *page)
{
	unsigned char cw_cnt;
	unsigned long xdata *cw_buf;
	unsigned char idata bit_cnt;
	unsigned long idata bit_buf;
	unsigned int len;
	char xdata *str;
	char xdata *s;

	if (!page || page->cw_cnt <= 1)
		return NULL;

	cw_cnt = page->cw_cnt - 1;
	cw_buf = page->cw_buf + 1;

	/*
	 *	Benoetigte String-Laenge ausrechnen und
	 *	Speicher reservieren.
	 *
	 */
	len = ((unsigned int)cw_cnt * POC_BITS_PER_CW) / POC_BITS_PER_DIGIT + 2;

	if ( !(str = malloc(len)) )
		return NULL;

	s = str;
	bit_cnt = 0;
	bit_buf = 0;

	/*
	 *	Laeuft durch alle Codewoerter, gruppuert die Message-Bits
	 *	zusammen, extrahiert einzelne Zeichen daraus, decodiert
	 *	sie und speichert sie in dem Ergebnis-String.
	 *
	 */
	for (;;) {
		unsigned char ch;

		WATCHDOG();

		/*
		 *	Wenn nicht mehr genuegend Bits im Bit-Buffer
		 *	sind, wird er mit den Daten des naechsten
		 *	Codeworts gefuellt. Wenn kein Codewort mehr
		 *	vorhanden ist, ist die Decodierung abgeschlossen.
		 *
		 */
		if (bit_cnt < POC_BITS_PER_DIGIT) {
			if (cw_cnt == 0)
				break;
		
			bit_buf |= DECODE_MCW(*cw_buf) 
				   << (32 - bit_cnt - POC_BITS_PER_CW);
				   
			bit_cnt += POC_BITS_PER_CW;

			cw_buf++;
			cw_cnt--;
		}

		/*
		 *	Ziffer extrahieren, decodieren und im
		 *	String speichern.
		 *
		 */
		ch = (bit_buf & 0xf0000000UL) >> 28;
		bit_buf <<= POC_BITS_PER_DIGIT;
		bit_cnt  -= POC_BITS_PER_DIGIT;

		*s++ = poc_decode_digit(ch);
	}
	
	*s = 0;

	return str;
}


