
/*
 *	$Id: repeater.c,v 1.20 2002/05/12 12:35:04 Klaus Exp $
 *
 *	repeater.c	Implementation des Funkruf-Repeaters
 *
 *
 *
 *
 *	(c) 2001,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "repeater.h"
#include "config.h"
#include "scc.h"
#include "mm.h"
#include "pocsag.h"
#include "page.h"
#include "rtc.h"
#include "system.h"
#include "ax25.h"
#include "ax25_io.h"
#include "string.h"
#include "trace.h"

/*
 *	Konfigurationsdaten aus dem Konfigurationsmodul
 *	beziehen.
 *
 */
#define rp_ttab		conf.rp_ttab

#define rp_ident_addr	conf.ident_addr
#define rp_ident_text	conf.ident_text

/*
 *	Zustaende des Receivers.
 *
 */
#define STATE_OFF		0
#define	STATE_IDLE		1
#define STATE_RECEIVE		2
#define STATE_XMIT		3
#define STATE_WAIT_ON_XMIT	4

/*
 *	Aktueller Zustand des Repeaters
 *
 */
static unsigned char idata state;

/*
 *	Empfangsdaten
 *
 */
static unsigned long xdata *rx_cwbuf;	/* Empfangspuffer */
static stime_t              rx_time;	/* Empfangszeitpunkt */

/*
 *	Liste der Funkrufe, die im REPEAT-State
 *	ausgesendet werden sollen.
 *	tx_tstat zeigt auf die Status-Instanz des Triggers,
 *	von dem die Sendedaten stammen.
 *
 */
static struct PageQueue xdata tx_queue;
static unsigned char xdata tx_speed;
static struct TriggerState xdata *tx_tstat;

/*
 *	Statistik-Daten jedes Triggers
 *
 */
static struct TriggerState xdata rp_tstat[TRIGGER_MAX];


static char rp_mode (unsigned char on);

/*--------------------------------------------------------------------------*/
/*									    */
/*	VERWALTUNGSFUNKTIONEN RUND UM DIE REPEATER-TRIGGER		    */
/*									    */
/*--------------------------------------------------------------------------*/


/*
 *	Gibt den Index in der Trigger-Tabelle des Triggers mit der 
 *	Adresse <addr> zurueck. Wenn keine Trigger mit der Adresse
 *	<addr> existiert, wird -1 zurueckggeben.
 *
 */
static char rp_trigger_lookup (unsigned long addr)
{
	unsigned char idx;

	for (idx = 0; idx < TRIGGER_MAX; idx++)
		if (rp_ttab[idx].in_use && rp_ttab[idx].addr == addr)
			return idx;

	return -1;
}


/*
 *	Preuft, ob ein Trigger in die Trigger-Tabelle eingetragen
 *	ist. Wenn ja, wird 1 zurueckgegeben, ansonsten 0.
 *
 */
static char rp_trigger_active ()
{
	struct Trigger xdata *t;

	for (t = rp_ttab; t != rp_ttab + TRIGGER_MAX; t++)
		if (t->in_use)
			return 1;

	return 0;
}


/*
 *	Loescht alle Trigger aus der Trigger-Tabelle.
 *	
 *	An dieser Stelle darf rp_mode(0) nicht aufgerufen
 *	werden, weil der Repeater ggf. noch nicht
 *	initialisiert ist.
 *
 */
void rp_trigger_clear ()
{
	struct Trigger xdata *t;	

	for (t = rp_ttab; t != rp_ttab + TRIGGER_MAX; t++)
		t->in_use = 0;
}


/*
 *	Fuegt einen Trigger mit der Adresse <addr> hinzu.
 *	Wenn kein Trigger mehr hinzugefuegt werden kann, wird 0
 *	zurueckgegeben, ansonsten 1.
 *
 */
char rp_trigger_add (unsigned long addr)
{
	unsigned char idx;

	if (rp_trigger_lookup(addr) != -1)
		return 0;

	for (idx = 0; idx < TRIGGER_MAX; idx++) {
		struct Trigger xdata *t = &rp_ttab[idx];

		if (!t->in_use) {
			t->addr   = addr;
			t->in_use = 1;

			memset_x(&rp_tstat[idx], 0, 
				 sizeof(struct TriggerState));

			rp_mode(1);
			return 1;
		}
	}

	return 0;
}


/*
 *	Entfernt den Trigger mit der Adresse <addr> aus der
 * 	Trigger-Tabelle. Wenn der Trigger entfernt wurde
 *	wird 1 zurueckgegeben, ansonsten 0.
 *
 */
char rp_trigger_del (unsigned long addr)
{
	char idx;

	if ( (idx = rp_trigger_lookup(addr)) == -1 )
		return 0;

	rp_ttab[idx].in_use = 0;

	if (!rp_trigger_active())
		rp_mode(0);
		
	return 1;
}


/*
 *	Gibt die Trigger-Liste mit ein paar Statistik-Daten
 *	aus.
 *
 */
void rp_trigger_list (struct AX25_cb xdata *axcb)
{
	unsigned char idx;

	ax25_printf(axcb, "address  lcwc        rx        tx  loss\r");
	/*                 2000000  2097  29682273  29546294    2% */

	for (idx = 0; idx < TRIGGER_MAX; idx++) {
		struct TriggerState xdata *ts = &rp_tstat[idx];
		unsigned int loss;

		if (!rp_ttab[idx].in_use)
			continue;

		if (ts->rx != 0)
			loss = (100 * (ts->rx - ts->tx)) / ts->rx;
		else
			loss = 0;

		ax25_printf(axcb, "%7lu%6u%10lu%10lu%5u%%\r",
			    (unsigned long)rp_ttab[idx].addr,
			    (unsigned int)ts->lcwc,
			    (unsigned long)ts->rx,
			    (unsigned long)ts->tx,
			    (unsigned int)loss);
	}
}


/*--------------------------------------------------------------------------*/
/*									    */
/*	REPEATER-FUNKTIONEN						    */
/*									    */
/*--------------------------------------------------------------------------*/


/*
 *	Schaltet den Repeater ein (on != 0) oder
 *	aus (on == 0).
 *	Im Fehlerfall wird 0 zurueckgegeben, ansonsten 1.
 *
 */
static char rp_mode (unsigned char on)
{
	if (state != STATE_OFF && !on) {
		/* Receiver abschalten */
		scc_recv_mode(0);

		/* reservierten Speicher freigeben */
		pq_clear(&tx_queue);
		free(rx_cwbuf);

		/* Zustand auf "aus" stellen */
		state = STATE_OFF;
		return 1;
	}

	if (state == STATE_OFF && on) {
		if ( !(rx_cwbuf = malloc(RP_CW_MAX * 4)) )
			return 0;

		pq_init(&tx_queue);
		scc_recv_mode(1);
		state = STATE_IDLE;
	}

	return 1;
}


/*
 *	Initialisiert den Repeater.
 *
 */
void rp_init ()
{
	state = STATE_OFF;

	memset_x(rp_tstat, 0, TRIGGER_MAX * sizeof(struct TriggerState));	

	if (rp_trigger_active())
		rp_mode(1);
}


/*
 *	Empfangsfunktion des Repeaters. Hier werden
 *	alle Empfangenen Sequenzen entgegengenommen, ausgewertet
 *	und ggf. eine Wieder-Aussendung angestossen.
 *
 */
static void rp_receive ()
{
	struct PageSequence xdata ps;
	struct Page xdata *p;
	unsigned long xdata addr;
	char idx;

	/*
	 *	Warten bis etwas empfangen wurde. Wenn Daten
	 *	anstehen wird das Page-Sequenz-Objekt initialisiert
	 *	und die Statistik-Daten gesetzt.
	 *
	 */
	if ( !(ps.cw_cnt = scc_recv_poll(&tx_speed)) )
		return;

	ps.cw_off = 0;
	ps.cw_buf = rx_cwbuf;

	rx_time = rtc_time();

	trace_print_ps('R', tx_speed, &ps);

	/*
	 *	Der erste Funkruf in der Sequenz muss vom Master
	 *	kommen. Wenn der erste Funkruf nicht abgespalten
	 *	werden kann oder nicht von einem unserer Master kommt,
	 *	wird die empfangene Sequenz verworfen.
	 *
	 */
	if ( (p = poc_splitoff(&ps)) == NULL ) {
		state = STATE_IDLE;
		return;
	}

	poc_decode_addr(p, &addr, NULL);

	if ( (idx = rp_trigger_lookup(addr)) == -1 ) {
		page_delete(p);
		state = STATE_IDLE;
		return;
	}

	tx_tstat = &rp_tstat[idx];
	tx_tstat->lcwc = ps.cw_cnt;
	tx_tstat->rx++;

	/*
	 *	Ok, der Funkruf stammt tatsaechlich von unserem
	 *	Master. Nun wird der Master-Identity-Funkruf
	 *	weggeworfen und ggf. ein eigener generiert und in
	 *	die TX-Queue eingehaengt. der Rest der Page-Sequenz
	 *	wird in einen Funkruf umgewandelt und ebenfalls in
	 *	die Queue gehaengt.
	 *
	 */
	page_delete(p);

	if (rp_ident_addr) {
		p = poc_encode(rp_ident_addr, 1, rp_ident_text);

		if (p != NULL) {
			p->domain = PAGE_DOMAIN_SYSTEM;
			pq_enqueue(&tx_queue, p, 1);
		}
	}
	
	poc_trim(&ps);

	while ( (p = page_split(&ps, PAGE_CWS_MAX)) != NULL )
		pq_enqueue(&tx_queue, p, 0);

	/*
	 *	Wenn sich nun keine Funkrufe in der TX-Queue befinden
	 *	wird die Page-Sequenz verworfen. 
	 *
	 */
	if (!pq_first(&tx_queue)) {
		state = STATE_IDLE;
		return;
	}

	/*
	 *	Aussendung anstossen
	 *
	 */
	state = STATE_XMIT;
}



/*
 *	Hauptschleife des Repeaters
 *
 */
void rp_run ()
{
	switch (state) {
	case STATE_OFF:
		break;
		
	case STATE_IDLE:
		if (!scc_recv(rx_cwbuf, RP_CW_MAX))
			break;

		state = STATE_RECEIVE;
		break;

	case STATE_RECEIVE:
		rp_receive();
		break;

	case STATE_XMIT:
		if (scc_xmit(tx_speed, &tx_queue)) {
			state = STATE_WAIT_ON_XMIT;
			break;
		}
		
		if (rtc_time() - rx_time > RP_DELAY_MAX) {
			pq_clear(&tx_queue);
			state = STATE_IDLE;
		}
		
		break;

	case STATE_WAIT_ON_XMIT:
		if (scc_xmitr_state()) {
			tx_tstat->tx++;
			pq_clear(&tx_queue);
			state = STATE_IDLE;
		}
		
		break;

	default:
		state = STATE_OFF;
 	}
}



