
/*
 *	$Id: rtc.c,v 1.3 2002/04/01 12:56:57 klaus Exp $
 *
 *	rtc.c		Emulation einer Real Time Clock
 *
 *			Das Zeitnormal fuer die RTC wird von
 *			der Compare-Logik des 80C552 erzeugt.
 *			Somit ist die RTC quarzstabil, d.h.
 *			ihre Genauigkeit haengt von der
 *			Genauigkeit des Prozessortaktes ab.
 *			Fuer unsere Anwendungen sollte das
 *			ausreichen.
 *
 *
 *
 *	(c) 2000,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "system.h"
#include "rtc.h"
#include "timer.h"
#include "mm.h"
#include "sched.h"

/*
 *	Variablen fuer die Alarm-Funktion der RTC. Wennder aktuelle 
 *	Slot gleich <alarm_slot> ist, wird die Funktion <alarm_func>
 *	aufgerufen.
 *
 */
static unsigned char idata alarm_slot;
static void (* idata alarm_func)();

/*
 *	Funkrufzeit. Sie wird aus folgender Formel berechnet
 *
 *	(<UNIX-Zeit * 10> + <1/10 Sekunden>) & 0xffff
 *
 */
stime_t idata systime;

/*
 *	letzter gesetzter CM2 Registerwert
 *
 */
static unsigned int idata last_cm2;

/*---------------------------------------------------------------------------*/


/*
 *	Initialisiert die RTC. Sie beginnt bei 00:00:00 an zu laufen und
 *	muss spaeter mit rtc_set() auf die tatsaechliche Uhrzeit gesetzt
 *	werden.
 *
 */
void rtc_init ()
{
	alarm_func = NULL;

	systime = 0;

	last_cm2 = RTC_INTERVAL;
	CML2 = last_cm2 & 0xff;
	CMH2 = last_cm2 >> 8;

	TM2CON = 0x05;	/* T2_clk_src = Fosc/12, prescale = 2 */
	PCM2 = 0;	/* niedrige Prioritaet fuer CM2 */
	ECM2 = 1;	/* Interrupt fuer CM2 zulassen */
}


/*
 *	Aktiviert/Deaktiviert die Alarm-Funktion.
 *	Wenn die Systemzeit gleich <time> ist, wird die
 *	Funktion <func> aufgerufen. Wenn <func> == NULL ist,
 *	ist die Alarm-Funktion ausgeschaltet.
 *
 *	ACHTUNG: Die Funktion <func> muss die Registerbank REGB_LP
 *		 benutzen!
 *
 */
void rtc_alarm (unsigned char slot, void (*func)())
{
	ECM2 = 0;
	alarm_slot = slot;
	alarm_func = func;
	ECM2 = 1;
}


/*
 *	Diese Funktion wird alle 100ms aufgerufen und
 *	zaehlt die Uhr weiter...
 *
 */
void rtc_isr () interrupt INT_CM2 using REGB_LP
{
	last_cm2 += RTC_INTERVAL;
	CML2 = last_cm2 & 0xff;
	CMH2 = last_cm2 >> 8;
	CMI2 = 0;

	/*
	 *	Timer-Register vom Timer-Modul
	 *	incrementieren.
	 *
	 */
	t_reg++;

	EA = 0;	/* Increment atomar machen */
	systime++;
	EA = 1;

	/*
	 *	Alarm-Zeit vergleichen und ggf. die
	 *	Alarm-Funktion aufrufen.
	 *	
	 */
	if (alarm_func != NULL && SCHED_SLOT(systime) == alarm_slot) {
		alarm_func();
		alarm_func = NULL;
	}
}


/*
 *	Gibt die Systemzeit zurueck.
 *
 */
stime_t rtc_time ()
{
	register stime_t time;

	ECM2 = 0;
	time = systime;
	ECM2 = 1;
	
	return time;
}


/*
 *	Setzt die angegebene Systemzeit
 *
 */
void rtc_set (stime_t stime)
{
	ECM2 = 0;
	systime = stime;
	ECM2 = 1;
}

