
/*
 *	$Id: scc.c,v 1.20 2002/05/18 11:46:12 Klaus Exp $
 *
 *	scc.c		Serial Communication Controller
 *
 *
 *
 *
 *	(c) 2000,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "scc.h"
#include "system.h"
#include "mm.h"
#include "config.h"
#include "dectabs.h"
#include "rtc.h"
#include "page.h"

/*
 *	Zustaende des Transmitters 
 *
 */
#define STATE_TX_IDLE		0	
#define STATE_TX_CAL_START	1
#define STATE_TX_CAL_RESTART	2
#define STATE_TX_CAL		3
#define STATE_TX_CAL_PAUSE	4
#define STATE_TX_START		5
#define STATE_TX_HEAD		6		
#define STATE_TX_PREAMBLE	7	
#define STATE_TX_DATA		8
#define STATE_TX_TAIL		9
#define STATE_TX_OFF		10

/*
 *	Zustaende des Receivers
 *
 */
#define STATE_RX_IDLE		0
#define STATE_RX_SYNC		1
#define STATE_RX_DATA		2
#define STATE_RX_DISCARD	3

/*
 *	Zeitangaben fuer den Kalibrationsmodus in 1/10 Sekunden
 *
 */
#define CAL_DURATION		600	/* CAL Durchgangsdauer */
#define CAL_PAUSE		10	/* Pausendauer */

/*
 *	AutoSensing-Parameter
 *
 */
#define AS_SAMPLE_CNT		32
#define AS_THRESHOLD		20

/*
 *	DCD-Parameter
 *
 */
#define DCD_CLIP_LOW		0
#define DCD_CLIP_HIGH		38	
#define DCD_THRESHOLD		30	
#define DCD_XMIT_DELAY		2	/* in 1/10 Sekunden */

/*
 *	Parameter aus dem Konfigurations-Modul beziehen.
 *
 */
#define scc_txhead 	conf.scc_txhead
#define scc_txtail 	conf.scc_txtail
#define scc_txinvert 	conf.scc_txinvert
#define scc_rxinvert	conf.scc_rxinvert
#define scc_preamble	conf.scc_preamble

/*
 * Tabelle, mit der die speed-Angaben SCC_B512, SCC_B1200, etc.
 * in Zeitkonstanten umgewandelt werden.
 *
 */
static unsigned int code tc_tab[SCC_BCNT] = {
	SCC_TC_512BPS,
	SCC_TC_1200BPS,
	SCC_TC_2400BPS
};


/*--------------------------------------------------------------------------*/
/*									    */
/*			DCD- und Autosensing-Code			    */
/*									    */
/*--------------------------------------------------------------------------*/

static struct DecTab code * idata rx_dectab;	/* Zeiger auf Dec-Tabelle */

static unsigned char idata as_sample_cnt;	/* Anzahl Flankenwechsel */	
static unsigned char idata as_oktab[SCC_BCNT];

static bit dcd;					/* DCD-Zustand */
static unsigned char idata dcd_reg;
static stime_t xdata dcd_drop_time;		/* Zeitpunk, zu dem die DCD
						   abgeschaltet wurde */

/*
 *	Initialisiert das Auto-Sensing
 *
 */
static void scc_as_init () using REGB_HP
{
	unsigned char speed;
	
	rx_dectab = NULL;

	for (speed = SCC_BMIN; speed <= SCC_BMAX; speed++)
		as_oktab[speed] = 0;

	as_sample_cnt = AS_SAMPLE_CNT;
}


/*
 *	Versucht anhand der AutoSensing-Tabelle zu ermitteln, auf welche
 *	Datenrate der Empfaenger eingestellt werden muss. Das AutoSensing
 *	funktioniert in der Regel nur dann, wenn gerade eine Praeambel
 *	empfangen wird.
 *
 */
static void scc_as (unsigned int dt) using REGB_HP
{
	register unsigned char speed;
	unsigned char s;
	
	if (dt < sizeof(astab)) {
		speed = astab[dt];

		if (speed & 0x80)
			as_oktab[speed & 0x7f]++;
	}

	if (--as_sample_cnt != 0)
		return;

	/*
	 *	Tabellen-Index (speed) von as_oktab heraussuchen,
	 *	der den hoechsten Zaehlerstand besitzt. Wenn dieser
	 *	Zaehlerzustand dann noch groesser oder gleich dem
	 *	Erkennungs-Schwellwert ist, dann haben wir die
	 *	korrekte Geschwindigkeit gefunden.
	 *
	 */
	speed = SCC_BMIN;

	for (s = SCC_BMIN + 1; s <= SCC_BMAX; s++)
		if (as_oktab[s] > as_oktab[speed])
			speed = s;

	if (as_oktab[speed] >= AS_THRESHOLD) {
		rx_dectab = &dectab[speed];
		return;
	}

	scc_as_init();	
}


/*
 *	Setzt die Zaehler zurueck, die zum Generieren des DCD-Signals verwendet
 *	werden. Ausserdem wird die DCD-LED ausgeschaltet.
 *
 */
static void scc_dcd_init () using REGB_HP
{
	if (dcd)
		dcd_drop_time = systime;

	OFF(LED_DCD);

	dcd_reg = 0;
	dcd     = 0;
}


/*
 *	Algorithmus zur Generierung des DCD-Signals. Der
 *	DCD-Algorithmus ist so implementiert, dass er moeglichst
 *	schnell abgearbeitet werden kann.
 *
 */
static void scc_dcd_compute (unsigned char rdcd) using REGB_HP
{
	if (rdcd) {
		if (dcd_reg < DCD_CLIP_HIGH)
			dcd_reg++;

		if (dcd)
			return;
		
		if (dcd_reg >= DCD_THRESHOLD) {
			dcd = 1;
			ON(LED_DCD);
		}
	} else {
		if (dcd_reg > DCD_CLIP_LOW)
			dcd_reg--;

		if (!dcd)
			return;

		if (dcd_reg < DCD_THRESHOLD)
			scc_dcd_init();
	}
}


/*--------------------------------------------------------------------------*/
/*									    */
/*			R E C E I V E R 				    */
/*									    */
/*--------------------------------------------------------------------------*/

static unsigned char volatile idata rx_state;	/* Zustand des Receivers */
static bit rx_enabled;				/* 1 == Receiver an */
static char xdata rx_cw_cntr;			/* Codewortzaehler */

/*
 *	Puffer, in den die empfangenen Codewoerter abgelegt werden.
 *	Synchron-Codewoerter werden nicht mit abgespeichert!
 *
 */
static unsigned long xdata * idata rx_buf; 	/* Zeiger auf naechstes CW */
static unsigned int idata rx_buflen;	/* urspruenglicher Platz im Puffer */
static unsigned int idata rx_bufroom;	/* Anzahl CWs noch Platz im Puffer */

static unsigned char xdata rx_speed;    /* Datenrate mit der empfangen wurde */
static bit rx_buf_ready;		/* Empfangsdaten liegen im Puffer */


/*
 *	Schreibt das uebergebene Codewort <cw> in den Empfangspuffer.
 *	Dabei werden alle Synchron-CWs herausgefiltert. 
 *	Es wird die Anzahl der gespeicherten Codewoerter zurueckgegeben.
 *
 */
static char scc_recv_cw (unsigned long cw) using REGB_HP
{
	if (rx_bufroom == 0)
		return 0;

	/*
	 *	Synchron-Codewoerter werden aus dem Datenstrom 
	 *	herausgefiltert, um Platz zu sparen. Es wird nicht
	 *	geprueft, ob das Synchroncodewort korrekt ist, weil
	 *	ein Bitfehler den Empfangsdurchgang nicht abbrechen
	 * 	koennen soll!
	 *
	 */
	if (rx_cw_cntr++ == -1)
		return 1;

	if (rx_cw_cntr >= 16)
		rx_cw_cntr = -1;

	/*
	 *	Codewort im Puffer speichern.
	 *
	 */
	*rx_buf++ = cw;
	rx_bufroom--;

	return 1;
}


/*
 *	Nach einem kompletten Empfangsdurchlauf oder nach
 *	dem Empfangsabbruch durch DCD- oder Synchronisationsverlust
 *	wird diese Funktion aufgerufen, um den Zustand der
 *	empfangenen Daten zu beurteilen. Wenn Daten Empfangen
 *	wurden, wird das durch rx_buf_ready angezeigt. Um
 *	zu kennzeichnen, dass der Empfangspuffer nun nicht mehr
 *	gueltig ist, wird er auf NULL gesetzt.
 *
 */
static void scc_recv_resume () using REGB_HP
{
	if (rx_state == STATE_RX_DATA && rx_buflen != rx_bufroom) {
		rx_buf = NULL;
		rx_speed = rx_dectab->speed;
		rx_buf_ready = 1;
	}

	rx_state = STATE_RX_IDLE;
}


/*
 *	Gruppiert Bitsequenzen zu Codewoertern. In dieser Funktion wird
 *	die komplette Receiver-Statemachine abgewickelt.
 *
 */
static void scc_receiver (unsigned char b, unsigned char bcnt) using REGB_HP
{
	static unsigned long idata cw;
	static unsigned char idata bits_left;

	switch (rx_state) {
	case STATE_RX_IDLE:
		/*
		 *	In diesem Zustand wird die Empfangs-State-
		 *	Machine intialisiert.
		 *
		 */
		cw = 0;
		rx_state = STATE_RX_SYNC;
		
		/* kein break, da bcnt evtl. > 0 ist */
	
	case STATE_RX_SYNC:
		/*
		 *	Sucht nach dem Synchroncodewort. Wenn das 
		 *	Synchron-Codewort gefunden wurde, wird in den
		 *	Daten-Empfangs-Zustand gewechselt.
		 *
		 */
		for (; bcnt != 0 && cw != CW_SYNC; bcnt--) {
			cw <<= 1;
			
			if (b) 
				cw |= 1;
		}

		if (cw != CW_SYNC)
			break;

		if (rx_buf != NULL) {
			rx_state = STATE_RX_DATA;
		} else {
			rx_state = STATE_RX_DISCARD;
			break;
		}

		rx_bufroom = rx_buflen;
		rx_cw_cntr = 0;		/* Mit CW 0/Frame 0 beginnen */

		bits_left  = 32;	/* 32 Bits zu empfangen */

		/* kein break, da bcnt evtl. > 0 ist */
	
	case STATE_RX_DATA:
		/*
		 *	Empfangene Bits zu Codewoertern gruppieren und
		 *	im Empfangspuffer ablegen.
		 *
		 */
		for (; bcnt != 0; bcnt--) {
			cw <<= 1;

			if (b)
				cw |= 1;

			if (--bits_left == 0) {
				bits_left = 32;
				scc_recv_cw(cw);
			}
		}
		break;

	case STATE_RX_DISCARD:
		/*
		 *	Alle empfangenen Daten verwerfen. Dieser Zustand
		 *	wird eingenommen, wenn Daten empfangen werden,
		 *	jedoch kein Puffer zur Aufnahme der Daten bereit-
		 *	steht.
		 *
		 */
		break;

	default:
		scc_recv_resume();
	}
}


/*
 *	Schaltet um ein 3/4-Bit verzoegert den Capture-
 *	Interrupt wieder ein. So wird die Systembelastung
 *	bei anliegendem Rauschen stark reduziert.
 *
 */
void scc_t0_isr () interrupt INT_T0 using REGB_HP
{
	TF0  = 0;	/* T0 overflow reset */
	ECT0 = 1;	/* capture interrupt 0 einschalten */
}


/*
 *	Nach einem Flankenwechsel an RXD wird diese Funktion aufgerufen.
 *	Im Capture-Register CT0 ist dann die Zeit in T2-Ticks zwischen den 
 *	letzten beiden Flankenwechseln abgelegt.
 *
 */
void scc_ct0_isr () interrupt INT_CT0 using REGB_HP
{
	static unsigned int idata last_t2;
	static bit dcd_old = 0;
	register unsigned int ct0;
	register unsigned int dt;
	register unsigned char b;
	register unsigned char bcnt;

	ct0 = ((unsigned int)CTH0 << 8) | (unsigned int)CTL0;
	b   = (~P1 & RXD);

	CTI0 = 0;
	ECT0 = 0;

	/*
	 *	Die Zeitdifferenz wird auf 1/16 heruntergerechnet, damit
	 *	mit diesem Wert die Decoder-Tabelle indiziert werden kann.
	 *	Die verbleibende Timer-Aufloesung entspricht 24 Abtastungen
	 *	pro Bit bei 1200bps.
	 *
	 */
	dt = ct0 - last_t2;
	last_t2 = ct0;
	dt >>= 4;

	/*
	 *	Der Capture-Interrupt wird erst verzoegert wieder
	 *	aktiviert. Das bewirkt, dass z.B. bei anliegendem
	 *	Rauschen das System nicht zu stark belastet wird.
	 *
	 */
	if (!rx_dectab) {
		TH0 = SCC_TC_AS >> 8;
		TL0 = SCC_TC_AS & 0xff;
		TR0 = 1;
		scc_as(dt);
		return;
	}
	
	TH0 = rx_dectab->tc_bitlen_75 >> 8;
	TL0 = rx_dectab->tc_bitlen_75 & 0xff;
	TR0 = 1;

	/*
	 *	Znaechst wird das DCD-Signal aus den roh-Daten generiert.
	 *	Wenn die roh-Daten nicht mit der aktuellen decoder-Tabelle
	 *	decodiert werden koennen, wird einfach der letzte Wert aus
	 *	der Decoder-Tabelle benutzt.
	 *
	 */
	if (dt >= rx_dectab->tablen)
		dt = rx_dectab->tablen - 1;

	bcnt = rx_dectab->tab[dt];
	scc_dcd_compute(bcnt & 0x80);

	if (!dcd) {
		if (dcd_old) {
			scc_recv_resume();
			scc_as_init();
			dcd_old = 0;
		}

		return;
	}

	dcd_old = 1;

	/*
	 *	Empfgangene Bit-Sequenz ggf. invertieren und dann
	 *	an die Receiver-Statemachine weitergeben.
	 *
	 */
	if (scc_rxinvert)
		b ^= RXD;
	
	scc_receiver(b, bcnt & 0x7f);
}


/*
 * 	Wenn der Receiver eingeschaltet ist, wird er bis
 *	zum naechsten Aufruf von scc_recv_unlock() deaktiviert.
 *
 */
static void scc_recv_lock ()
{
	if (rx_enabled) {
		ET0  = 0;
		ECT0 = 0;
	}
}


/*
 *	Aktiviert den durch scc_recv_lock() deaktivierten
 *	Receiver wieder.
 *
 */
static void scc_recv_unlock ()
{
	if (rx_enabled) {
		ET0  = 1;
		ECT0 = 1;
	}
}


/*--------------------------------------------------------------------------*/
/*									    */
/*			T R A N S M I T T E R				    */
/*									    */
/*--------------------------------------------------------------------------*/

static unsigned char volatile idata tx_state;	/* Zustand des Trabsmitters */
static unsigned int idata tx_bitlen;		/* Bitlaenge in T2-Ticks */
						/* bei akt. Datenrate */

/*
 *	Die folgenden beiden Variablen werden zur grundlegenden Aussendung
 *	von Datenbits verwendet. Die Aussendung erfolgt Codewort-weise,
 *	beginnend mit dem MSB.
 *	D.h. in <tx_cw> wird das zu sendende Codewort geschrieben und
 *	<tx_bits_left> muss auf die Laenge des Codeworts in Bits (hier
 *	immer 32 Bits) gesetzt werden. Der Sende-Algorithmus sendet dann
 *	alle Bits aus und ruft danach die Funktion scc_loadtxcw() auf,
 *	um <tx_cw>/<tx_bits_left> nachzuladen.
 *
 */
static unsigned long idata tx_cw;
static unsigned char idata tx_bits_left;

/*
 *	Zeiger auf aktuellen Funkruf. In Wirklichkeit wird auf eine
 *	Funkrufliste gezeigt, die abgearbeitet werden soll.
 *	Zusaetzlich wird der Codewort-Puffer in eigenen Variablen gehalten,
 *	damit die Page-Instanz nicht modifiziert werden muss.
 *
 */
static struct Page xdata * idata   tx_page;
static unsigned char idata 	   tx_cw_cnt;
static unsigned long xdata * idata tx_cw_buf;

/*
 *	Um Speicher zu sparen werden Synchron- und Idle-Codewoerter
 *	erst beim Aussenden in die Funkrufe eingebaut. Damit der richtige
 *	Platz des entsprechenden Code-Worts gefunden werden kann wird
 *	u.a. der Codewort-Zaehler <tx_cw_cntr> benoetigt. Er wir
 *	zyklisch von -1 bis 15 durchgezaehlt. Wobei bei -1 ein Synchron-CW
 *	in den Sende-Datenstrom eingebaut werden muss. Ansonsten wird
 *	<tx_cw_cntr> noch zum Bestimmen der aktuellen Framenummer
 *	verwendet.
 *	<tx_lock> zeigt an, ob momentan ein Funkruf ausgesendet wird oder
 *	ob bis zum Anfang des naechsten Funkrufs Idle-Codewoerter ausgesendet 
 *	werden muessen.
 *
 */
static char idata tx_cw_cntr;
static bit tx_locked;

/*
 *	Maximale Dauer einer Kalibrationsaussendung in 1/10-Sekunden.
 *	Wenn die Dauer 0 ist, gibt es keine Zeitbeschraenkung.
 *
 */
static unsigned int xdata tx_cal_duration;


/*
 *	Laedt das naechste Codewort in das Senderegister. Es
 *	wird die Anzahl der geladenen Codewoerter zurueckgegeben
 *	(0 oder 1).
 *
 */
static char scc_loadtxcw () using REGB_HP
{
	/*
	 *	Synchron-Codewort einschieben.. Wenn keine Sendedaten
	 *	mehr anstehen wird das der aufrufenden Funktion
	 *	signalisiert und kein Synchron-CW gesendet. So
	 *	wird sichergestellt, dass nur ganze Batches 
	 *	ausgestrahlt werden.
	 *
	 */
	if (tx_cw_cntr == -1) 
	{
		if (!tx_page)
			return 0;
	
		tx_cw = CW_SYNC;
	} 
	else if (!tx_page) 
	{
		/*
		 *	Da keine Sendedaten mehr anstehen muss das
		 *	Batch bis zum Ende mit Idle-Codewoertern
		 *	gefuellt werden.
		 *
		 */
		tx_cw = CW_IDLE;
	} 
	else if (!tx_locked && (tx_page->frame & 0x07) != tx_cw_cntr / 2) 
	{
		/*
		 *	Solange der Sender und die Page-Liste nicht
		 *	synchron sind, muessen Idle-Codewoerter
		 *	ausgesendet werden.
		 *
		 */
		tx_cw = CW_IDLE;
	} 
	else 
	{
		tx_locked = 1;

		tx_cw = *tx_cw_buf;

		tx_cw_buf++;
		tx_cw_cnt--;

		if (!tx_cw_cnt) {
			tx_locked = 0;
			
			tx_page   = pq_next(tx_page);
			tx_cw_cnt = tx_page->cw_cnt;
			tx_cw_buf = tx_page->cw_buf;
		}
	}
		
	tx_bits_left = 32;

	/*
	 *	Nach 16 ausgesendeten Codewoertern ist das Batch fertig und
	 *	es muss ein neuer Zyklus mit einem Synchron-CW angefangen
	 *	werden.
	 *
	 */
	if (++tx_cw_cntr >= 16)
		tx_cw_cntr = -1;
	
	return 1;
}


/*
 *	Ermittelt das naechste auszusendende Bit
 *
 *	Rueckgabewert:
 *
 *	0	0-Bit
 *	1	1-Bit
 *	-1	keine weiteren Sendedaten	
 *
 */
static char scc_gettxbit () using REGB_HP
{
	register char b;

	if (!tx_bits_left && !scc_loadtxcw())
		return -1;

	b = tx_cw & 0x80000000 ? 1 : 0;

	tx_cw <<= 1;
	tx_bits_left--;

	return b;
}


/*
 *	Setzt nach <ticks> T2-Ticks den TXD-Pegel auf den
 *	logischen Wert <l>.
 *
 *	ACHTUNG: Laut POCSAG-Protokoll ist die Frequenzlage.
 *	         invertiert!
 *
 */
static void scc_set_txd (unsigned int ticks, unsigned char l) using REGB_HP
{
	static unsigned int idata last_t2;

	last_t2 += ticks;

	if (scc_txinvert)
		l = !l;

	/*
	 *	Laut POCSAG-Spezifikation sind logische
	 *	Pegel folgendermassen zugeordnet:
	 *
	 *	0 -> f_t + f_h
	 *	1 -> f_t - f_h
	 *
	 *	Dabei ist f_t die Traegerfrequenz und f_h der Hub.
	 *
	 */
	if (!l) {
		CML0 = last_t2 & 0xff;
		CMH0 = last_t2 >> 8;
		STE |= TXD;
		CMI0 = 0;
		ECM0 = 1;
	} else {
		CML1 = last_t2 & 0xff;
		CMH1 = last_t2 >> 8;
		RTE |= TXD;
		CMI1 = 0;
		ECM1 = 1;
	} 
}


/*
 *	Die Transmitter-State-Machine
 *
 */
static void scc_xmitr_state_machine () using REGB_HP
{
	static unsigned int idata cntr;
	static stime_t t_start;
	static stime_t t_phase_start;
	register char b;
	
	switch (tx_state) {
	/*
	 *	States fuer Kalibrationsmodus
	 *
	 */
	case STATE_TX_CAL_START:
		t_start = systime;
		/* kein break! */

	case STATE_TX_CAL_RESTART:
		ON(PTT);
		tx_state = STATE_TX_CAL;
		t_phase_start = systime;
		/* kein break! */
	
	case STATE_TX_CAL:
		scc_set_txd(tx_bitlen,cntr & 1);
		cntr++;

		if (tx_cal_duration && systime - t_start >= tx_cal_duration) {
			tx_state = STATE_TX_OFF;
			break;
		}

		if (systime - t_phase_start > CAL_DURATION) {
			OFF(PTT);
			tx_state = STATE_TX_CAL_PAUSE;
			t_phase_start = systime;
		}

		break;

	case STATE_TX_CAL_PAUSE:
		/* TX-NF wird trotzdem weiterhin erzeugt! */
		scc_set_txd(tx_bitlen,cntr & 1);
		cntr++;

		if (systime - t_phase_start > CAL_PAUSE)
			tx_state = STATE_TX_CAL_RESTART;
		
		break;
	
	/*
	 *	States zum Aussenden von Funkrufen
	 *
	 */
	case STATE_TX_START:
		ON(PTT);
		tx_state = STATE_TX_HEAD;
		cntr = scc_txhead;
		/* kein break! */

	case STATE_TX_HEAD:
		if (cntr) {
			scc_set_txd(SCC_TC_10MS,1);
			cntr--;
		} else {
			scc_set_txd(tx_bitlen,0);
			tx_state = STATE_TX_PREAMBLE;

			/* 
			 *	das erste Bit der Praeambel wurde 
			 *	schon gesendet, deshalb "... - 1" 
			 *
			 */
			cntr = scc_preamble - 1;	
		}
		break;
	
	case STATE_TX_PREAMBLE:
		scc_set_txd(tx_bitlen,cntr & 1);
		
		if (cntr)
			cntr--;
		else	
			tx_state = STATE_TX_DATA;

		break;
		
	case STATE_TX_DATA:
		b = scc_gettxbit();

		if (b == -1) {
			scc_set_txd(tx_bitlen,0);
			cntr = scc_txtail;
			tx_state = STATE_TX_TAIL;
		} else
			scc_set_txd(tx_bitlen,b);
		break;

	case STATE_TX_TAIL:
		if (cntr) {
			scc_set_txd(SCC_TC_10MS,0);
			cntr--;
			break;
		}

		/* kein break! */

	case STATE_TX_OFF:
	default:
		/* Transmitter ausschalten */
		ECM0 = 0;
		ECM1 = 0;
		CMI0 = 0;
		CMI1 = 0;
		STE &= ~TXD;
		RTE &= ~TXD;
		
		tx_state = STATE_TX_IDLE;
		OFF(PTT);

		/* Receiver wieder freischalten */
		scc_recv_unlock();
	}
}


/*
 *	Wird aufgerufen, wenn TXD auf 1 gesetzt wurde
 *
 */
void scc_cm0_isr () interrupt INT_CM0 using REGB_HP
{
	ECM0 = 0;
	CMI0 = 0;

	STE &= ~TXD;
	
	T2 = 0;		/* clock fuer kommerzielle FSK-Sender */
	T2 = 1;
	
	scc_xmitr_state_machine();
}


/*
 *	Wird aufgerufen, wenn TXD auf 0 gesetzt wurde
 *
 */
void scc_cm1_isr () interrupt INT_CM1 using REGB_HP
{
	ECM1 = 0;
	CMI1 = 0;

	RTE &= ~TXD;
	
	T2 = 0;		/* clock fuer kommerzielle FSK-Sender */
	T2 = 1;
	
	scc_xmitr_state_machine();
}


/*--------------------------------------------------------------------------*/
/*									    */
/*			I N T E R F A C E				    */
/*									    */
/*--------------------------------------------------------------------------*/


/*
 *	Initialisiert den SCC
 *
 */
void scc_init ()
{
	/*
	 *	Timer ist schon durch das RTC-Modul
	 *	initialisiert!!!
	 *
	 * 	TM2CON = 0x05;	 T2_clk_src = Fosc/12, prescale = 2
	 */

	/*
	 *	Initialisierungen fuer den Receiver. Die Restlichen
	 * 	Einstellungen werden in den recv-Interface-Funktionen
	 *	durchgefuehrt.
	 *
	 */
	rx_state   = STATE_RX_IDLE;
	rx_enabled = 0;	/* Receiver ist aus */

	CTCON |= 0x03;	/* capture bei steigender und fallender */
			/* Flanke an CT0I aktivieren */

	PCT0 = 1;	/* capture interrupt 0 Prioritaet hoch */

	TR0 = 0;	/* T0 stop */
	TF0 = 0;	/* T0 overflow reset */
	ET0 = 0;	/* T0 interrupt abschalten */
	PT0 = 1;	/* T0 interrupt priority hight */
	TMOD &= 0xf0;	
	TMOD |= 0x01;	/* T0 16bit one-shot Timer */

	/*
	 *	Initialisierungen fuer den Transmitter
	 *
	 */
	tx_state = STATE_TX_IDLE;
	
	ECM0 = 0;	/* Compare-Interrupts abschalten */
	ECM1 = 0;

	CMI0 = 0;	/* Anstehende Compare-Interrupts verwerfen */
	CMI1 = 0;

	PCM0 = 1;	/* hohe Prioritaet fuer compare interrupts */
	PCM1 = 1;
}


/*
 *	Schaltet den Receiver an (on != 0) oder aus (on == 0).
 *
 */
void scc_recv_mode (unsigned char on)
{
	/*
	 *	Zunaechst wird der Receiver komplett abgeschaltet
	 *	auch dann, wenn er eigentlich eingeschaltet werden
	 *	soll. Das wird gemacht, um auf jeden Fall einen
	 *	bekannten Ausgangszustand zu haben. 
	 *
	 */
	TR0  = 0;	/* T0 anhalten */
	ET0  = 0;	/* T0 interrupt ausschalten */
	ECT0 = 0;	/* capture interrupt 0 ausschalten */
	CTI0 = 0;	/* Anstehende CT0 interrupts verwerfen */

	rx_enabled = 0;	/* Receiver aus */
	rx_state   = STATE_RX_IDLE;
	
	scc_dcd_init();

	/*
	 *	Wenn der Reciever abgeschaltet werden sollte, sind
	 *	wir hier bereits fertig.
	 *
	 */
	if (!on)
		return;
	
	/*
	 *	Receiver einschalten.
	 *
	 */
	scc_as_init();

	rx_buf       = NULL;
	rx_buflen    = 0;
	rx_buf_ready = 0;
	
	rx_state = STATE_RX_IDLE;
	scc_dcd_init();
		
	rx_enabled = 1;	/* Receiver ein */

	ET0  = 1;	/* T0 interrupts einschalten */
	ECT0 = 1;	/* capture interrupt 0 einschalten */
}


/*
 *	Setzt den Empfangspuffer, in den der Empfaenger seine
 *	Daten schreibt. 
 *
 */
char scc_recv (unsigned long xdata *buf, unsigned int size)
{
	if (!ET0)
		return 0;

	rx_buf        = buf;
	rx_buflen     = size;
	rx_buf_ready = 0;

	return 1;
}


/*
 *	Prueft, ob eine Aussendung empfangen wurde. Es wird
 *	die Anzahl der Empfangenen Codewoerter zurueckgegeben.
 *
 */
unsigned int scc_recv_poll (unsigned char xdata *speed)
{
	if (!rx_buf_ready)
		return 0;

	rx_buf_ready = 0;

	if (speed)
		*speed = rx_speed;

	return rx_buflen - rx_bufroom;
}


/*
 *	Sendet alle Funkrufe in der Page-Liste <page_list> mit der
 *	Geschwindigkeit <speed> (SCC_Bxxxx, etc.) aus.
 *	Wenn der Sender besetzt ist, wird 0 zurueckgegeben,
 *	ansonsten 1.
 *
 */
char scc_xmit (unsigned char speed, const struct PageQueue xdata *pq) 
{
	if (tx_state != STATE_TX_IDLE || rx_state != STATE_RX_IDLE)
		return 0;

	if (rx_enabled && rtc_time() - dcd_drop_time <= DCD_XMIT_DELAY)
		return 0;

	/* Receiver fuer den Sendedurchgang abschalten */
	scc_recv_lock();

	tx_page      = pq_first(pq);
	tx_cw_cnt    = tx_page->cw_cnt;
	tx_cw_buf    = tx_page->cw_buf;
	tx_bits_left = 0;
	tx_cw_cntr   = -1;
	tx_locked    = 0;
	
	tx_bitlen    = tc_tab[speed];

	tx_state = STATE_TX_START;
	CMI0 = 1;
	ECM0 = 1;

	return 1;
}


/*
 *	Startet den Kalibrationsmodus mit der Bitrate
 *	<bps> und der maximalen Dauer <duration> (in Sekunden).  
 *	Wenn <bps> gleich 0 ist, wird der Kalibrationsmodus abgeschaltet.
 *	Wenn <duration> gleich 0 ist, gibt es keine Zeitbeschraenkung.
 *
 *	Im Fehlerfall wird 0 zurueckgegeben, ansonsten 1.
 *
 */
char scc_calibrate (unsigned int bps, unsigned char duration)
{
	/*
	 *	Kalibrationsmodus abschalten?
	 *
	 */
	if (!bps) {
		switch (tx_state) {
		case STATE_TX_CAL_START:
		case STATE_TX_CAL:
		case STATE_TX_CAL_PAUSE:
			/*
			 *	Das Schreiben von tx_state ist atomar, daher
			 *	muessen die Interrupts hier nicht kurzzeitig
			 *	abgeschaltet werden.
			 */
			tx_state = STATE_TX_OFF;
			break;

		case STATE_TX_IDLE:
			break;

		default:
			return 0;
		}

		return 1;
	}

	/*
	 *	Kalibrationsmodus einschalten. Hier muss nicht
	 *	darauf geachtet werden, ob der Receiver laeuft,
	 *	schliesslich soll immer kalibriert werden koennen.
	 *
	 */
	if (tx_state != STATE_TX_IDLE)
		return 0;

	/* Receiver fuer den Sendedurchgang abschalten */
	scc_recv_lock();

	tx_cal_duration = (unsigned int)duration * 10;
	tx_bitlen       = SCC_BPS2TC(bps);

	tx_state = STATE_TX_CAL_START;
	CMI0 = 1;
	ECM0 = 1;

	return 1;
}


/*
 *	Gibt den Zustand des Transmitters zurueck:
 *
 *	0	Transmitter ist blockiert
 *		Entweder wird gerade gesendet oder empfangen.
 *
 *	!= 0	Transmitter ist zum Senden bereit.
 *
 */
char scc_xmitr_state ()
{
	return (tx_state == STATE_TX_IDLE);
}





