
/*
 *	$Id: sched.c,v 1.20 2002/05/12 13:00:44 klaus Exp $
 *
 *	sched.c		Funkruf-Scheduler. Dieses Modul koordiniert
 *			alle eintreffendenden Funkrufe und wickelt 
 *			die Zeitsteuerung des Senders ab.
 *
 *
 *
 *
 *	(c) 2001,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "sched.h"
#include "mm.h"
#include "rtc.h"
#include "scc.h"
#include "ax25_io.h"
#include "config.h"
#include "pocsag.h"
#include "page.h"
#include "trace.h"

/*
 *	Der Scheduler und der Repeater teilen sich diese
 *	Konfigurationsvariable.
 */
#define	sched_ident_addr	conf.ident_addr
#define sched_ident_text	conf.ident_text

/*
 *	Zustaende des Schedulers
 *
 */
#define STATE_IDLE		0
#define STATE_WAIT_FOR_XMIT	1
#define STATE_KICK_XMITR	2
#define STATE_WAIT_ON_XMIT	3
 
/*
 *	Aktueller Zustand der Scheduler-State-Machine
 *
 */
static unsigned char idata state;

/*
 *	Fuer jede unterstuetzte Baudrate wird eine eigene Page-Queue
 *	benoetigt. 
 *
 */
static struct PageQueue xdata pqtab[SCHED_SPEED_CNT];

/*
 *	Liste der im naechsten freien Slot auszusendenden Funkrufe +
 *	Sendegeschwindigkeit + erster Sendeslot (nur fuer die Status-
 *	ausgabe)
 *
 */
static struct PageQueue xdata tx_queue;
static unsigned char tx_speed;
static unsigned char tx_slot;

/*
 *	Slot-Tabelle + Bitmaske, die Anzeigt, in welchen Slots gesendet
 *	werden darf
 *
 */
static struct Slot xdata stab[SCHED_SLOT_CNT];
static unsigned int smask;

/*
 *	Tabelle wird zur Bestimmung der max. Batch-Anzahl bei einem
 *	Sendedurchgang gebraucht:
 *
 *	bmax_tab[speed][slot-Anzahl] =  max. Anzahl Batches, die hinter-
 *					einander gesendet weden duerfen.
 *
 *	In der Tabelle sind sowohl Preamble-Dauer, mindest-Sendepause
 *	(2 Sekunden), sowie max. Sendedauer (30 Sekunden) beruecksichtigt.
 *
 */
static unsigned char code bmax_tab[SCHED_SPEED_CNT][SCHED_SLOT_CNT + 1] = 
{
	{ 
		 0,					/* 512bps */
		 3,    9,  15,  21,  27,  27,  27,  27,  
	  	27,   27,  27,  27,  27,  27,  27,  27 
	},
	{
		 0,					/* 1200bps */
		 8,   22,  36,  51,  65,  65,  65,  65,  
	        65,   65,  65,  65,  65,  65,  65,  65 
	},
	{
		 0,					/* 2400bps */
		18,   46,  74, 103, 131, 131, 131, 131, 
	       131,  131, 131, 131, 131, 131, 131, 131 
	}
};

/*--------------------------------------------------------------------------*/


/*
 *	Initialisiert den Scheduler
 *
 */
void sched_init ()
{
	struct PageQueue xdata *pq;

	for (pq = pqtab; pq != pqtab + SCHED_SPEED_CNT; pq++)
		pq_init(pq);

	pq_init(&tx_queue);

	sched_set(0x0000);

	state = STATE_IDLE;
}


/*
 *	Setzt die Slots in denen der Scheduler den Sender anwerfen
 *	darf. Jeder Slot ist als Bit in <slot_mask> codiert.
 *	slot_mask == 0x0005 bedeutet demzufolge, dass in den Slots
 *	0 und 2 gesendet werden darf.
 *
 */
void sched_set (unsigned int slot_mask)
{
	struct Slot xdata *slot;
	struct Slot xdata *gap;
	unsigned char seq_len;

	smask = slot_mask;

	/*
	 *	Wenn alle Slots freigegeben sind kann die Slot-Tabelle
	 *	mit standard-Werten gefuellt werden.
	 *
	 */
	if (slot_mask == 0xffff) {
		for (slot = stab; slot != stab + SCHED_SLOT_CNT; slot++) {
			slot->tx_unlock = 1;
			slot->seq_len = SCHED_SLOT_CNT;
		}

		return;
	}

	/*
	 *	tx_unlock-Flags initialisieren und Sequenzlaengen
	 *	berechnen.
	 *
	 */
	for (slot = stab; slot != stab + SCHED_SLOT_CNT; slot++) {
		if ( !(slot_mask & 1) )
			gap = slot;
	
		slot->tx_unlock = slot_mask & 1;
		slot_mask >>= 1;
	}

	slot = gap;
	
	do {
		if (slot->tx_unlock)
			slot->seq_len = ++seq_len;
		else
			slot->seq_len = seq_len = 0;

		if (slot == stab)
			slot = stab + SCHED_SLOT_CNT - 1;
		else
			slot--;
	} while (slot != gap);
}


/*
 *	Verbietet dem Scheduler in allen Slots zu senden,
 *	deren korrespondierende Bits in <slot_mask>
 *	gesetzt sind.
 *
 */
void sched_lock (unsigned int slot_mask)
{
	sched_set(smask & ~slot_mask);
}


/*
 *	Erlaubt dem Scheduler in allen Slots zu senden,
 *	deren korrespondierende Bits in <slot_mask>
 *	gesetzt sind.
 *
 */
void sched_unlock (unsigned int slot_mask)
{
	sched_set(smask | slot_mask);
}


/*
 *	Gibt die Groesse des freien Speichers zurueck, der noch
 *	zum Speichern von Funkrufen verwendet werden kann.
 *
 */
static unsigned int sched_room ()
{
	if (mm_coreleft <= SCHED_BUSY_MARK)
		return 0;

	return mm_coreleft - SCHED_BUSY_MARK;
}


/*
 *	Fuegt den Funkruf <page> zu der durch <speed> angegebenen Page-Queue
 *	hinzu. Wenn <head> ungleich 0 ist, wird der Funkruf am Anfang der
 *	Queue hinzugefuegt, ansonsten am Ende.
 *
 *	Im Fehlerfall wird 0 zurueckgegeben, ansonsten 1.
 *
 */
char sched_addpage (unsigned char speed, struct Page xdata *page,
		    unsigned char head)
{
	if (!smask || !sched_room() || speed > SCHED_SPEED_MAX) 
		return 0;

	pq_enqueue(&pqtab[speed], page, head);

	return 1;
}


/*
 *	Gibt den Scheduler-Status aus.
 *
 */
void sched_print_state (struct AX25_cb xdata *axcb)
{
	struct Slot xdata *slot;
	unsigned char slot_nr;
	unsigned char sched_slot;
	unsigned char act_slot;
	char symbol;

	act_slot = SCHED_SLOT(rtc_time());

	switch (state) {
	case STATE_WAIT_FOR_XMIT:
		symbol = '+';
		sched_slot = tx_slot;
		break;
	
	case STATE_WAIT_ON_XMIT:
		symbol = '*';
		sched_slot = tx_slot;
		break;

	default:
		sched_slot = 0xff;
	}

	ax25_printf(axcb, "slot  state\r");

	slot_nr = 0;

	for (slot = stab; slot != stab + SCHED_SLOT_CNT; slot++) {
		char sym_left;
		char sym_right;

		sym_left  = (slot_nr == act_slot)   ? '>'    : ' ';
		sym_right = (slot_nr == sched_slot) ? symbol : ' ';
	
		ax25_printf(axcb, "%c%3u%c%6b\r",
			    (char)sym_left, 
			    (unsigned int)slot_nr, 
			    (char)sym_right,
			    (unsigned char)slot->tx_unlock);

		slot_nr++;
	}
}


/*
 *	Zeigt den Queue-Zustand des Schedulers an. Diese
 *	Ausgabe ist nur fuer debug-Zwecke interessant.
 *
 */
void sched_print_qstate (struct AX25_cb xdata *axcb)
{

	ax25_printf(axcb," 512  1200  2400  txq   room\r"
			 "%4u%6u%6u%5u%7u\r",
			 (unsigned int)pq_page_cnt(&pqtab[SCHED_SPEED_512]),
			 (unsigned int)pq_page_cnt(&pqtab[SCHED_SPEED_1200]),
			 (unsigned int)pq_page_cnt(&pqtab[SCHED_SPEED_2400]),
			 (unsigned int)pq_page_cnt(&tx_queue),
			 (unsigned int)sched_room());
}

/*--------------------------------------------------------------------------*/


/*
 *	Stoesst das Aussenden von Funkrufen an. Da in dieser Funktion
 *	aufgrund des Registerbank-Konflikts sched_alarm() <-> scc_xmit()
 * 	der Transmitter nicht direkt gestartet werden kann, wird das
 *	ueber die Scheduler-State-Machine abgewickelt. Diese Methode
 *	ist nicht ideal und kann ggf. zu Verzoegerungen fuehren.
 *
 */
static void sched_alarm () using REGB_LP
{
	if (state == STATE_WAIT_FOR_XMIT)
		state = STATE_KICK_XMITR;
}


/*
 *	Prueft nacheinander alle Page-Queues, ob Funkrufe anstehen.
 *	Wenn Funkrufe vorhanden sind, wird fuer einen Slot eine Aussendung
 *	zusammengestellt.
 *
 */
static void sched_state_idle ()
{
	static unsigned char speed = SCHED_SPEED_MAX;
	struct PageQueue xdata * idata pq;
	unsigned char slot;

	/*
	 *	Slot ermitteln, in dem fruehestens eine Aussendung
	 *	stattfinden kann. Wenn in dem errechneten Slot nicht 
	 *	gesendet werden darf, werden die PageQueues nicht geprueft. 
	 *
	 *	Die 5 in der Berechnung des naechst moeglichen Sendeslots
	 *	entspricht der Ausloesezeit des Watchdogs (500ms). Es wird
	 *	so der schlimmste Fall beruecksichtigt, dass die
	 *	Zusammenstellung der Funkrufe knapp 500ms dauern kann.
	 *
	 */
	slot = SCHED_SLOT(rtc_time() + 5 + SCHED_SLOT_DURATION);

	if (!stab[slot].tx_unlock)
		return;

	/*
	 *	Pruefen, ob ein Funkruf in der aktuellen 
	 *	Sendequeue steht.
	 *
	 */
	if (++speed > SCHED_SPEED_MAX)
		speed = SCHED_SPEED_MIN;

	pq = &pqtab[speed];

	if (!pq_first(pq))
		return;

	/*
	 *	Sendedaten generieren und RTC programmieren..
	 *	Wenn eine Ident-Adresse gesetzt ist, wird ein
	 *	Ident-Funkruf erzeugt.
	 *
	 */
	if (sched_ident_addr) {
		struct Page xdata *p;

		p = poc_encode(sched_ident_addr, 0, sched_ident_text);

		if (p != NULL) {
			p->domain = PAGE_DOMAIN_SYSTEM;
			pq_enqueue(&tx_queue, p, 1);
		}
	}

	pq_batch(&tx_queue, pq, bmax_tab[speed][stab[slot].seq_len]);

	if (!pq_first(&tx_queue))
		return;
		
	tx_slot  = slot;
	tx_speed = speed;

	state = STATE_WAIT_FOR_XMIT;

	rtc_alarm(slot, sched_alarm);
}


/*
 *	Kettet alle Funkrufe aus der tx_queue wieder in die zugehoerige
 *	Queue ein. Dabei werden alle SYSTEM-Funkrufe entfernt. 
 *
 */
static void sched_tx_resume ()
{
	pq_clear_dom(&tx_queue, PAGE_DOMAIN_SYSTEM);
	pq_insert(&pqtab[tx_speed], &tx_queue);
}


/*
 *	State-Machine des Schedulers
 *
 */
void sched_run ()
{
	switch (state) {
	case STATE_IDLE:
		sched_state_idle();
		break;

	case STATE_WAIT_FOR_XMIT:
		break;

	case STATE_KICK_XMITR:
		if (scc_xmit(tx_speed, &tx_queue)) {
			state = STATE_WAIT_ON_XMIT;
		} else {
			sched_tx_resume();
			state = STATE_IDLE;
		}
		
		break;
		
	case STATE_WAIT_ON_XMIT:
		if (scc_xmitr_state()) {
			trace_print_pq('T', tx_speed, &tx_queue);
			pq_clear(&tx_queue);
			state = STATE_IDLE;
		}

		break;

	default:
		state = STATE_IDLE;
	}
}



