
/*
 *	$Id: scp.c,v 1.7 2002/03/17 12:36:31 klaus Exp $
 *
 *	scp.c		Slave Control Protocol
 *		
 *
 *
 *	(c) 2001,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "scp.h"
#include "mm.h"
#include "ax25.h"
#include "ax25_io.h"
#include "string.h"
#include "sched.h"
#include "pocsag.h"
#include "system.h"
#include "passwd.h"
#include "master.h"
#include "version.h"
#include "timer.h"
#include "page.h"

/*
 *	Rueckgabewerte der Kommandoauswertungsfunktionen
 *
 */
#define E_SUCCESS	1
#define E_ERROR		2
#define E_RETRY		3

/*
 *	Laufdauern der SCP-Timer
 *
 */
#define T_TIMEOUT	30	/* in 10 Sekunden-Schritten */
#define T_ACK		5	/* in Sekunden */

/*
 *	Die unterschiedlichen Funkruf-Codierungstypen
 *
 */
#define ENC_NUMERIC	0
#define ENC_ALPHA	1
#define ENC_SCRAMBLED	2

/*
 *	SID (System ID)
 *
 *	Die SID ist in 3 Teile aufgeteilt
 *	
 *
 *	[RPC/XOS 2.1.0-SCP-#234567]
 *         (1)         (2)    (3)
 *
 *	(1)	Versions-String der Slave-Software/Firmware
 *	(2)	Kennzeichnung fuer das SCP (Slave Control Protocol)
 *	(3)	Unterstuetzte Protokollelemente
 *
 */
#define SCP_SID		"[" VERSION_STRING "-SCP-#2345678]"

/*
 *	Liste, in der alle SCP-Ports eingetragen sind
 *
 */
static struct SCP xdata * scptab = NULL;

/*
 *	Im naechsten Zyklus zu bearbeitender SCP-Port
 *
 */
static struct SCP xdata * scp_next = NULL;

/*--------------------------------------------------------------------------*/


/*
 *	Erzeugt einen neuen SCP-Port, initialisiert ihn und
 *	verbindet ihn mit dem uebergebenen AX25-Controlblock
 *	<axcb>
 *
 */
void scp_setup (struct AX25_cb xdata * axcb)
{
	struct SCP xdata * idata scp;
	struct Master xdata *m;

	m = m_lookup(&axcb->path.path.addr.peer);

	if (!m) {
		ax25_close(axcb);
		return;
	}

	scp = (struct SCP xdata *)malloc(sizeof(struct SCP));

	if (!scp) {
		/*
		 *	Wenn keine SCP-Instanz erzeugt werden kann,
		 *	wird die AX.25-Verbindung geschlossen.
		 *
		 */
		ax25_close(axcb);
		return;
	}

	/*
	 *	Magic an den Master schicken, damit der erkennt,
	 *	welche SCP-Version wir hier verwenden.
	 *
	 */
	ax25_printf(axcb, SCP_SID);

	if (*m->passwd) {
		if ( !(scp->passwd = pw_new(axcb, m->passwd, 0)) ) {
			free(scp);
			ax25_close(axcb);
			return;
		}
	} else {
		scp->passwd = NULL;
		ax25_printf(axcb, "\r");
	}

	scp->axcb    = axcb;
	scp->seq     = 0;

	t_set(&scp->t_tout, T_TIMEOUT, T_RES10S);
	t_set(&scp->t_ack,  T_ACK, T_RES1S);

	scp->next = scptab;
	scptab = scp;

	ON(LED_CON);

	t_start(&scp->t_tout);
}


/*
 *	Schliesst den SCP-Port <scp> und gibt alle von ihm
 *	belegten Resourcen wieder frei.
 *
 */
static void scp_shutdown (struct SCP xdata * scp)
{
	struct SCP xdata * idata s;
	struct SCP xdata * idata prev;

	if (scp == scp_next)
		scp_next = scp_next->next;
	
	prev = NULL;

	for (s = scptab; s != NULL && s != scp; s = s->next) 
		prev = s;

	if (!prev)
		scptab = scptab->next;
	else
		prev->next = scp->next;

	if (scp->passwd)
		pw_delete(scp->passwd);

	ax25_close(scp->axcb);

	t_stop(&scp->t_tout);
	t_stop(&scp->t_ack);

	free(scp);

	if (!scptab)
		OFF(LED_CON);	
}


/*
 *	Gibt alle scp-User aus
 *
 */
void scp_usage (struct AX25_cb xdata *axcb)
{
	struct SCP xdata *scp;

	ax25_printf(axcb, "scp:\r");

	for (scp = scptab; scp != NULL; scp = scp->next)
		ax25_printf(axcb, 
		            "%a\r",
			    (__XPTR__)&scp->axcb->path.path.addr.peer);
}


/*
 *	Gibt != 0 zurueck, wenn momentan ein Master
 *	connected ist.
 *
 */
char scp_connection ()
{
	return (scptab != NULL);
}


/*--------------------------------------------------------------------------*/


/*
 *	Generiert aus ASCII-Hex codierten Codewort-Sequenzen einen
 *	Funkruf.
 *
 *	Im Fehlerfall wird NULL zurueckgegeben, ansonsten der
 *	Zeiger auf den erzeugten Funkruf.
 *
 */
static struct Page xdata * scp_hexstr2page (unsigned char frame,
					    char xdata *hexstr)
{
	unsigned int len;
	struct Page xdata *page;
	unsigned long xdata *cw_buf;

	len = strlen_x(hexstr);

	/*
	 *	Jedes Codewort ist 32 Bits gross und wird als Hex-Zahl
	 *	durch 8 Ziffern representiert. Daher muss die Laenge 
	 *	von <hexstr> ein Vielfaches von 8 sein!
	 *
	 */
	if ( (len % 8) != 0 )
		return NULL;

	if ( !(page = page_new(len / 8, frame)) )
		return NULL;

	cw_buf = page->cw_buf;

	while (*hexstr) {
		char xdata *endptr;
		char xdata *tail;
		char saved_ch;

		tail = hexstr + 8;
		saved_ch = *tail;
		*tail = 0;

		*cw_buf++ = strtoul(hexstr, &endptr, 16);

		if (*endptr) {
			page_delete(page);
			return NULL;
		}
	
		*tail = saved_ch;
		hexstr = tail;
	}

	return page;
}


/*--------------------------------------------------------------------------*/


/*
 *	KOMMANDO 2
 *
 *	SYNTAX:		2:<ident>
 *	BEISPIEL:	2:echo1
 *
 *	RUECKGABE:	2:<ident>:<rpc system time>
 *
 *	Fuehrt ein Echo mit hinzugefuegter Systemzeit aus. <ident> darf
 *	nicht leer sein und darf auch keine Doppelpunkte enthalten.
 *
 */
static char scp_process_2 (struct SCP xdata *scp)
{
	char xdata *ident;

	if ( !(ident = strtok()) )
		return E_ERROR;

	ax25_printf(scp->axcb, 
		    "2:%s:%04X\r", 
		    (__XPTR__)ident, 
		    (unsigned int)rtc_time());

	return E_SUCCESS;
}


/*
 *	KOMMANDO 3
 *
 *	SYNTAX:		3:[+|-]<stime>
 *	BEISPIEL:	3:+1A
 *
 *	Korrigiert die Systemzeit um die angegebene Anzahl Zehntelsekunden.
 *	Wenn weder + noch - angegeben sind, wird die Systemzeit auf <stime>
 *	gesetzt.
 *
 */
static char scp_process_3 ()
{
	char xdata *adjstr;
	char xdata *endptr;
	char sign;
	unsigned long adj;
	stime_t stime;

	if ( !(adjstr = strtok()) )
		return E_ERROR;

	/*
	 *	Vorzeichen extrahieren.
	 *
	 */
	if (*adjstr == '+' || *adjstr == '-') {
		sign = *adjstr;
		adjstr++;
	} else
		sign = 0;

	/*
	 *	Korrekturwert extrahieren. Der
	 *	Korrekturwert darf nur 16-bittig sein!
	 *
	 */
	adj = strtoul(adjstr, &endptr, 16);

	if (*endptr)
		return E_ERROR;

	if (adj & 0xffff0000UL)
		return E_ERROR;

	/*
	 *	Systemzeit anpassen
	 *
	 */
	stime = rtc_time();
	
	switch (sign) {
	case '+':
		stime += (stime_t)adj;
		break;
	
	case '-':
		stime -= (stime_t)adj;
		break;
	
	default:
		stime = (stime_t)adj;
	}

	rtc_set(stime);

	return E_SUCCESS;	
}


/*
 *	KOMMANDO 4
 *
 *	SYNTAX:		4:<slots>
 *	BEISPIEL:	4:0123ABCD
 *
 *	Gibt die angegebenen Slots beim Scheduler fuer 
 *	Funkruf-Aussendungen frei.
 *
 */
static char scp_process_4 ()
{
	char xdata *slots;
	char xdata *endptr;
	unsigned int slot_mask;

	if ( !(slots = strtok()) )
		return E_ERROR;

	slot_mask = bstrtoui(slots, &endptr);

	if (*endptr)
		return E_ERROR;

	if (strtok_tail())
		return E_ERROR;

	sched_set(slot_mask);

	return E_SUCCESS;
}


/*
 *	KOMMANDOS 5, 6, 8
 *
 *	SYNTAX:		6:<speed>:<address>:<func>:<text>
 *
 *	mit		<speed> =	0 -> 512bps
 *					1 -> 1200bps
 *					2 -> 2400bps
 *
 *	BEISPIEL:	6:1:18272:0:ein kurzer Funkruf
 *
 *	Codiert den Funkruf <text> und uebergibt ihn an den
 *	Scheduler zum Aussenden. In <enc> muss der Codierungstyp
 *	angegeben werden.
 *
 */
static char scp_process_568 (unsigned char enc)
{
	char xdata *token;
	char xdata *endptr;
	unsigned char speed;
	unsigned long addr;
	unsigned char func;
	struct Page xdata *page;

	/*
	 *	Sendegeschwindigkeit extrahieren
	 *
	 */
	if ( !(token = strtok()) )
		return E_ERROR;
	
	speed = *token - '0';

	if (speed > 2)
		return E_ERROR;

	if (token[1])
		return E_ERROR;

	/*
	 *	Adresse extrahieren
	 *
	 */
	if ( !(token = strtok()) )
		return E_ERROR;

	addr = strtoul(token, &endptr, 16);

	if (*endptr)
		return E_ERROR;

	/*
	 *	Funktionsbits extrahieren
	 *
	 */
	if ( !(token = strtok()) )
		return E_ERROR;

	func = (unsigned char)strtoul(token, &endptr, 16);

	if (*endptr)
		return E_ERROR;

	if (func > 3)
		return E_ERROR;

	/*
	 *	Funkruftext extrahieren, codieren und an den
	 *	Scheduler uebergeben.
	 *
	 */
	if ( !(token = strtok_tail()) )
		return E_ERROR;

	switch (enc) {
	case ENC_NUMERIC:
		page = poc_encode_num(addr, func, token);
		break;
	
	case ENC_ALPHA:
		page = poc_encode_str(addr, func, token);
		break;
		
	case ENC_SCRAMBLED:
		page = poc_encode_scrambled(addr, func, token);
		break;
		
	default:
		return E_ERROR;
	}

	if (!page)
		return E_RETRY;

	if (!sched_addpage(speed, page, 0)) {
		page_delete(page);
		return E_RETRY;
	}

	return E_SUCCESS;
}


/*
 *	KOMMANDO 7
 *
 *	USAGE:		7:<speed>:<frame>:<ascii hex coded cws>
 *
 *	mit		<speed>	=	0 -> 512bps
 *					1 -> 1200bps
 *					2 -> 2400bps
 *
 *	BEISPIEL:	7:1:0:3A09265EFE022836
 *
 *	Gibt den ASCII-Hex codierten Funkruf zum Aussenden an den
 *	Scheduler weiter.
 *
 */
static char scp_process_7 ()
{
	char xdata *token;
	unsigned char speed;
	unsigned char frame;
	struct Page xdata *page;
	
	/*
	 *	Sendegeschwindigkeit extrahieren
	 *
	 */
	if ( !(token = strtok()) )
		return E_ERROR;
	
	speed = *token - '0';

	if (speed > 2)
		return E_ERROR;

	if (token[1])
		return E_ERROR;

	/*
	 *	Frame extrahieren
	 *
	 */
	if ( !(token = strtok()) )
		return E_ERROR;
	
	frame = *token - '0';

	if (frame > 7)
		return E_ERROR;

	if (token[1])
		return E_ERROR;

	/*
	 *	CW-Sequenz extrahieren, in einen Funkruf
	 *	umwandeln und an den Scheduler uebergeben.
	 *
	 */
	if ( !(token = strtok_tail()) )
		return E_ERROR;

	if ( !(page = scp_hexstr2page(frame, token)) )
		return E_ERROR;
	
	if (!sched_addpage(speed, page, 0)) {
		page_delete(page);
		return E_RETRY;
	}

	return E_SUCCESS;
}


/*
 *	Kommandozeile auswerten
 *	
 *	Die Rueckgabewerte haben folgende Bedeutung:
 *
 *	0		Nichts ausgewertet
 *	E_SUCCESS	Erfolgreich ausgewertet
 *	E_ERROR		Fehler beim Auswerten
 *	E_RETRY		Auswertung Ok aber Aktion im Moment
 *			nicht ausfuehrbar.
 *
 */
static char scp_process (struct SCP xdata *scp, char xdata * cmd)
{
	char rc;

	strtok_init(cmd, ":", 0);
	
	if (!strtok())
		return 0;

	switch (*cmd) {
	case '2':
		/* echo mit System-Zeit */
		rc = scp_process_2(scp);
		break;
	
	case '3':
		/* System-Zeit-Korrektur */
		rc = scp_process_3();
		break;
		
	case '4':
		/* Sende-Slots setzen */
		rc = scp_process_4();
		break;

	case '5':
		/* numerischen Funkruf aussenden */
		rc = scp_process_568(ENC_NUMERIC);
		break;

	case '6':
		/* Alphanumerisch codierten Funkruf aussenden */
		rc = scp_process_568(ENC_ALPHA);
		break;
		
	case '7':
		/* raw-codierten Funkruf aussenden */
		rc = scp_process_7();
		break;

	case '8':
		/* gescrambledten alphanumerischen Funkruf aussenden */
		rc = scp_process_568(ENC_SCRAMBLED);
		break;
	
	default:
		rc = E_ERROR;
	}

	/*
	 *	Nur nach erfolgreich ausgefuehrten Kommandos
	 *	wird der Timeout-Timer zurueckgesetzt.
	 *
	 */
	if (rc == E_SUCCESS || rc == E_RETRY)
		t_start(&scp->t_tout);

	return rc;
}


/*
 *	Bearbeitet alle asynchronen Kommandos. Wenn ein Protokoll-Fehler
 *	aufgetreten ist, wird 0 zurueckgegeben, ansonsten 1.
 *
 */
static char scp_process_async (struct SCP xdata *scp, char xdata *cmd)
{
	char xdata *str;
	char xdata *endptr;
	unsigned long seq;
	
	/*
	 *	Das erste Zeichen in der Zeile ist ein #. Da wir
	 *	das bereits wissen. kann es entfernt werden.
	 *
	 */
	cmd++;

	strtok_init(cmd, " \t", 1);

	if ( !(str = strtok()) )
		return 0;

	seq = strtoul(str, &endptr, 16);

	if (*endptr)
		return 0;
		
	/*
	 *	Die Sequenznummer darf nur im Bereich von
	 *	0 bis 255 liegen.
	 *
	 */
	if (seq	& 0xffffff00UL)
		return 0;

	/*
	 *	Master und Slave asynchron?
	 *
	 */
	if ((unsigned char)seq != scp->seq)
		return 0;

	scp->seq++;

	/*
	 *	Wenn hinter der Sequenznummer kein Befehl folgt
	 *	kann so eine sofortige Bestaetigung aller bisher
	 *	eingetroffenen Kommandos erzwungen werden. Ansonsten
	 *	wuerden T_ACK Sekunden gewartet bevor
	 *	die Bestaetigung geschickt wird.
	 *	Mit dieser Methode kann der Durchsatz auf dem Link
	 *	erhoeht werden.
	 *
	 */
	if ( !(str = strtok_tail()) ) {
		ax25_printf(scp->axcb, "#%02X +\r", (unsigned int)scp->seq);
		t_stop(&scp->t_ack);
		return 1;
	}
	
	t_start(&scp->t_ack);

	switch (scp_process(scp, str)) {
	case 0:
		break;
	
	case E_ERROR:
		ax25_printf(scp->axcb, "#%02X -\r", (unsigned int)seq);
		break;

	case E_RETRY:
		ax25_printf(scp->axcb, "#%02X %%\r", (unsigned int)seq);
		break;
	}

	return 1;
}


/*
 *	Bearbeitet die SCP-Timer. Wenn die SCP-Instanz
 *	entsorgt wurde, wird 1 zurueckgegeben, ansonsten 0.
 *
 */
static char scp_timer (struct SCP xdata *scp)
{
	if (t_isexpired(&scp->t_ack))
		ax25_printf(scp->axcb, 
			    "#%02X +\r", 
			    (unsigned int)scp->seq);

	if (t_isexpired(&scp->t_tout)) {
		ax25_printf(scp->axcb, "! timeout\r");
		scp_shutdown(scp);
		return 1;
	}

	return 0;
}


/*--------------------------------------------------------------------------*/


/*
 *	Wickelt haeppchenweise das Slave Control Protokoll ab
 *	
 */
void scp_run ()
{
	struct SCP xdata * idata scp;
	char xdata * str;

	if (!scp_next) {
		if (!scptab)
			return;

		scp_next = scptab;
	}

	scp = scp_next;
	scp_next = scp_next->next;

	switch (ax25_event(scp->axcb)) {
		case EVENT_TXQEMPTY:
		/*
		 *	Nur wenn die TX-Queue leer ist, werden
		 *	Kommandos bearbeitet.
		 *
		 */
		break;

	case EVENT_LINKRESET:
	case EVENT_LINKFAILURE:
	case EVENT_DISCONNECT:
		scp_shutdown(scp);
		return;

	default:
		return;
	}

	/*
	 *	Passwort-Abfrage
	 *
	 */
	if (scp->passwd) {
		switch (pw_run(scp->passwd)) {
		case 0:
			break;
			
		case 1:
			pw_delete(scp->passwd);
			scp->passwd = NULL;
			ax25_printf(scp->axcb, "+\r");
			break;
		
		default:
			ax25_printf(scp->axcb, "-\r");
			scp_shutdown(scp);
		}

		return;
	}
	
	if (scp_timer(scp))
		return;

	/*
	 *	Kommando bearbeiten
	 *
	 */
	if ( !(str = ax25_gets(scp->axcb, 512)) )
		return;

	/*
	 *	asynchrone Kommandos werden gesondert behandelt
	 *
	 */
	if (*str == '#') {
		if (!scp_process_async(scp, str)) {
			ax25_printf(scp->axcb, 
			            "! protocol error (seq=%02X)\r",
				    (unsigned int)scp->seq);

			scp_shutdown(scp);
		}
	} else
		switch (scp_process(scp, str)) {
		case E_SUCCESS:
			ax25_printf(scp->axcb, "+\r");
			break;
			
		case E_ERROR:
			ax25_printf(scp->axcb, "-\r");
			break;

		case E_RETRY:
			ax25_printf(scp->axcb, "%%\r");
		}

	free(str);
}


