
/*
 *	$Id: timer.c,v 1.1.1.1 2002/02/15 22:36:40 Klaus Exp $
 *
 *	timer.c		Funktionen zum Verwalten von Timern
 *
 *			Der Tackt fuer die Timer-Funktionen wird
 *			von der RTC (Real Time Clock) bezogen. Der
 *			Tackt wird mit einer Aufloesung von 100ms
 *			angeliefert.
 *
 * 			Es werden unterschiedliche Timer-Aufloesungen 
 *			benutzt, damit so der Timeralgorithums effizienter 
 *			arbeiten kann und das MCS51-System nicht so stark
 *			belastet wird.
 *
 *
 *	(c) 2000,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "system.h"
#include "timer.h"
#include "mm.h"
#include "types.h"

/*
 *	Tabellen fuer die unterschiedlichen Timeraufloesungen 
 *	100ms, 1s, 10s.
 *
 */
static struct TimerTab idata ttab[3] = { 
	{ NULL, 0 },
	{ NULL, 0 },
	{ NULL, 0 }
};

/*
 *	Timer-Register muss extern incrementiert werden
 *
 */
volatile unsigned char idata t_reg = 0; 

/*---------------------------------------------------------------------------*/


/*
 *	traegt den Timer <t> aus der entsprechenden Timer-Tabelle
 *	aus.
 *
 *	In dieser Funktion muss t->next unveraendert bleiben, da
 *	sonst t_release() in einer Schleife, die ueber alle Timer 
 *	laeuft nicht verwendet werden kann!
 *
 */
static void t_release (struct Timer xdata *t)
{
	struct Timer xdata * idata tp;

	if (t->ttab->first == t) {
		t->ttab->first = t->next;
		return;
	}

	for (tp = t->ttab->first; tp != NULL; tp = tp->next)
		if (tp->next == t) {
			tp->next = t->next;
			return;
		}
	
	/*
	 *	Hmm, die Funktion duerfte niemals an dieser Stelle
	 *	beendet werden. Wenn das doch passiert, ist die
	 *	Timer-Tabelle oder eine Funktion nicht korrekt.
	 *
	 */
}


/*
 *	Bearbeitet alle Timer in der Tabelle <tt>. Dabei wird
 *	der Wert <val> von allen Timer->values abgezogen. Wenn
 *	ein Timer abgelaufen ist, wird er aus der Timer-Tabelle
 *	ausgetragen.
 *
 */
static void t_process (struct TimerTab idata *tt,unsigned char val)
{
	struct Timer xdata * idata tnext;

	tnext = tt->first;

	while (tnext) {
		struct Timer xdata * idata t;
	
		t = tnext;
		tnext = tnext->next;

		if (t->value > val) {
			t->value -= val;
			continue;
		}
		
		t_release(t);
		t->state = T_EXPIRED;
	}
}


/*
 *	Diese Funktion muss zyklisch aufgerufen werden. Sie
 *	dekrementiert ggf. Timer->values und aendert den Zustand
 *	eines Timers, wenn er abgelaufen ist.
 *
 */
void t_run ()
{
	unsigned char idata reg;
	struct TimerTab idata * idata tt; 

	reg = t_reg;
	t_reg = 0;

	tt = ttab;

	while (reg) {
		t_process(tt,reg);

		tt++;

		if (tt == ttab + 3)
			break;

		tt->reg += reg;
		reg = tt->reg / 10;
		tt->reg %= 10;
	}
}


/*---------------------------------------------------------------------------*/


/*
 *	Initialisiert den Timer <t> mit den Werten <initial> und <res>	
 *
 */
void t_set (struct Timer xdata *t,unsigned char initial,unsigned char res)
{
	t->state      = T_STOPPED;
	t->initial    = initial;
	t->ttab       = &ttab[res];
	t->next       = NULL;
}


/*
 *	Startet den Timer <t>
 *
 */
void t_start (struct Timer xdata *t) 
{
	t->value = t->initial;

	/*
	 *	Wenn der Timer noch in einer Tabelle gefuehrt wird,
	 *	darf er nicht erneut eingetragen werden!
	 *	
	 */
	if (t->state == T_RUNNING)
		return;

	t->next = t->ttab->first;
	t->ttab->first = t;
	t->state = T_RUNNING;
}


/*
 *	Haelt den Timer <t> an.
 *
 */
void t_stop (struct Timer xdata *t)
{
	if (t->state == T_RUNNING)
		t_release(t);

	t->state = T_STOPPED;
}


/*
 *	Liefert != 0 zurueck, wenn der Timer abgelaufen ist.
 *
 */
char t_isexpired (struct Timer xdata *t)
{
	if (t->state == T_EXPIRED) {
		t->state = T_STOPPED;
		return 1;
	}
	
	return 0;
}


/*
 *	Liefert != 0 zurueck, wenn der Timer <t> laeuft.
 *
 */
char t_isrunning (struct Timer xdata *t)
{
	return (t->state == T_RUNNING);
}



