
/*
 *	$Id: trace.c,v 1.8 2002/05/12 12:35:04 Klaus Exp $
 *
 *	trace.c		Modul zum Mitlesen von gesendeten und Empfangenen
 *			Funkrufen.
 *			Das Trace-Modul kann nur an eine einzige Infobox
 *			gebunden werden und unterscheidet sich deshalb auch
 *			etwas im Interface von den anderen Infobox-Modulen.
 *
 *
 *
 *	(c) 2002,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "trace.h"
#include "ax25.h"
#include "page.h"
#include "string.h"
#include "pocsag.h"
#include "mm.h"
#include "scc.h"

/*
 *	 Zeiger auf den AX.25-Controlblock ueber den der Trace
 *	 ausgegeben werden soll. Wenn axcb == NULL ist, wird
 *	 nichts ausgegeben.
 *
 */
static struct AX25_cb xdata *axcb = NULL;

/*--------------------------------------------------------------------------*/


/*
 *	Aktiviert das Trace-Modul und laesst alle Trace-Ausgaben
 *	ueber den AX.25-Controlblock <_axcb> laufen. Wenn der Trace
 *	bereits aktiv ist, wird 0 zurueckgegeben, ansonsten 1.
 *
 */
char trace_new (struct AX25_cb xdata *_axcb)
{
	if (axcb)
		return 0;

	axcb = _axcb;

	return 1;
}


/*
 *	Beendet den laufenden Trace.
 *
 */
void trace_delete ()
{
	axcb = NULL;
}


/*
 *	Wandelt die <speed> Angabe in Bit pro Sekunde um.
 *
 */
static unsigned long trace_speed_to_bps (unsigned char speed)
{
	switch (speed) {
	case SCC_B512:
		return 512;

	case SCC_B1200:
		return 1200;

	case SCC_B2400:
		return 2400;
	}

	return 0;
}


/*
 *	Wandelt den Funkruf <p> in Klartext um und gibt
 *	diesen aus. Damit der Funkruf zugeordnet werden kann,
 *	wird dem Funkruf das Zeichen <prefix> vorangestellt.
 *
 */
static void trace_print_page (char prefix, 
			      unsigned char speed,
			      const struct Page xdata *p)
{
	unsigned long xdata addr;
	unsigned char xdata func;
	char xdata *str;
	char type;

	/*
	 *	zunaechst wird die Adresse decodiert
	 *	und ausgegeben. Danach wird entschieden,
	 *	wie der Rest zu decodieren ist.
	 *
	 */
	poc_decode_addr(p, &addr, &func);

	ax25_printf(axcb, "%c%04u %07lu.%u ",
		    (char)prefix,
		    (unsigned int)trace_speed_to_bps(speed),
		    (unsigned long)addr,
		    (unsigned int)func);

	/*
	 *	Zur Wahl der korrekten Decodierung werden folgende
	 *	Annahmen gemacht:
	 *
	 *	Adressinterval  Funktionsbits	Codierung
	 *	------------------------------------------------------
	 *	4000..5000	3		scrambled alphanumeric
	 *	2000..3000	0		numeric
	 *	*               *		alphanumerisch
	 *      ------------------------------------------------------
	 *
	 */
	if (addr >= 4000 && addr <= 5000 && func == 3) {
		str = poc_decode_scrambled(p);
		type = '$';
	} else if (addr >= 2000 && addr <= 3000 && func == 0) {
		str = poc_decode_num(p);
		type = '#';
	} else {
		str = poc_decode_str(p);
		type = '@';
	}

	/*
	 *	Klartext-Funkruf formatiert ausgeben.
	 *
	 */
	if (str == NULL) {
		ax25_printf(axcb, "*\r");
	} else {
		ax25_printf(axcb, "%c %s\r", (char)type, (__XPTR__)str);
		free(str);
	}
}


/*
 *	Gibt den Inhalt der Page-Queue <pq> als Klartext aus.
 *	Damit die einzelnen Funkrufe zuordbar sind, wird der
 *	Prefix <prefix> vorangestellt.
 *
 */
void trace_print_pq (char prefix, 
		     unsigned char speed,
		     const struct PageQueue *pq)
{
	struct Page xdata *p;

	if (!axcb)
		return;

	ax25_printf(axcb, "\r");

	for ( p = pq_first(pq); p != NULL; p = pq_next(p) )
		trace_print_page(prefix, speed, p);
}


/*
 *	Wenn das Tracen aktiviert ist, decodiert diese Funktion die
 *	Page-Sequenz <pseq> in Klartext und gibt ihn ueber den
 *	aktuellen AX.25-Controlblock aus. Um Die Sequenz zuordnen
 *	zu koennen, wird jedem Funkruf das Zeichen <prefix>
 *	vorangestellt.
 *
 */
void trace_print_ps (char prefix, 
		     unsigned char speed,
		     const struct PageSequence xdata *pseq)
{
	struct PageSequence xdata ps;
	struct Page xdata *p;

	if (!axcb)
		return;

	memcpy_xx(&ps, pseq, sizeof(ps));

	ax25_printf(axcb, "\r");

	/*
	 * 	Page-Sequenz in einzelne Funkrufe zerlegen, 
	 *	decodieren und danach ausgeben.
	 *
	 */
	while ( (p = poc_splitoff(&ps)) != NULL ) {
		trace_print_page(prefix, speed, p);
		page_delete(p);			    
	}
}


/*
 *	Sobald ueber den AX.25-Controlblock etwas gelesen
 *	werden kann, wird der Trace-Modus abgebrochen.
 *	Wenn der Trace-Modus beendet ist, wird 1 zurueckgegeben,
 *	ansonsten 0.
 *
 */
char trace_run ()
{
	char xdata *str;
	
	if (!axcb)
		return 1;

	if ( !(str = ax25_gets(axcb, 10)) )
		return 0;

	free(str);
	trace_delete();

	return 1;
}




