
/*
 *	$Id: upload.c,v 1.3 2002/04/29 10:45:27 klaus Exp $
 *
 *	upload.c	Modul zum uploaden der Firmware ueber AX.25
 *			Die Daten muessen beim upload mit dem
 *			Autobin-Protokoll angeliefert werden.
 *
 *
 *
 *
 *	(c) 2001,	Klaus Hennemann, DL3KHB
 *
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation; either version 2, or (at your option) any
 *	later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *	In other words, you are welcome to use, share and improve this program.
 *	You are forbidden to forbid anyone else to use, share and improve
 *	what you give them.   Help stamp out software-hoarding!  
 */

#include "upload.h"
#include "mm.h"
#include "ax25.h"
#include "ax25_io.h"
#include "flash.h"
#include "string.h"
#include "crc.h"
#include "buf.h"
#include "system.h"
#include "boot.h"

/*
 *	Zustaende des Upload-Objekts
 *
 */
#define STATE_START_MSG		0
#define STATE_WAIT_FOR_BIN	1
#define STATE_RECV_DATA		2
#define STATE_SUCCESS		3	
#define STATE_FAILURE		4

/*--------------------------------------------------------------------------*/


/*
 *	Erzeugt ein neues Upload-Objekt und gibt den Zeiger darauf
 *	zurueck. Wenn kein Objekt erzeugt werden konnte, wird
 *	NULL zurueckgegeben.
 *
 */
struct Upload xdata * up_new (struct AX25_cb xdata *axcb, unsigned char bank)
{
	struct Upload xdata *up;

	if (IS_CSEG_FLASH())
		return NULL;

	if (bank != BOOT_FLASH0 && bank != BOOT_FLASH1)
		return NULL;
	
	if ( !(up = malloc(sizeof(struct Upload))) )
		return NULL;

	up->state = STATE_START_MSG;
	up->axcb  = axcb;
	up->flash_bank  = bank;

	return up;
}


/*
 *	Loescht das Upload-Objekt. 
 *
 */
void up_delete (struct Upload xdata *up)
{
	/*
	 *	Autobin-Rueckstaende loeschen.
	 *
	 */
	ax25_clean_io(up->axcb);

	free(up);	
}


/*
 *	Wartet darauf, dass der User den #BIN#-Header schickt.
 *	Wenn der Header erkannt wurde, wird er ausgewertet und
 *	der Upload gestartet.
 *
 */
static void up_wait_for_bin (struct Upload xdata *up)
{
	char xdata *str;
	char xdata *token;
	unsigned char n;
	unsigned long size, crc;

	if ( !(str = ax25_gets(up->axcb, 80)) )
		return;

	for (n = 0;n < 5; n++)
		if (str[n] != "#BIN#"[n]) {
			free(str);
			return;
		}

	/*
	 *	Bin-Header auswerten. Es werden lediglich die
	 *	Filegroesse und die CRC aus dem Header extrahiert.
	 *
	 *	Header-Format:
	 *
	 *	#BIN#<size>#|<crc>#$<dos time>#<file name>\r
	 *
	 */
	strtok_init(str, "#", 1);
	strtok();		/* BIN wegwerfen */

	if ( (token = strtok()) != NULL ) {
		char xdata *endptr;

		size = strtoul(token, &endptr, 10);

		if (*endptr)
			size = 0;
	}

	while ( (token = strtok()) != NULL )
		if (*token == '|') {
			char xdata *endptr;

			crc = strtoul(token + 1, &endptr, 10);

			if (!*endptr)
				break;
		}

	free(str);
	
	/*
	 *	Es wird eine CRC benoetigt, um den Transfer abzusichern.
	 *
	 */
	if (!token || (crc & 0xffff0000UL) ) {
		ax25_printf(up->axcb, "\r#NO#\rinvalid crc\r");
		up->state = STATE_FAILURE;
		return;
	}

	/*
	 *	die Filegroesse muss im Interval [1 .. UP_BINLEN_MAX]
	 *	liegen, damit das Uploaden durchgefuehrt wird.
	 *
	 */
	if (!size || size > UP_BINLEN_MAX) {
		ax25_printf(up->axcb, "\r#NO#\rinvalid file size\r");
		up->state = STATE_FAILURE;
		return;
	}

	/*
	 *	Alles fuer den Autobin-Empfang vorbereiten
	 *
	 */
	up->state    = STATE_RECV_DATA;
	
	up->flash_addr = 0;
	
	up->bin_crc  = (unsigned int)crc;
	up->bin_size = size;
	up->bin_left = up->bin_size;
	
	up->crc      = CRC_THP_INITIAL;

	flash_erase(up->flash_bank);

	ax25_printf(up->axcb, "#OK#\r");
}


/*
 *	Schreibt alle empfangenen Daten in die entsprechende
 *	Flash-Bank.
 *
 */
static void up_recv_data (struct Upload xdata *up)
{
	struct Buffer xdata *buf;
	unsigned char xdata *dat;
	unsigned int size;

	if ( !(buf = ax25_recv_buf(up->axcb)) )
		return;

	dat = buf_data(buf);
	size = buf_size(buf);

	if (size > up->bin_left)
		size = up->bin_left;

	flash_write(up->flash_bank, up->flash_addr, dat, size);
	up->crc = crc_thp_buf(up->crc, dat, size);
	
	up->flash_addr += size;
	up->bin_left -= size;

	buf_delete(buf);

	/*
	 *	Sind wir noch nicht fertig?
	 *
	 */
	if (up->bin_left)
		return;

	if (up->crc == up->bin_crc) {
		ax25_printf(up->axcb, "transfer successful\r");
		boot_enable_bank(up->flash_bank, up->bin_crc);
		up->state = STATE_SUCCESS;
	} else {
		ax25_printf(up->axcb, "transfer failed\r");
		flash_erase(up->flash_bank);
		boot_disable_bank(up->flash_bank);
		up->state = STATE_FAILURE;
	}
}


/*
 *	Upload State-Machine ausfuehren
 *
 *	Rueckgabe:
 *	-1	upload fehlgeschlagen
 *	0	upload laeuft noch
 *	1	upload erfolgreich
 *
 */
char up_run (struct Upload xdata *up)
{
	switch (up->state) {
	case STATE_START_MSG:
		ax25_printf(up->axcb, "please upload firmware\r");
		up->state = STATE_WAIT_FOR_BIN;
		break;

	case STATE_WAIT_FOR_BIN:
		up_wait_for_bin(up);
		break;

	case STATE_RECV_DATA:
		up_recv_data(up);
		break;

	case STATE_SUCCESS:
		return 1;

	case STATE_FAILURE:
	default:
		return -1;
	}

	return 0;
}




